/*ScianCalculator.c
  Eric Pepke
  March 17,  1994

  The calculator maintains a variety of semantic classes which can be created
  by the parser and later evaluated.  Evaluation normally proceeds recursively.
  Maybe some day I'll write a compiler.
*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianArrays.h"
#include "ScianIDs.h"
#include "ScianEvents.h"
#include "ScianRecorders.h"
#include "ScianColors.h"
#include "ScianWindows.h"
#include "ScianObjWindows.h"
#include "ScianDialogs.h"
#include "ScianVisWindows.h"
#include "ScianSliders.h"
#include "ScianTimers.h"
#include "ScianMethods.h"
#include "ScianLists.h"
#include "ScianErrors.h"
#include "ScianGarbageMan.h"
#include "ScianSpaces.h"
#include "ScianIcons.h"
#include "ScianSockets.h"
#include "ScianObjFunctions.h"
#include "ScianGlobalFunctions.h"
#include "ScianCurrentFunctions.h"
#include "ScianWindowFunctions.h"
#include "ScianRepobjFunctions.h"
#include "ScianHelp.h"
#include "ScianStyle.h"
#include "ScianNames.h"
#include "ScianTextFiles.h"
#include "ScianDatabase.h"
#include "ScianSymbols.h"
#include "ScianControls.h"
#include "ScianScripts.h"
#include "ScianFixedClasses.h"
#include "ScianParser.h"
#include "ScianCalculator.h"
#include "ScianPointers.h"
#include "ScianFiles.h"
#include "ScianFileSystem.h"

/*Semantic classes*/
ObjPtr windowVarClass;			/*Class for window reference variables*/
ObjPtr classVarClass;			/*Class for class reference variables*/
ObjPtr varReferenceClass;		/*Class for var reference*/
ObjPtr typeCastClass;			/*Class for a type cast*/
ObjPtr scriptVarClass;			/*Class for a script variable*/
ObjPtr codeSequenceClass;		/*Class for a sequence of code*/
ObjPtr setVarClass;			/*Class for variable assignment*/
ObjPtr setObjVarClass;			/*Class for setting object variable*/
ObjPtr fileReadClass;			/*Class for a file read*/

ObjPtr allCalcFunctions;		/*All the calculator functions*/

ObjPtr curFileCommand = NULLOBJ;	/*Current file command*/

static Bool calcErrorMessages = true;	/*True iff calc error should produce messages*/

int calcErrorsInhibited = 0;		/*Nonzero if calc errors inhibited*/

#ifdef HAVE_PROTOTYPES
void InhibitCalcErrors(Bool whether)
#else
void InhibitCalcErrors(whether)
Bool whether;
#endif
/*Chooses whether or not to inhibit calc errors*/
{
    if (whether)
    {
	++calcErrorsInhibited;
    }
    else
    {
	--calcErrorsInhibited;
    }
}

#ifdef HAVE_PROTOTYPES
void AddHistory(ObjPtr object, ObjPtr newHist)
#else
void AddHistory(object, newHist)
ObjPtr object;
ObjPtr newHist;
#endif
/*Adds historical event newHist to object*/
{
    ObjPtr history;

    if (!newHist) return;

    history = GetVar(object, HISTORY);
    if (!history)
    {
	history = NewList();
    }
    history = AddToSet(history, newHist);
    SetVar(object, HISTORY, history);
}

#ifdef HAVE_PROTOTYPES
void EmitCalcMessages(Bool whether)
#else
void EmitCalcMessages(whether)
Bool whether;
#endif
/*Sets whether or not to emit calc messages*/
{
    calcErrorMessages = whether;
}

#ifdef HAVE_PROTOTYPES
void CalcError(char *msg)
#else
void CalcError(msg)
char *msg;
#endif
/*Emits a calculation error*/
{
    if (!calcErrorsInhibited)
    {
	ParseError(msg);
    }
}

#ifdef HAVE_PROTOTYPES
ObjPtr TypeCast(ObjPtr object, int type)
#else
ObjPtr TypeCast(object, type)
ObjPtr object;
int type;
#endif
/*Casts object to type.  Returns cast object or ERROBJ*/
{
    ObjPtr retVal;

    retVal = ERROBJ;

    switch (type)
    {
	case TC_NULL:
	    retVal = object;
	    break;
	case TC_INTEGER:
	    if (IsInt(object))
	    {
		retVal = object;
	    }
	    else if (IsReal(object))
	    {
		retVal = NewInt((int) (GetReal(object)));
	    }
	    break;
	case TC_SHORTARRAY:
	    if (IsShortArray(object))
	    {
		retVal = object;
	    }
	    break;
	case TC_BYTEARRAY:
	    if (IsByteArray(object))
	    {
		retVal = object;
	    }
	    break;
	case TC_REALARRAY:
	    if (IsRealArray(object))
	    {
		retVal = object;
	    }
	    break;
	case TC_POINTERARRAY:
	    if (IsPointerArray(object))
	    {
		retVal = object;
	    }
	    break;
	case TC_OBJECTARRAY:
	    if (IsObjArray(object))
	    {
		retVal = object;
	    }
	    break;
    }

    return retVal;
}

#ifdef HAVE_PROTOTYPES
ObjPtr Eval(ObjPtr object)
#else
ObjPtr Eval(object)
ObjPtr object;
#endif
/*Evaluates an object*/
{
    FuncTyp method;

    if (!object) return NULLOBJ;
    if (!IsObject(object)) return object;

    method = GetMethod(object, EVAL);
    if (method)
    {
	return (*method)(object);
    }
    else
    {
	return object;
    }
}

#ifdef HAVE_PROTOTYPES
char *EvalExpressionArg(char *s, ObjPtr *o)
#else
char *EvalExpressionArg(s, o)
char *s;
ObjPtr *o;
#endif
/*
    Parses an expression at s and evaluates it
    Returns
	s if the object was syntactically or semantically bad or null
	the first character after the object if it was OK.
	Puts in o the object if it was semantically valid
        Leavs o alone if it was not
    Generates a script error if syntactically or semantically invalid.
*/
{
    char *r;

    r = ParseExpression(s, o);
    if (*o != ERROBJ)
    {
	*o = Eval(*o);
    }
    return r;
}

#ifdef HAVE_PROTOTYPES
void AppendStatementToCode(ObjPtr codeSeq, ObjPtr stmt)
#else
void AppendStatementToCode(codeSeq, stmt)
ObjPtr codeSeq;
ObjPtr stmt;
#endif
/*Appends statement stmt to codeSeq*/
{
    ObjPtr var;

    var = GetListVar("AppendStatementToCode", codeSeq, ARGUMENTS);
    if (!var) var = NewList();
    PostfixList(var, stmt);
    SetVar(codeSeq, ARGUMENTS, var);
}

/*****  Creation functions for semantic classes  *****/

/* These all generate error messages through ParseError
   and return NULLOBJ if bad */

#ifdef HAVE_PROTOTYPES
ObjPtr NewCalcFunction(char *name, ObjPtr args)
#else
ObjPtr NewCalcFunction(name, args)
char *name;
ObjPtr args;
#endif
/*Creates a new function with name and args if possible*/
{
    ThingListPtr runner;
    ObjPtr var;
    int nArgs;
    char msg[200];
    ObjPtr retVal;

    for (runner = LISTOF(allCalcFunctions); runner; runner = runner -> next)
    {
	var = GetStringVar("NewCalcFunction", runner -> thing, NAME);
	if (var)
	{
	    if (strcmp2(GetString(var), name) == 0) break;
	}
    }
    if (runner)
    {
	/*A function has been found!*/
	var = GetIntVar("NewCalcFunction", runner -> thing, NARGS);
	if (var)
 	{
	    nArgs = ListCount(args);
	    if (GetInt(var) == nArgs || GetInt(var) == -1)
	    {
		retVal = NewObject(runner -> thing, 0L);
		if (GetInt(var) == -1 || nArgs >= 3)
		{
		    SetVar(retVal, ARGUMENTS, args);
		}
		else if (nArgs == 1)
		{
		    SetVar(retVal, ARGUMENT1, LISTOF(args) -> thing);
		}
		else if (nArgs == 2)
		{
		    SetVar(retVal, ARGUMENT1, LISTOF(args) -> thing);
		    SetVar(retVal, ARGUMENT2, LISTOF(args) -> next -> thing);
		}
		return retVal;
	    }
	    else
	    {
		sprintf(msg, "%d argument%s expected, %d %s given", 
		    GetInt(var), (GetInt(var) == 1) ? " was" : "s were",
		    nArgs, (nArgs == 1) ? " was" : "s were");
		ParseError(msg);
	    }
	}
    }
    else
    {
	sprintf(msg, "There is no function %s", name);
	ParseError(msg);
    }
    return NULLOBJ;
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewWindowVariable(WinInfoPtr win, char *name)
#else
ObjPtr NewWindowVariable(win, name)
WinInfoPtr win;
char *name;
#endif
/*Returns a new window variable*/
{
    ObjPtr retVal;
    retVal = NewObject(windowVarClass, 0L);
    SetVar(retVal, NAME, NewString(name));
    SetVar(retVal, OWNERWINDOW, (ObjPtr) win);
    return retVal;
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewClassVariable(int classID, char *name)
#else
ObjPtr NewClassVariable(classID, name)
int classID;
char *name;
#endif
/*Returns a new class variable*/
{
    ObjPtr retVal;

    retVal = NewObject(classVarClass, 0L);
    SetVar(retVal, NAME, NewString(name));
    SetVar(retVal, REPCLASSID, NewInt(classID));
    return retVal;
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewScriptVariable(ScriptVarPtr var)
#else
ObjPtr NewScriptVariable(var)
ScriptVarPtr var;
#endif
/*Returns a new script variable*/
{
    ObjPtr retVal;
    retVal = NewObject(scriptVarClass, 0L);
    SetVar(retVal, WHICHVAR, NewPointer(var));
    return retVal;
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewVarReference(ObjPtr object, NameTyp varID)
#else
ObjPtr NewVarReference(object, varID)
ObjPtr object;
NameTyp varID;
#endif
/*Returns a new variable reference, which evals to the value of the variable*/
{
    ObjPtr retVal;
    retVal = NewObject(varReferenceClass, 0L);
    SetVar(retVal, WHICHVAR, NewSymbol(varID));
    SetVar(retVal, ARGUMENT1, object);
    return retVal;
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewTypeCast(ObjPtr object, int typ)
#else
ObjPtr NewVarReference(object, typ)
ObjPtr object;
int typ;
#endif
/*Returns a new type cast*/
{
    ObjPtr retVal;
    retVal = NewObject(typeCastClass, 0L);
    SetVar(retVal, ARGUMENT2, NewInt(typ));
    SetVar(retVal, ARGUMENT1, object);
    return retVal;
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewCodeSequence(void)
#else
ObjPtr NewCodeSequence()
#endif
/*Returns a new code sequence*/
{
    ObjPtr retVal;

    retVal = NewObject(codeSequenceClass, 0L);
    SetVar(retVal, ARGUMENTS, NewList());

    return retVal;
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewSetVar(ScriptVarPtr var, ObjPtr value)
#else
ObjPtr NewSetVar(var, value)
ScriptVarPtr var;
ObjPtr value;
#endif
/*Returns a new set var*/
{
    ObjPtr retVal;

    retVal = NewObject(setVarClass, 0L);
    SetVar(retVal, ARGUMENT1, NewPointer(var));
    SetVar(retVal, ARGUMENT2, value);

    return retVal;
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewSetObjVar(ObjPtr object, NameTyp whichVar, ObjPtr value)
#else
ObjPtr NewSetVar(object, whichVar, value)
ObjPtr object;
NameTyp whichVar;
ObjPtr value;
#endif
/*Returns a new set object var*/
{
    ObjPtr retVal;

    retVal = NewObject(setObjVarClass, 0L);
    SetVar(retVal, ARGUMENT1, object);
    SetVar(retVal, ARGUMENT2, value);
    SetVar(retVal, WHICHVAR, NewSymbol(whichVar));

    return retVal;
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewFileRead(ObjPtr path, ObjPtr fileName, ObjPtr fileFormat)
#else
ObjPtr NewFileRead(object, fileName, fileFormat)
ObjPtr object;
ObjPtr fileName;
ObjPtr fileFormat;
#endif
/*Returns a new file read semantic class*/
{
    ObjPtr retVal;

    retVal = NewObject(fileReadClass, 0L);
    SetVar(retVal, NAME, fileName);
    SetVar(retVal, PATH, path);
    SetVar(retVal, FORMAT, fileFormat);

    return retVal;
}

/*****  EVAL methods for semantic classes  *****/

static ObjPtr EvalWindowVar(var)
ObjPtr var;
/*Evaluates a window variable*/
{
    ObjPtr list, win, name;
    char msg[300];

    win = GetVar(var, OWNERWINDOW);
    if (!win) return ObjFalse;
    name = GetStringVar("EvalWindowVar", var, NAME);
    if (!name) return ObjFalse;

    list = FindNamedObject(win, GetString(name));
    if ((!list) || !LISTOF(list))
    {
	ObjPtr winName;

	winName = GetVar(win, NAME);
	sprintf(msg, "There is no object named '%s' in window %s.\n",
	    GetString(name), winName ? GetString(winName) : "?");
	CalcError(msg);
	return ERROBJ;
    }
    else if (LISTOF(list) -> next)
    {
	ObjPtr winName;

	winName = GetVar(win, NAME);
	sprintf(msg, "Name '%s' is not unique in window %s.\n",
	    GetString(name), winName ? GetString(winName) : "?");
	CalcError(msg);
	return ERROBJ;
    }
    else
    {
	return LISTOF(list) -> thing;
    }
}

static ObjPtr EvalClassVar(var)
ObjPtr var;
/*Evaluates a class variable, effectively looking it up in the database*/
{
    ObjPtr list, classID, name;
    char msg[300];
    ObjPtr keyList;

    name = GetStringVar("EvalClassVar", var, NAME);
    if (!name) return ObjFalse;
    classID = GetIntVar("EvalClassVar", var, REPCLASSID);
    if (!classID) return ObjFalse;

    keyList = NewList();
    PostfixList(keyList, NewSymbol(NAME));
    PostfixList(keyList, name);
    PostfixList(keyList, NewSymbol(CLASSID));
    PostfixList(keyList, classID);

    list = SearchDatabase(keyList);
    if ((!list) || !LISTOF(list))
    {
	sprintf(msg, "There is no object named '%s' of class %s in the database.\n",
	    GetString(name), ClassIDToName(GetInt(classID)));
	CalcError(msg);
	return ERROBJ;
    }
    else if (LISTOF(list) -> next)
    {
	sprintf(msg, "Object '%s' of class %s is not unique in the database.\n",
	    GetString(name), ClassIDToName(GetInt(classID)));
	CalcError(msg);
	return ERROBJ;
    }
    else
    {
	return LISTOF(list) -> thing;
    }
}

static ObjPtr EvalVarReference(ref)
ObjPtr ref;
/*Evaluates a variable reference, which means getting the variable*/
{
    NameTyp id;
    ObjPtr argument, var;

    var = GetSymbolVar("EvalVarReference", ref, WHICHVAR);
    if (!var) return ERROBJ;
    id = GetSymbolID(var);

    argument = GetVar(ref, ARGUMENT1);
    argument = Eval(argument);

    if (argument && argument != ERROBJ)
    {
	return GetVar(argument, id);
    }
    else
    {
	return argument;
    }
}

static ObjPtr EvalTypeCast(tc)
ObjPtr tc;
/*Evaluates a type cast*/
{
    int typ;
    ObjPtr argument, var;

    var = GetIntVar("EvalTypeCast", tc, ARGUMENT2);
    if (!var) return ERROBJ;
    typ = GetInt(var);

    argument = GetVar(tc, ARGUMENT1);
    argument = Eval(argument);

    if (argument && argument != ERROBJ)
    {
	return TypeCast(argument, typ);
    }
    else
    {
	return argument;
    }
}

static ObjPtr EvalScriptVar(object)
ObjPtr object;
/*Evaluates a script variable*/
{
    ObjPtr var;
    ScriptVarPtr varPtr;

    var = GetVar(object, WHICHVAR);
    if (var)
    {
	varPtr = GetPointer(var);
	if (varPtr -> value == ERROBJ)
	{
	    /*Try to evaluate the definition*/
	    Eval(varPtr -> definition);
	}
	if (varPtr -> value == ERROBJ)
	{
	    char mes[200];
	    sprintf(mes, "Variable '%s' is undefined\n", varPtr -> name);
	    CalcError(mes);
	}
	return varPtr -> value;
    }
    else
    {
	return ERROBJ;
    }
}

static ObjPtr EvalCodeSequence(seq)
ObjPtr seq;
/*Evaluates a code sequence*/
{
    ObjPtr var;
    ThingListPtr runner;

    var = GetListVar("EvalCodeSequence", seq, ARGUMENTS);
    for (runner = LISTOF(var); runner; runner = runner -> next)
    {
	Eval(runner -> thing);
    }

    return ObjTrue;
}


static ObjPtr EvalSetVar(setVar)
ObjPtr setVar;
/*Evaluates a set var*/
{
    ObjPtr var;
    ObjPtr value;
    ScriptVarPtr scriptVar;

    var = GetVar(setVar, ARGUMENT1);
    if (!var) return ObjFalse;

    scriptVar = GetPointer(var);

    value = GetVar(setVar, ARGUMENT2);
    value = Eval(value);

    if (value != ERROBJ)
    {
	scriptVar -> value = value;
    }

    return value;
}

static ObjPtr EvalNewFunction(object)
ObjPtr object;
/*Evaluates a new(classname) function*/
{
    ObjPtr var;
    ObjPtr arg;
    ObjPtr retVal;

    arg = GetVar(object, ARGUMENT1);
    if (arg)
    {
	var = GetVar(arg, CLASSID);
	if (var)
	{
	    if (GetInt(var) == CLASS_CLASSVAR)
	    {
		/*It's a class variable.  Create a new one.*/
		var = GetVar(arg, REPCLASSID);
		if (var)
		{
		    retVal = GetFixedClassObject(GetInt(var));
		    if (retVal)
		    {
			retVal = NewObject(retVal, 0L);
			SetVar(retVal, NAME, GetVar(arg, NAME));

			return retVal;
		    }	    
		}
	    }
	    else
	    {
		/*Evaluate it and make a new one*/
		retVal = Eval(arg);
		if (retVal && retVal != ERROBJ)
		{
		    return NewObject(retVal, 0L);
		}
		else
		{
		    return retVal;
		}
	    }
	}
    }
    return NULLOBJ;
}

static ObjPtr EvalSetObjVar(setVar)
ObjPtr setVar;
/*Evaluates a setObjVar class*/
{
    ObjPtr object, value;
    ObjPtr var;
    NameTyp whichVar;

    object = Eval(GetVar(setVar, ARGUMENT1));
    value = Eval(GetVar(setVar, ARGUMENT2));
    var = GetSymbolVar("EvalSetObjVar", setVar, WHICHVAR);
    if (!var) return NULLOBJ;
    whichVar = GetSymbolID(var);

    SetVar(object, whichVar, value);

    return ObjTrue;
}

static ObjPtr EvalFileRead(fileRead)
ObjPtr fileRead;
/*Evaluates a file read*/
{
    ObjPtr path, fileName, fileFormat;
    ObjPtr retVal;
    ObjPtr rdr;

    path = Eval(GetVar(fileRead, PATH));
    if (!IsString(path))
    {
	CalcError("The path of a file read must be a string");
	return ObjFalse;
    }

    fileName = Eval(GetVar(fileRead, NAME));
    if (!IsString(fileName))
    {
	CalcError("The name of a file read must be a string");
	return ObjFalse;
    }

    fileFormat = Eval(GetVar(fileRead, FORMAT));
    if (!IsString(fileFormat))
    {
	CalcError("The format of a file read must be a string");
	return ObjFalse;
    }
 
    if ((rdr = FindFormatReader(GetString(fileFormat))) == NULLOBJ)
    {
	char msg[250];
	sprintf(msg, "File format '%s' does not exist", GetString(fileFormat));
	CalcError(msg);
	return ObjFalse;
    }
    if (!ChangeDir(GetString(path)))
    {
	return ObjFalse;
    }

    curFileCommand = fileRead;
    retVal = ReadFormattedFile(rdr, GetString(fileName)); /* fcn returns ObjTrue/False */
    curFileCommand = NULLOBJ;
    
    RestoreDir();

    LogObject(fileRead);

    return retVal;
}

/*****  Log methods for semantic classes  *****/

static ObjPtr LogFileRead(fileRead)
ObjPtr fileRead;
/*Logs a file read*/
{
    ObjPtr path, fileName, fileFormat;
    ObjPtr retVal;
    ObjPtr rdr;

    path = (GetVar(fileRead, PATH));

    fileName = (GetVar(fileRead, NAME));

    fileFormat = (GetVar(fileRead, FORMAT));

    InhibitWindow(true);
    Log("read ");
    LogObject(path);
    Log(", ");
    LogObject(fileName);
    Log(", ");
    LogObject(fileFormat);
    Log("\n");
    InhibitWindow(false);

}

#ifdef HAVE_PROTOTYPES
void InitCalculator(void)
#else
void InitCalculator()
#endif
/*Initializes the calculator*/
{
    ObjPtr func;

    allCalcFunctions = NewList();
    AddToReferenceList(allCalcFunctions);

    windowVarClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(windowVarClass);
    SetMethod(windowVarClass, EVAL, EvalWindowVar);

    classVarClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(classVarClass);
    SetMethod(classVarClass, EVAL, EvalClassVar);
    DefineFixedClass(classVarClass, CLASS_CLASSVAR, "classVar");

    scriptVarClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(scriptVarClass);
    SetMethod(scriptVarClass, EVAL, EvalScriptVar);

    varReferenceClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(varReferenceClass);
    SetMethod(varReferenceClass, EVAL, EvalVarReference);

    typeCastClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(typeCastClass);
    SetMethod(typeCastClass, EVAL, EvalTypeCast);

    codeSequenceClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(codeSequenceClass);
    SetMethod(codeSequenceClass, EVAL, EvalCodeSequence);

    setVarClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(setVarClass);
    SetMethod(setVarClass, EVAL, EvalSetVar);

    setObjVarClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(setObjVarClass);
    SetMethod(setObjVarClass, EVAL, EvalSetObjVar);

    fileReadClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(fileReadClass);
    SetMethod(fileReadClass, EVAL, EvalFileRead);
    SetMethod(fileReadClass, LOGEXPR, LogFileRead);

    /*Functions*/
    func = NewObject(NULLOBJ, 0L);
    AddToReferenceList(func);
    SetVar(func, NAME, NewString("new"));
    SetVar(func, NARGS, NewInt(1));
    SetMethod(func, EVAL, EvalNewFunction);
    PrefixList(allCalcFunctions, func);
}

#ifdef HAVE_PROTOTYPES
void KillCalculator(void)
#else
void KillCalculator()
#endif
/*Kills the calculator*/
{
    RemoveFromReferenceList(fileReadClass);
    RemoveFromReferenceList(setObjVarClass);
    RemoveFromReferenceList(setVarClass);
    RemoveFromReferenceList(codeSequenceClass);
    RemoveFromReferenceList(typeCastClass);
    RemoveFromReferenceList(varReferenceClass);
    RemoveFromReferenceList(classVarClass);
    RemoveFromReferenceList(windowVarClass);
    RemoveFromReferenceList(allCalcFunctions);
}
