/*ScianColorControls.c
  Eric Pepke
  2 April 1993

  Color controls in SciAn
*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianIDs.h"
#include "ScianArrays.h"
#include "ScianWindows.h"
#include "ScianObjWindows.h"
#include "ScianControls.h"
#include "ScianButtons.h"
#include "ScianColors.h"
#include "ScianEvents.h"
#include "ScianErrors.h"
#include "ScianScripts.h"
#include "ScianPictures.h"
#include "ScianDialogs.h"
#include "ScianSliders.h"
#include "ScianScales.h"
#include "ScianDatasets.h"
#include "ScianStyle.h"
#include "ScianTitleBoxes.h"
#include "ScianTextBoxes.h"
#include "ScianLists.h"
#include "ScianMethods.h"
#include "ScianDraw.h"
#include "ScianIcons.h"
#include "ScianFiles.h"
#include "ScianVisWindows.h"
#include "ScianObjFunctions.h"
#include "ScianTemplates.h"
#include "ScianTemplateHelper.h"
#include "ScianSciences.h"
#include "ScianSymbols.h"
#include "ScianGarbageMan.h"
#include "ScianFontSystem.h"
#include "ScianDepend.h"
#include "ScianColorControls.h"
#include "ScianHelp.h"

static ObjPtr swatchClass;
static ObjPtr mappingControlClass;
static ObjPtr componentControlClass;

/*Separate color components*/
#define SEP_RED		0	/*Red*/
#define SEP_GREEN	1	/*Green*/
#define SEP_BLUE	2	/*Blue*/
#define SEP_HUE		3	/*Hue (HSV or HLS)*/
#define SEP_SATURATION	4	/*Saturation (HSV or HLS)*/
#define SEP_VALUE	5	/*Value*/
#define SEP_LIGHTNESS	6	/*Lightness*/
#define SEP_Y		7	/*Y*/
#define SEP_I		8	/*I*/
#define SEP_Q		9	/*Q*/
#define NSEPCOMPONENTS	10

int sepComponents[NCOLORMODELS][3] =
    {
	{SEP_RED,	SEP_GREEN,	SEP_BLUE	},
	{SEP_HUE,	SEP_SATURATION,	SEP_VALUE	},
	{SEP_HUE,	SEP_LIGHTNESS,	SEP_SATURATION	},
	{SEP_Y,		SEP_I,		SEP_Q		},
    };

float componentColors[256][NSEPCOMPONENTS][3];

#ifdef HAVE_PROTOTYPES
void RGB2RGB(real *r2, real *g2, real*b2, real r, real g, real b)
#else
void RGB2RGB(r2, g2, b2, r, g, b)
real *r2, *g2, *b2, r, g, b;
#endif
/*Just copies r, g, b to r2, g2, b2.  Pretty dull*/
{
    *r2 = r;
    *g2 = g;
    *b2 = b;
}

#ifdef HAVE_PROTOTYPES
typedef void (*CCFunc)(real *, real *, real*, real, real, real);
#else
typedef void (*CCFunc)();
#endif

static struct
    {
	CCFunc ToRGBFunc;
	CCFunc FromRGBFunc;
    } colorModelInfo[NCOLORMODELS] =
    {
	{RGB2RGB, RGB2RGB},
	{HSV2RGB, RGB2HSV},
	{HLS2RGB, RGB2HLS},
	{YIQ2RGB, RGB2YIQ}
    };

#ifdef HAVE_PROTOTYPES
ObjPtr NewIndexedSwatch(int l, int r, int b, int t, char *name, ObjPtr repObj, NameTyp whichVar, int index)
#else
ObjPtr NewIndexedSwatch(l, r, b, t, name, repObj, whichVar, index)
int l, r, b, t;
char *name;
ObjPtr repObj;
NameTyp whichVar;
int index;
#endif
/*Creates an indexed swatch, which prints out the color in the index entry of
whichVar of its repobj, which is presumed to be an array of short3.  
If index is negative, it's assumed to be from the end + 1 of the array*/
{
    ObjPtr retVal;

    retVal = NewObject(swatchClass, 0L);
    Set2DIntBounds(retVal, l, r, b, t);
    SetVar(retVal, NAME, NewString(name));
    SetVar(retVal, WHICHVAR, NewSymbol(whichVar));
    SetVar(retVal, WHICHINDEX, NewInt(index));
    SetVar(retVal, REPOBJ, repObj);
    DeclareIndirectDependency(retVal, APPEARANCE, REPOBJ, whichVar);

    return retVal;
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewDirectSwatch(int l, int r, int b, int t, char *name, ObjPtr repObj, NameTyp whichVar)
#else
ObjPtr NewDirectSwatch(l, r, b, t, name, repObj, whichVar)
int l, r, b, t;
char *name;
ObjPtr repObj;
NameTyp whichVar;
#endif
/*Creates a direct swatch, which prints out the color in 
whichVar of its repobj, which is presumed to be an array of three
reals.*/
{
    ObjPtr retVal;

    retVal = NewObject(swatchClass, 0L);
    Set2DIntBounds(retVal, l, r, b, t);
    SetVar(retVal, NAME, NewString(name));
    SetVar(retVal, WHICHVAR, NewSymbol(whichVar));
    SetVar(retVal, REPOBJ, repObj);
    DeclareIndirectDependency(retVal, APPEARANCE, REPOBJ, whichVar);

    return retVal;
}

static ObjPtr DrawSwatch(swatch)
ObjPtr swatch;
/*Draws a swatch*/
{
    int l, r, b, t;
    ObjPtr repObj;
    ObjPtr var;
    NameTyp whichVar;
    int index;

    Get2DIntBounds(swatch, &l, &r, &b, &t);

    repObj = GetVar(swatch, REPOBJ);
    if (!repObj)
    {
	return ObjFalse;
    }

    var = GetVar(swatch, WHICHVAR);
    if (var)
    {
	whichVar = GetSymbolID(var);
    }
    else
    {
	return ObjFalse;
    }

    var = GetVar(swatch, WHICHINDEX);
    if (var)
    {
	/*It's indexed*/
	short3 *elements;
	float clr[3];

	index = GetInt(var);

	MakeVar(repObj, whichVar);
	var = GetVar(repObj, whichVar);
	if ((!IsShortArray(var)) || (RANK(var) != 2) || (DIMS(var)[1] != 3))
	{
	    ReportError("DrawSwatch", "Bad color array variable");
	    return ObjFalse;
	}
	elements = ELEMENTS(var);
	if (index < 0) index += DIMS(var)[0];

	if (index < 0 || index >= DIMS(var)[0])
	{
	    sprintf(tempStr, "Index %d out of range %d\n",
		index, DIMS(var)[0]);
	    ReportError("DrawSwatch", tempStr);
	    return ObjFalse;
	}
	clr[0] = ITOFC(elements[index][0]);
	clr[1] = ITOFC(elements[index][1]);
	clr[2] = ITOFC(elements[index][2]);
	RGBC(clr);
    }
    else
    {
	/*It isn't indexed*/
	var = GetVar(repObj, whichVar);
	if (var)
	{
	    SetObjectColor(var);
	}
	else
	{
	    SetUIColor(UIBACKGROUND);
	}
    }

    OptimizeColor();
    FillRect(l, r, b, t);
    OptimizeSharpness();
    SetUIColor(UIBLACK);
    FrameRect(l, r, b, t);
    OptimizeSharpness();

    return ObjTrue;
}

static MakeMethod MakeSwatchAppearance(swatch)
ObjPtr swatch;
/*Makes a swatch's appearance*/
{
    SetVar(swatch, APPEARANCE, ObjTrue);
    ImInvalid(swatch);
    return ObjTrue;
}

static MakeMethod MakeMappingControlAppearance(control)
ObjPtr control;
/*Makes a mapping control's appearance*/
{
    SetVar(control, APPEARANCE, ObjTrue);
    ImInvalid(control);

    return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
static void IndentMappingBounds(ObjPtr control, int *l, int *r, int *b, int *t)
#else
static void IndentMappingBounds(control, l, r, b, t)
ObjPtr control;
int flags; int *l; int *r; int *b; int *t;
#endif
/*Adjusts the mapping bounds to the interior of the control*/
{
    int lb = 0, rb = 0, bb = 0, tb = 0;
    ObjPtr var;
    int flags;
    short *scaleMargin, *extraMargin;

    var = GetVar(control, CONTROLFLAGS);
    if (!var)
    {
	return;
    }
    flags = GetInt(var);

    var = GetVar(control, SCALEMARGIN);
    if (!var)
    {
	ReportError("IndentMappingBounds", "No scale margin");
	return;
    }
    scaleMargin = ELEMENTS(var);

    var = GetVar(control, EXTRAMARGIN);
    if (!var)
    {
	ReportError("IndentMappingBounds", "No extra margin");
	return;
    }
    extraMargin = ELEMENTS(var);

    lb = extraMargin[0] + ((flags & MC_SCALELEFT) ? scaleMargin[0] : 0);
    rb = extraMargin[1] + ((flags & MC_SCALERIGHT) ? scaleMargin[1] : 0);
    bb = extraMargin[2] + ((flags & MC_SCALEBOTTOM) ? scaleMargin[2] : 0);
    tb = extraMargin[3] + ((flags & MC_SCALETOP) ? scaleMargin[3] : 0);

    if ((flags & MC_SCALETOP) || (flags & MC_SCALEBOTTOM))
    {
	lb = MAX(lb, scaleMargin[0] / 2 - 10);
	rb = MAX(rb, scaleMargin[1] / 2 - 10);
    }
    *l += lb;
    *r -= rb;
    *b += bb;
    *t -= tb;
}

#ifdef HAVE_PROTOTYPES
static int ValueToPixel(real value, real min, real max, int pMin, int pMax)
#else
static int ValueToPixel(value, min, max, pMin, pMax)
real value;
real min;
real max;
int pMin;
int pMax;
#endif
/*Returns a pixel value from a value, from pMin to pMax - 1*/
{
    int offset;

    offset = (int) (rfloor((pMax - pMin) * (value - min) /
	 (max - min)));

    return (offset >= pMax - pMin) ? pMax - 1 : pMin + offset;
}

#ifdef HAVE_PROTOTYPES
static int ValueToRectPixel(real value, real min, real max, int pMin, int pMax)
#else
static int ValueToRectPixel(value, min, max, pMin, pMax)
real value;
real min;
real max;
int pMin;
int pMax;
#endif
/*Returns a rectpixel value from a value, from pMin to pMax - 1*/
{
    int offset;

    offset = (int) (rfloor((pMax - pMin + 1) * (value - min) /
	 (max - min)));

    return (offset > pMax - pMin) ? pMax : pMin + offset;
}

#ifdef HAVE_PROTOTYPES
static real PixelToValue(int pixel, real min, real max, int pMin, int pMax)
#else
static real PixelToValue(pixel, min, max, pMin, pMax)
int pixel;
real min;
real max;
int pMin;
int pMax;
#endif
/*Returns a pixel value from a value, from pMin to pMax - 1*/
{
    int offset;
    real retVal;

    offset = pixel - pMin;

    retVal = ((real) offset) / ((real) (pMax - pMin - 1)) * (max - min) + min;
    if (retVal < min) return min;
    if (retVal > max) return max;
    return retVal;
}

#ifdef HAVE_PROTOTYPES
static int MapPixel(int pixel, int pMin1, int pMax1, int pMin2, int pMax2)
#else
static int MapPixel(pixel, pMin1, pMax1, pMin2, pMax2)
int pixel;
int pMin1;
int pMax1;
int pMin2;
int pMax2;
#endif
/*Maps pixel in range pMin1 to pMax1 to within range pMin2 to pMax2*/
{
    return pMin2 + floor(((real) (pixel - pMin1)) / ((real) (pMax1 - pMin1))
	* ((real) (pMax2 - pMin2)));
}

static ObjPtr DrawMapping(control)
ObjPtr control;
/*Draws the mapping within a mapping control*/
{
    int l, r, b, t;
    int flags;
    ObjPtr var;
    ObjPtr value;
    real *elements;
    real range[2];
    real domain[2];
    int nEntries;
    int left, right, bottom, top;

    Get2DIntBounds(control, &left, &right, &bottom, &top);
    var = GetVar(control, CONTROLFLAGS);
    if (var)
    {
	flags = GetInt(var);
    }
    else
    {
	flags = 0;
    }

    /*Get the interior*/
    l = left + EDGE + MINORBORDER;
    r = right - EDGE - MINORBORDER;
    b = bottom + EDGE + MINORBORDER;
    t = top - EDGE - MINORBORDER;

    IndentMappingBounds(control, &l, &r, &b, &t);

    value = GetValue(control);
    if (value)
    {
	/*There's a value, let's draw it.*/
	elements = ELEMENTS(value);

	MakeVar(control, NENTRIES);
	var = GetIntVar("DrawMapping", control, NENTRIES);
	if (!var)
	{
	    return ObjFalse;
	}
	nEntries = GetInt(var);

	var = GetFixedArrayVar("DrawMapping", control, XDOMAIN, 1, 2L);
	if (!var) return ObjFalse;
	Array2CArray(domain, var);

	var = GetFixedArrayVar("DrawMapping", control, YRANGE, 1, 2L);
	if (!var) return ObjFalse;
	Array2CArray(range, var);

	if (0)
	{
	    /*Draw white rectangles*/
	    int x1, x2, y, k;

	    x1 = l;

	    SetUIColor(UIWHITE);

	    for (k = 0; k < nEntries; ++k)
	    {
		x2 = (k == nEntries - 1) ? r : MapPixel(k + 1, 0, nEntries, l, r);

		y = ValueToRectPixel(elements[k], range[0], range[1], b, t);
		if (x2 > x1)
		{
		    FillRect(x1, x2, y, t);
		}
		x1 = x2;
	    }
	}
	else
	{
	    /*Draw lines from value centers*/
	    int x1, y1, x2, y2;
	    int k;

	    x1 = l;
	    y1 = ValueToPixel(elements[0], range[0], range[1], b, t);

	    for (k = 0; k < nEntries; ++k)
	    {
		x2 = MapPixel(k* 2 + 1, 0, nEntries * 2, l, r);
		y2 = ValueToPixel(elements[k], range[0], range[1], b, t);
		DrawUILine(x1, y1, x2, y2, UIBLACK);
		x1 = x2;
		y1 = y2;
	    }
	    DrawUILine(x1, y1, r - 1, y1, UIBLACK);
	}
    }

    /*Draw the mapping box*/
    FrameUIRect(l - 1, r + 1, b - 1, t + 1, UIBLACK);
    SetUIColor(UIBACKGROUND);
    FrameWideRect(l - 5, r + 5, b - 5, t + 5, 4);

    return ObjTrue;
}

static ObjPtr PressMappingControl(control, x, y, flags)
ObjPtr control;
int x, y;
long flags;
/*Does a press in a mapping control beginning at x and y.  Returns
  true iff the press really was in the control.*/
{
#ifdef INTERACTIVE
    int left, right, bottom, top;
    int l, r, b, t;
    int k;
    ObjPtr var;
    ObjPtr value;
    real *elements;
    real range[2];
    real domain[2];
    int nEntries;

    Get2DIntBounds(control, &left, &right, &bottom, &top);

    left = left + EDGE + MINORBORDER;
    right = right - EDGE - MINORBORDER;
    bottom = bottom + EDGE + MINORBORDER;
    top = top - EDGE - MINORBORDER;

    l = left;
    r = right;
    b = bottom;
    t = top;

    IndentMappingBounds(control, &l, &r, &b, &t);

    if (x >= left && x <= right && y >= bottom && y <= top)
    {
	/*Hey!  It really was a click in the control*/

	if (TOOL(flags) == T_HELP)
	{
	    ContextHelp(control);
	    return ObjTrue;
	}

	MakeVar(control, NENTRIES);
	var = GetIntVar("DrawComponentMapping", control, NENTRIES);
	if (!var)
	{
	    return ObjFalse;
	}
	nEntries = GetInt(var);

	var = GetFixedArrayVar("DrawComponentMapping", control, YRANGE, 1, 2L);
	if (!var) return ObjFalse;
	Array2CArray(range, var);

	var = GetFixedArrayVar("DrawComponentMapping", control, XDOMAIN, 1, 2L);
	if (!var) return ObjFalse;
	Array2CArray(domain, var);

	MakeVar(control, VALUE);
	value = GetVar(control, VALUE);
	if (!value) return ObjTrue;
	elements = ELEMENTS(value);

	SaveForUndo(control);

	if (x >= l - SLOP && x <= r + SLOP && y >= b - SLOP && y <= t + SLOP)
	{
	    /*It's in the edit portion.  Do stuff based on edit mode*/
	    int editMode;
	    int lastX, lastY;
	    int lastIndex;
	    real lastValue;
	    int k;
	    int x1, x2;
	    int newX, newY;
	    int newIndex;
	    real newValue;
	    ObjPtr panel, button;
	    ObjPtr var;

	    var = GetIntVar("PressMappingControl", control, EDITMODE);
	    if (!var)
	    {
		return ObjFalse;
	    }
	    editMode = GetInt(var);

	    if (editMode == PT_FREEFORM)
	    {
		int firstAltered, lastAltered;
		Bool constrainX, constrainY;

		/*Free form*/
		lastIndex = -1;
		lastValue = 0.0;

		constrainX = constrainY = false;

		while (Mouse(&newX, &newY))
		{
		    if (newX >= l - SLOP && newX <= r + SLOP && newY >= b - SLOP && newY <= t + SLOP)
		    {
			if (flags & F_SHIFTDOWN)
			{
			    if (!constrainX && !constrainY)
			    {
				int xDiff, yDiff;

				/*Must make some constraints*/
				xDiff = ABS(newX - x);
				yDiff = ABS(newY - y);
				if (xDiff >= MINCONSTRAINT &&
				    yDiff >= MINCONSTRAINT)
				{
				    if (yDiff > xDiff)
				    {
					constrainX = true;
				    }
				    else
				    {
					constrainY = true;
				    }
				}
			    }
			    if (constrainX) newX = x;
			    if (constrainY) newY = y;
			}

			/*Calculate newValue*/
			newValue = PixelToValue(newY, range[0], range[1], b, t);

			/*Calculate newIndex*/
			newIndex = MapPixel(newX, l, r, 0, nEntries);
			if (newIndex < 0) newIndex = 0;
			if (newIndex > nEntries - 1) newIndex = nEntries - 1;

			if (newIndex != lastIndex || newValue != lastValue)
			{
			    /*Set the new color to be correct*/
			    if (lastIndex < 0 || lastIndex == newIndex)
			    {
				/*New in this range, just set this value*/
				elements[newIndex] = newValue;
				SetVar(control, VALUE, value);
			    }
			    else if (newIndex > lastIndex)
			    {
				/*Linearly interpolate up to this color*/
				for (k = lastIndex + 1; k <= newIndex; ++k)
				{
				    elements[k] = ((newIndex - k) * lastValue +
					  (k - lastIndex) * newValue) / 
						(newIndex - lastIndex);
				}
				SetVar(control, VALUE, value);
			    }
			    else if (newIndex < lastIndex)
			    {
				/*Linearly interpolate down to this color*/
				for (k = lastIndex - 1; k >= newIndex; --k)
				{
				    elements[k] = ((k - newIndex) * lastValue +
					  (lastIndex - k) * newValue) / 
						(lastIndex - newIndex);
				}
				SetVar(control, VALUE, value);
			    }
			    if (lastIndex >= 0)
			    {
				x1 = MapPixel(MIN(newIndex, lastIndex), 0, nEntries, l, r) - 1;
				x2 = MapPixel(MAX(newIndex, lastIndex) + 1, 0, nEntries, l, r) + 1;
			    }
			    else
			    {
				x1 = MapPixel(newIndex, 0, nEntries, l, r) - 1;
				x2 = MapPixel(newIndex + 1, 0, nEntries, l, r) + 1;
			    }
			    DrawMeInBounds(control, x1, x2, bottom, top);
 			    lastValue = newValue;
			    lastIndex = newIndex;
			}
		    }
		}
		ChangedValue(control);
		LogControl(control);
	    }
	}
	return ObjTrue;
    }
#endif
    return ObjFalse;
}

static ObjPtr DrawComponentMapping(control)
ObjPtr control;
/*Draws the mapping within a component control*/
{
    int l, r, b, t;
    int flags;
    ObjPtr var;
    ObjPtr value;
    real *elements;
    real range[2];
    real domain[2];
    int nEntries;
    int left, right, bottom, top;
    int x1, x2, y, k;
    float clr[3];
    int sepComponent;
    int editComponent;
    int colorModel;

    Get2DIntBounds(control, &left, &right, &bottom, &top);
    var = GetVar(control, CONTROLFLAGS);
    if (var)
    {
	flags = GetInt(var);
    }
    else
    {
	flags = 0;
    }

    /*Get the interior*/
    l = left + EDGE + MINORBORDER;
    r = right - EDGE - MINORBORDER;
    b = bottom + EDGE + MINORBORDER;
    t = top - EDGE - MINORBORDER;

    IndentMappingBounds(control, &l, &r, &b, &t);

    value = GetValue(control);
    if (value)
    {
	/*There's a value, let's draw it.*/
	elements = ELEMENTS(value);

	MakeVar(control, NENTRIES);
	var = GetIntVar("DrawComponentMapping", control, NENTRIES);
	if (!var)
	{
	    return ObjFalse;
	}
	nEntries = GetInt(var);

	var = GetFixedArrayVar("DrawComponentMapping", control, XDOMAIN, 1, 2L);
	if (!var) return ObjFalse;
	Array2CArray(domain, var);

	var = GetFixedArrayVar("DrawComponentMapping", control, YRANGE, 1, 2L);
	if (!var) return ObjFalse;
	Array2CArray(range, var);

	var = GetIntVar("DrawComponentMapping", control, EDITCOMPONENT);
	if (!var) return ObjFalse;
	editComponent = GetInt(var);

	var = GetIntVar("DrawComponentMapping", control, COLORMODEL);
	if (!var) return ObjFalse;
	colorModel = GetInt(var);

	x1 = l;

	OptimizeColor();

	sepComponent = sepComponents[colorModel][editComponent];

	for (k = 0; k < nEntries; ++k)
	{
	    x2 = (k == nEntries - 1) ? r : MapPixel(k + 1, 0, nEntries, l, r);

	    y = ValueToRectPixel(elements[k], range[0], range[1], b, t);
	    if (x2 > x1)
	    {
		FillUIRect(x1, x2, y, t, UIGRAY62);
		RGBC(componentColors[FTOIC(elements[k])][sepComponent]);
		FillRect(x1, x2, b, y);
	    }
	    x1 = x2;
	}

	OptimizeSharpness();

	if (1)
	{
	    /*Draw white rectangles*/

	    x1 = l;

	    SetUIColor(UIGRAY62);

	    for (k = 0; k < nEntries; ++k)
	    {
		x2 = (k == nEntries - 1) ? r : MapPixel(k + 1, 0, nEntries, l, r);

		y = ValueToRectPixel(elements[k], range[0], range[1], b, t);
		if (x2 > x1)
		{
		    FillRect(x1, x2, y, t);
		}
		x1 = x2;
	    }
	}
	else
	{
	    /*Draw lines from value centers*/
	    int x1, y1, x2, y2;
	    int k;

	    x1 = l;
	    y1 = ValueToPixel(elements[0], range[0], range[1], b, t);

	    for (k = 0; k < nEntries; ++k)
	    {
		x2 = MapPixel(k* 2 + 1, 0, nEntries * 2, l, r);
		y2 = ValueToPixel(elements[k], range[0], range[1], b, t);
		DrawUILine(x1, y1, x2, y2, UIBLACK);
		x1 = x2;
		y1 = y2;
	    }
	    DrawUILine(x1, y1, r - 1, y1, UIBLACK);
	}
    }

    /*Outline the mapping box*/
    FrameUIRect(l - 1, r + 1, b - 1, t + 1, UIBLACK);
    SetUIColor(UIBACKGROUND);
    FrameWideRect(l - 5, r + 5, b - 5, t + 5, 4);

    return ObjTrue;
}

static ObjPtr DrawComponentExtras(control)
ObjPtr control;
/*Draws the extra bits within a component control*/
{
    int l, r, b, t;
    ObjPtr var;
    ObjPtr value;
    int nEntries;
    int left, right, bottom, top;
    int y1, y2, x1, x2, k;
    real fc;
    float clr[3];
    int sepComponent;
    int editComponent;
    int colorModel;

    Get2DIntBounds(control, &left, &right, &bottom, &top);

    /*Get the interior*/
    left = left + EDGE + MINORBORDER;
    right = right - EDGE - MINORBORDER;
    bottom = bottom + EDGE + MINORBORDER;
    top = top - EDGE - MINORBORDER;

    l = left;
    r = right;
    b = bottom;
    t = top;

    IndentMappingBounds(control, &l, &r, &b, &t);

    var = GetIntVar("DrawComponentExtras", control, EDITCOMPONENT);
    if (!var) return ObjFalse;
    editComponent = GetInt(var);

    var = GetIntVar("DrawComponentExtras", control, COLORMODEL);
    if (!var) return ObjFalse;
    colorModel = GetInt(var);

    MakeVar(control, NENTRIES);
    var = GetIntVar("DrawComponentExtras", control, NENTRIES);
    if (!var)
    {
	return ObjFalse;
    }
    nEntries = GetInt(var);

    OptimizeColor();

    if (!IsDrawingRestricted(right - CC_COLORBARWIDTH - 1, right + 1, b - 1, t + 1))
    {
	/*Draw the color range box*/
  
	OptimizeColor();
    
	sepComponent = sepComponents[colorModel][editComponent];
    
	y1 = b;
    
	for (k = 0; k < 255; ++k)
	{
	    fc = ITOFC(k);

	    y2 = ValueToRectPixel(fc, 0.0, 1.0, b, t);
	    if (y2 > y1)
	    {
		RGBC(componentColors[k][sepComponent]);
		FillRect(right - CC_COLORBARWIDTH, right, y1, y2);
	    }
	    y1 = y2;
	}

	FrameUIRect(right - CC_COLORBARWIDTH - 1, right + 1, b - 1, t + 1, UIBLACK);
    }

    if (!IsDrawingRestricted(l - 1, r + 1, bottom - 1, bottom +  CC_COLORBARWIDTH + 1))
    {
	/*Draw the color bar*/
	ObjPtr palette;

	palette = GetPaletteVar("DrawComponentExtras", control, REPOBJ);
	if (palette)
	{
    	    short3 *colors;

	    MakeVar(palette, COLORS);
	    var = GetVar(palette, COLORS);
	    if (var)
	    {
		colors = ELEMENTS(var);

		x1 = l;
	    
		for (k = 0; k < nEntries; ++k)
		{
		    x2 = (k == nEntries - 1) ? r : MapPixel(k + 1, 0, nEntries, l, r);
	    
		    if (x2 > x1)
		    {
			clr[0] = ITOFC(colors[k + 2][0]);
			clr[1] = ITOFC(colors[k + 2][1]);
			clr[2] = ITOFC(colors[k + 2][2]);
			RGBC(clr);
			FillRect(x1, x2, bottom, bottom + CC_COLORBARWIDTH);
		    }
		    x1 = x2;
		}
	    }
	}
	else
	{
	    FillUIRect(l, r, bottom, bottom +  CC_COLORBARWIDTH, UIBACKGROUND);
	}

	FrameUIRect(l - 1, r + 1, bottom - 1, bottom +  CC_COLORBARWIDTH + 1, UIBLACK);
    }

    OptimizeSharpness();

    return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
static void DrawGenerator(ObjPtr generator, int l, int r, int b, int t, int nEntries, real domain[2], real range[2])
#else
static void DrawGenerator(generator, l, r, b, t, nEntries, domain, range)
ObjPtr generator;
int l, r, b, t, nEntries;
real range[2], domain [2];
#endif
/*Draws a generator within the given bounds over the range*/
{
    int x1, y1, x2, y2;
    int k;
    ObjPtr var;
    real *xVal;
    real *yVal;
    int nSamples;

    var = GetVar(generator, XDOMAIN);
    if (!var) return;
    nSamples = DIMS(var)[0];
    xVal = ELEMENTS(var);

    var = GetVar(generator, YRANGE);
    if (!var) return;
    yVal = ELEMENTS(var);

    /*Draw shadow*/
    SetUIColor(UIBLACK);

    SetLineWidth(2);

    x1 = ValueToPixel(xVal[0], domain[0], domain[1], l, r) + 2;
    y1 = ValueToPixel(yVal[0], range[0], range[1], b, t) - 2;

    for (k = 1; k < nSamples; ++k)
    {
	x2 = ValueToPixel(xVal[k], domain[0], domain[1], l, r) + 2;
	y2 = ValueToPixel(yVal[k], range[0], range[1], b, t) - 2;
	DrawLine(x1, y1, x2, y2);
	x1 = x2;
	y1 = y2;
    }

    SetLineWidth(1);

    x1 = ValueToPixel(xVal[0], domain[0], domain[1], l, r) + 2;
    y1 = ValueToPixel(yVal[0], range[0], range[1], b, t) - 2;
    FillUIRect(x1 - 2, x1 + 3, y1 - 2, y1 + 3, UIBLACK);

    for (k = 1; k < nSamples; ++k)
    {
	x2 = ValueToPixel(xVal[k], domain[0], domain[1], l, r) + 2;
	y2 = ValueToPixel(yVal[k], range[0], range[1], b, t) - 2;
	FillUIRect(x2 - 2, x2 + 3, y2 - 2, y2 + 3, UIBLACK);
	x1 = x2;
	y1 = y2;
    }

    /*Draw foreground*/
    SetUIColor(UIYELLOW);

    SetLineWidth(2);

    x1 = ValueToPixel(xVal[0], domain[0], domain[1], l, r);
    y1 = ValueToPixel(yVal[0], range[0], range[1], b, t);

    for (k = 1; k < nSamples; ++k)
    {
	x2 = ValueToPixel(xVal[k], domain[0], domain[1], l, r);
	y2 = ValueToPixel(yVal[k], range[0], range[1], b, t);
	DrawLine(x1, y1, x2, y2);
	x1 = x2;
	y1 = y2;
    }

    SetLineWidth(1);

    x1 = ValueToPixel(xVal[0], domain[0], domain[1], l, r);
    y1 = ValueToPixel(yVal[0], range[0], range[1], b, t);
    FillUIRect(x1 - 1, x1 + 2, y1 - 1, y1 + 2, UIBLACK);
    FrameUIRect(x1 - 2, x1 + 3, y1 - 2, y1 + 3, UIYELLOW);

    for (k = 1; k < nSamples; ++k)
    {
	x2 = ValueToPixel(xVal[k], domain[0], domain[1], l, r);
	y2 = ValueToPixel(yVal[k], range[0], range[1], b, t);
	FillUIRect(x2 - 1, x2 + 2, y2 - 1, y2 + 2, UIBLACK);
	FrameUIRect(x2 - 2, x2 + 3, y2 - 2, y2 + 3, UIYELLOW);
	x1 = x2;
	y1 = y2;
    }
}

static ObjPtr DrawMappingControl(control)
ObjPtr control;
/*Draws a mapping control*/
{
    int left, right, bottom, top;
    FuncTyp method;
    ObjPtr contents;
    ObjPtr var;
    int editMode;
    ObjPtr value;
    int l, r, b, t;
    real range[2], domain[2];
    ObjPtr generator;
    int nEntries;

    Get2DIntBounds(control, &left, &right, &bottom, &top);
    if (IsDrawingRestricted(left, right, bottom, top)) return ObjFalse;

    DrawRaisedEdge(left, right, bottom, top);

    method = GetMethod(control, DRAWMAPPING);
    if (method)
    {
	(*method)(control);
    }

    /*Get the interior*/
    l = left + EDGE + MINORBORDER;
    r = right - EDGE - MINORBORDER;
    b = bottom + EDGE + MINORBORDER;
    t = top - EDGE - MINORBORDER;

    IndentMappingBounds(control, &l, &r, &b, &t);

    var = GetFixedArrayVar("DrawMapping", control, XDOMAIN, 1, 2L);
    if (!var) return ObjFalse;
    Array2CArray(domain, var);

    var = GetFixedArrayVar("DrawMapping", control, YRANGE, 1, 2L);
    if (!var) return ObjFalse;
    Array2CArray(range, var);

    MakeVar(control, NENTRIES);
    var = GetIntVar("DrawMapping", control, NENTRIES);
    if (!var) return ObjFalse;
    nEntries = GetInt(var);

    contents = GetVar(control, CONTENTS);
    if (contents) DrawList(contents);

    var = GetVar(control, EDITMODE);
    if (var)
    {
	editMode = GetInt(var);

	if (editMode == CCT_LINEAR)
	{
	    MakeVar(control, GENERATOR);

	    generator = GetVar(control, GENERATOR);
	    SetUIColor(UIYELLOW);
	    DrawGenerator(generator, l, r, b, t, nEntries, domain, range);
	}
    }

    method = GetMethod(control, DRAWEXTRAS);
    if (method)
    {
	(*method)(control);
    }

    return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewMappingControl(ObjPtr superClass, int left, int right, int bottom, int top, char *name, int flags,
	real xMin, real xMax, real yMin, real yMax, int nEntries)
#else
ObjPtr NewMappingControl(superClass, left, right, bottom, top, name, flags, xMin, xMax, yMin, yMax, nEntries)
ObjPtr superClass, int left; int right; int bottom; int top; char *name, int flags;
real xMin, xMax, yMin, yMax;
int nEntries;
#endif
/*Returns a new mapping control with name within left, right, bottom, top
  xMin and xMax are the range of the mapping
  yMin and yMax are the domain of the mapping
  nEntries is the number of entries in the mapping*/
{
    ObjPtr retVal;
    ObjPtr contents;
    ObjPtr var;
    real *elements;
    ObjPtr scale;
    int l, r, b, t;

    retVal = NewObject(superClass ? superClass : mappingControlClass, 0L);
    Set2DIntBounds(retVal, left, right, bottom, top);
    SetVar(retVal, NAME, NewString(name));
    SetVar(retVal, CONTROLFLAGS, NewInt(flags));

    var = NewRealArray(1, 2L);
    elements = ELEMENTS(var);
    elements[0] = xMin;
    elements[1] = xMax;
    SetVar(retVal, XDOMAIN, var);

    var = NewRealArray(1, 2L);
    elements = ELEMENTS(var);
    elements[0] = yMin;
    elements[1] = yMax;
    SetVar(retVal, YRANGE, var);

    SetVar(retVal, NENTRIES, NewInt(nEntries));

    /*Add scales*/
    contents = NewList();
    SetVar(retVal, CONTENTS, contents);

    /*Get the interior*/
    left = left + EDGE + MINORBORDER;
    right = right - EDGE - MINORBORDER;
    bottom = bottom + EDGE + MINORBORDER;
    top = top - EDGE - MINORBORDER;
    l = left;
    r = right;
    b = bottom;
    t = top;

    IndentMappingBounds(retVal, &l, &r, &b, &t);

    /*Make the scales*/
    if (flags & MC_SCALELEFT)
    {
	scale = NewScale(left, l, b, t, SO_LEFT, false, "Left Scale");
	SetVar(scale, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM + STICKYTOP));
	SetVar(scale, INDENTATION, NewInt(0));
	SetVar(scale, PARENT, retVal);
	PrefixList(contents, scale);
	SetScaleRange(scale, yMin, yMax);
    }
    if (flags & MC_SCALERIGHT)
    {
	scale = NewScale(r, right, b, t, SO_RIGHT, false, "Right Scale");
	SetVar(scale, STICKINESS, NewInt(STICKYRIGHT + STICKYBOTTOM + STICKYTOP));
	SetVar(scale, INDENTATION, NewInt(0));
	SetVar(scale, PARENT, retVal);
	PrefixList(contents, scale);
	SetScaleRange(scale, yMin, yMax);
    }
    if (flags & MC_SCALEBOTTOM)
    {
	scale = NewScale(l, r, bottom, b, SO_BOTTOM, false, "Bottom Scale");
	SetVar(scale, STICKINESS, NewInt(STICKYLEFT + STICKYRIGHT + STICKYBOTTOM));
	SetVar(scale, INDENTATION, NewInt(0));
	SetVar(scale, PARENT, retVal);
	PrefixList(contents, scale);
	SetScaleRange(scale, xMin, xMax);
    }
    if (flags & MC_SCALETOP)
    {
	scale = NewScale(l, r, t, top, SO_TOP, false, "Top Scale");
	SetVar(scale, STICKINESS, NewInt(STICKYLEFT + STICKYRIGHT + STICKYTOP));
	SetVar(scale, INDENTATION, NewInt(0));
	SetVar(scale, PARENT, retVal);
	PrefixList(contents, scale);
	SetScaleRange(scale, xMin, xMax);
    }

    return retVal;
}

static MakeMethod MakeComponentControlAppearance(control)
ObjPtr control;
/*Makes the appearance of a component control*/
{
    FuncTyp method;

    MakeVar(control, VALUE);

    InhibitLogging(true);
    method = GetMethod(control, EXTRAAPPEARANCE);
    if (method)
    {
	(*method)(control);
    }
    ImInvalid(control);
    InhibitLogging(false);

    SetVar(control, APPEARANCE, ObjTrue);

    return ObjTrue;
}

static ObjPtr MakeComponentControlValue(control)
ObjPtr control;
/*Makes the value of a component control*/
{
    ObjPtr repObj, var, value;
    FuncTyp method;
    long nColors;
    short3 *colors;
    real *elements;
    long k;
    CCFunc FromRGBFunc;
    int colorModel;
    int componentIndex;
    real r, g, b;
    real c[3];

    repObj = GetVarSurely("MakeComponentControlValue", control, REPOBJ);
    if (!repObj)
    {
	return ObjFalse;
    }

    MakeVar(repObj, COLORS);
    var = GetVarSurely("MakeComponentControlValue", repObj, COLORS);
    if (!var)
    {
	return ObjFalse;
    }
    colors = ELEMENTS(var);
    nColors = DIMS(var)[0];

    MakeVar(control, NENTRIES);
    var = GetIntVar("MakeComponentControlValue", control, NENTRIES);
    if (!var)
    {
	return ObjFalse;
    }
    if (GetInt(var) != nColors - 3)
    {
	ReportError("MakeComponentControlValue", "Internal color number mismatch");
	return ObjFalse;
    }
    value = NewRealArray(1, GetInt(var));
    elements = ELEMENTS(value);

    var = GetIntVar("MakeComponentControlValue", control, EDITCOMPONENT);
    if (!var)
    {
	return ObjFalse;
    }
    componentIndex = GetInt(var);

    var = GetIntVar("MakeComponentControlValue", control, COLORMODEL);
    if (!var)
    {
	return ObjFalse;
    }
    colorModel = GetInt(var);

    /*Set the colors*/
    FromRGBFunc = colorModelInfo[colorModel] . FromRGBFunc;
    for (k = 0; k < nColors - 3; ++k)
    {
	r = ITOFC(colors[k + 2][0]);
	g = ITOFC(colors[k + 2][1]);
	b = ITOFC(colors[k + 2][2]);

	(*FromRGBFunc)(&(c[0]), &(c[1]), &(c[2]), r, g, b);

	elements[k] = c[componentIndex];
    }

    InhibitLogging(true);
    method = GetMethod(control, CHANGEDVALUE);
    SetMethod(control, CHANGEDVALUE, (FuncTyp) 0);
    SetValue(control, value);
    SetMethod(control, CHANGEDVALUE, method);
    InhibitLogging(false);

    return ObjTrue;
}

static ObjPtr MakeComponentControlNEntries(control)
ObjPtr control;
/*Makes the NENTRIES of a component control*/
{
    ObjPtr repObj, var, value;
    long nColors;

    repObj = GetVarSurely("MakeComponentControlValue", control, REPOBJ);
    if (!repObj)
    {
	return ObjFalse;
    }

    MakeVar(repObj, COLORS);
    var = GetVarSurely("MakeComponentControlValue", repObj, COLORS);
    if (!var)
    {
	return ObjFalse;
    }
    nColors = DIMS(var)[0];

    SetVar(control, NENTRIES, NewInt(nColors - 3));

    return ObjTrue;
}

static ObjPtr ChangeComponentPalette(control)
ObjPtr control;
/*Changes a palette in response to a change in a color control*/
{
    ObjPtr repObj, var, value;
    FuncTyp method;
    long nColors;
    short3 *colors;
    real *elements;
    long k;
    CCFunc ToRGBFunc;
    CCFunc FromRGBFunc;
    int colorModel;
    int componentIndex;
    real r, g, b;
    real c[3];

    repObj = GetVarSurely("ChangeComponentPalette", control, REPOBJ);
    if (!repObj)
    {
	return ObjFalse;
    }

    MakeVar(repObj, COLORS);
    var = GetVarSurely("ChangeComponentPalette", repObj, COLORS);
    if (!var)
    {
	return ObjFalse;
    }
    colors = ELEMENTS(var);
    SetVar(repObj, COLORS, var);
    nColors = DIMS(var)[0];

    MakeVar(control, NENTRIES);
    var = GetIntVar("ChangeComponentPalette", control, NENTRIES);
    if (!var)
    {
	return ObjFalse;
    }
    if (GetInt(var) != nColors - 3)
    {
	ReportError("ChangeComponentPalette", "Internal color number mismatch");
	return ObjFalse;
    }

    value = GetValue(control);
    if (!value) return ObjFalse;
    elements = ELEMENTS(value);

    var = GetIntVar("ChangeComponentPalette", control, EDITCOMPONENT);
    if (!var)
    {
	return ObjFalse;
    }
    componentIndex = GetInt(var);

    var = GetIntVar("ChangeComponentPalette", control, COLORMODEL);
    if (!var)
    {
	return ObjFalse;
    }
    colorModel = GetInt(var);

    /*Set the colors*/
    FromRGBFunc = colorModelInfo[colorModel] . FromRGBFunc;
    ToRGBFunc = colorModelInfo[colorModel] . ToRGBFunc;
    for (k = 0; k < nColors - 3; ++k)
    {
	r = ITOFC(colors[k + 2][0]);
	g = ITOFC(colors[k + 2][1]);
	b = ITOFC(colors[k + 2][2]);

	(*FromRGBFunc)(&(c[0]), &(c[1]), &(c[2]), r, g, b);

	/*Fiddle saturation, value, and lightness*/
	if (componentIndex == 0)
	{
	    if (colorModel == CM_HSV)
	    {
		if (c[1] <= 0.0) c[1] = 1.0;
		if (c[2] <= 0.0) c[2] = 0.0;
	    }
	    else if (colorModel == CM_HLS)
	    {
		if (c[1] <= 0.0) c[1] = 0.5;
		if (c[1] >= 1.0) c[1] = 0.5;
		if (c[2] <= 0.0) c[2] = 1.0;
	    }
	}

	c[componentIndex] = elements[k];

	(*ToRGBFunc)(&r, &g, &b, c[0], c[1], c[2]);

	colors[k + 2][0] = FTOIC(r);
	colors[k + 2][1] = FTOIC(g);
	colors[k + 2][2] = FTOIC(b);
    }

    SetVar(control, VALUE, value);

    return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewComponentControl(int l, int r, int b, int t, char *name, ObjPtr palette)
#else
ObjPtr NewComponentControl(l, r, b, t, name, palette)
int l; int r; int b; int t; char *name;
#endif
/*Returns a new component control*/
{
    ObjPtr retVal;
    float *elements;
    int nColors;
    ObjPtr var;

    MakeVar(palette, MINMAX);
    var = GetFixedArrayVar("NewComponentControl", palette, MINMAX, 1, 2L);
    if (!var) return NULLOBJ;
    elements = ELEMENTS(var);

    var = GetIntVar("NewComponentControl", palette, NCOLORS);
    if (!var) return NULLOBJ;
    nColors = GetInt(var);

    retVal = NewMappingControl(componentControlClass, l, r, b, t, name, MC_SCALELEFT + MC_SCALEBOTTOM,
	elements[0], elements[1], 0.0, 1.0, nColors - 3);
    SetVar(retVal, REPOBJ, palette);

    SetMethod(retVal, CHANGEDVALUE, ChangeComponentPalette);

    return retVal;
}

#define MINSLOPECHANGE	0.1	/*Minimum slope change for corner*/

static ObjPtr MakeMappingGenerator(control)
ObjPtr control;
/*Makes a mapping generator for a control*/
{
    ObjPtr value, generator;
    ObjPtr domain, range;
    short *xVal;
    real *yVal;
    long dimension;
    real *elements;
    ObjPtr var;
    int nEntries;

    MakeVar(control, VALUE);
    value = GetValue(control);
    if (!value) return ObjFalse;
    elements = ELEMENTS(value);

    MakeVar(control, NENTRIES);
    var = GetIntVar("MakeMappingGenerator", control, NENTRIES);
    if (!var) return ObjFalse;
    nEntries = GetInt(var);

    generator = NewObject(NULLOBJ, 0L);

    dimension = 2;
    domain = NewArray(AT_REAL, 1, &dimension);
    xVal = ELEMENTS(domain);
    range = NewArray(AT_REAL, 1, &dimension);
    yVal = ELEMENTS(range);

    var = GetVar(control, XDOMAIN);
    if (!var) return ObjFalse;
    Array2CArray(xVal, var);

    yVal[0] = elements[0];
    yVal[1] = elements[nEntries - 1];

    /*Now go through and find slope changes*/

    SetVar(generator, XDOMAIN, domain);
    SetVar(generator, YRANGE, range);
    SetVar(control, GENERATOR, generator);

    return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
void InitColorControls(void)
#else
void InitColorControls()
#endif
/*Initialize the color controls*/
{
    int c;
    real r, g, b, h, s, v, l, y, i, q;
    long dimension;
    ObjPtr var;
    short *elements;

    /*Fill componentColors*/
    for (c = 0; c < 256; ++c)
    {
	float fc;

	fc = ITOFC(c);

	/*RGB*/
	componentColors[c][SEP_RED][0] = fc;
	componentColors[c][SEP_RED][1] = 0.0;
	componentColors[c][SEP_RED][2] = 0.0;

	componentColors[c][SEP_GREEN][0] = 0.0;
	componentColors[c][SEP_GREEN][1] = fc;
	componentColors[c][SEP_GREEN][2] = 0.0;

	componentColors[c][SEP_BLUE][0] = 0.0;
	componentColors[c][SEP_BLUE][1] = 0.0;
	componentColors[c][SEP_BLUE][2] = fc;

	/*Hue*/
	h = fc;
	s = 1.0;
	v = 1.0;

	HSV2RGB(&r, &g, &b, h, s, v);
	componentColors[c][SEP_HUE][0] = r;
	componentColors[c][SEP_HUE][1] = g;
	componentColors[c][SEP_HUE][2] = b;

	/*Saturation*/
	h = 0.5;
	s = fc;
	v = 1.0;

	HSV2RGB(&r, &g, &b, h, s, v);
	componentColors[c][SEP_SATURATION][0] = r;
	componentColors[c][SEP_SATURATION][1] = g;
	componentColors[c][SEP_SATURATION][2] = b;

	/*Value*/
	componentColors[c][SEP_VALUE][0] = fc;
	componentColors[c][SEP_VALUE][1] = fc;
	componentColors[c][SEP_VALUE][2] = fc;

	/*Lightness*/
	h = 0.5;
	l = fc;
	s = 1.0;

	HLS2RGB(&r, &g, &b, h, l, s);
	componentColors[c][SEP_LIGHTNESS][0] = r;
	componentColors[c][SEP_LIGHTNESS][1] = g;
	componentColors[c][SEP_LIGHTNESS][2] = b;

	/*Y*/
	y = fc;
	i = 1.0;
	q = 0.0;

	YIQ2RGB(&r, &g, &b, y, i, q);
	componentColors[c][SEP_Y][0] = r;
	componentColors[c][SEP_Y][1] = g;
	componentColors[c][SEP_Y][2] = b;
	
	/*I*/
	y = 1.0;
	i = fc;
	q = 0.0;

	YIQ2RGB(&r, &g, &b, y, i, q);
	componentColors[c][SEP_I][0] = r;
	componentColors[c][SEP_I][1] = g;
	componentColors[c][SEP_I][2] = b;
	
	/*Q*/
	y = 1.0;
	i = 1.0;
	q = fc;

	YIQ2RGB(&r, &g, &b, y, i, q);
	componentColors[c][SEP_Q][0] = r;
	componentColors[c][SEP_Q][1] = g;
	componentColors[c][SEP_Q][2] = b;
    }

    /*Make the controls*/
    swatchClass = NewObject(controlClass, 0L);
    AddToReferenceList(swatchClass);
    SetVar(swatchClass, TYPESTRING, NewString("color swatch"));
    DeclareDependency(swatchClass, APPEARANCE, REPOBJ);
    SetMethod(swatchClass, APPEARANCE, MakeSwatchAppearance);
    SetVar(swatchClass, OPAQUE, ObjTrue);
    SetMethod(swatchClass, DRAW, DrawSwatch);

    /*Now mapping*/
    mappingControlClass = NewObject(controlClass, 0L);
    AddToReferenceList(mappingControlClass);
    SetMethod(mappingControlClass, DRAW, DrawMappingControl);
    SetMethod(mappingControlClass, PRESS, PressMappingControl);
    SetMethod(mappingControlClass, DRAWMAPPING, DrawMapping);
    DeclareDependency(mappingControlClass, APPEARANCE, VALUE);
    DeclareDependency(mappingControlClass, APPEARANCE, XSCALE);
    DeclareDependency(mappingControlClass, APPEARANCE, YSCALE);
    SetMethod(mappingControlClass, APPEARANCE, MakeMappingControlAppearance);
    dimension = 4;
    var = NewArray(AT_SHORT, 1, &dimension);
    elements = ELEMENTS(var);
    elements[0] = elements[1] = MC_SCALEWIDTH;
    elements[2] = elements[3] = MC_SCALEHEIGHT;
    SetVar(mappingControlClass, SCALEMARGIN, var);
    var = NewArray(AT_SHORT, 1, &dimension);
    elements = ELEMENTS(var);
    elements[0] = elements[1] = 0;
    elements[2] = elements[3] = 0;
    SetVar(mappingControlClass, EXTRAMARGIN, var);
    SetVar(mappingControlClass, OPAQUE, ObjTrue);
    SetVar(mappingControlClass, EDITMODE, NewInt(CCT_FREEFORM));
    AddSnapVar(mappingControlClass, EDITMODE);
    DeclareDependency(mappingControlClass, APPEARANCE, EDITMODE);
    DeclareDependency(mappingControlClass, GENERATOR, EDITMODE);
    DeclareDependency(mappingControlClass, GENERATOR, VALUE);
    SetMethod(mappingControlClass, GENERATOR, MakeMappingGenerator);

    /*Now component control*/
    componentControlClass = NewObject(mappingControlClass, 0L);
    AddToReferenceList(componentControlClass);
    DeclareDependency(componentControlClass, APPEARANCE, VALUE);
    DeclareDependency(componentControlClass, APPEARANCE, NENTRIES);
    DeclareDependency(componentControlClass, VALUE, COLORMODEL);
    DeclareDependency(componentControlClass, VALUE, EDITCOMPONENT);
    DeclareDependency(componentControlClass, APPEARANCE, VALUE);
    AddSnapVar(componentControlClass, COLORMODEL);
    AddSnapVar(componentControlClass, EDITCOMPONENT);
    DeclareIndirectDependency(componentControlClass, VALUE, REPOBJ, COLORS);
    DeclareIndirectDependency(componentControlClass, VALUE, REPOBJ, NCOLORS);
    DeclareIndirectDependency(componentControlClass, NENTRIES, REPOBJ, COLORS);
    DeclareIndirectDependency(componentControlClass, NENTRIES, REPOBJ, NCOLORS);
    SetMethod(componentControlClass, VALUE, MakeComponentControlValue);
    SetMethod(componentControlClass, NENTRIES, MakeComponentControlNEntries);
    SetMethod(componentControlClass, APPEARANCE, MakeComponentControlAppearance);
    SetVar(componentControlClass, EDITCOMPONENT, NewInt(0));
    SetVar(componentControlClass, COLORMODEL, NewInt(CM_HSV));
    SetMethod(componentControlClass, DRAWMAPPING, DrawComponentMapping);
    SetMethod(componentControlClass, DRAWEXTRAS, DrawComponentExtras);
    dimension = 4;
    var = NewArray(AT_SHORT, 1, &dimension);
    elements = ELEMENTS(var);
    elements[0] = elements[1] = 30;
    elements[2] = elements[3] = MC_SCALEHEIGHT;
    SetVar(componentControlClass, SCALEMARGIN, var);
    var = NewArray(AT_SHORT, 1, &dimension);
    elements = ELEMENTS(var);
    elements[0] = 0;
    elements[1] = CC_COLORBARWIDTH + MINORBORDER;
    elements[2] = CC_COLORBARWIDTH + MINORBORDER;
    elements[3] = 0;
    SetVar(componentControlClass, EXTRAMARGIN, var);
}

#ifdef HAVE_PROTOTYPES
void KillColorControls(void)
#else
void KillColorControls()
#endif
/*Initialize the color controls*/
{
    RemoveFromReferenceList(componentControlClass);
    RemoveFromReferenceList(mappingControlClass);
    RemoveFromReferenceList(swatchClass);
}

