/*	ScianFileSystem.c
	manages file system window for scian

	Jim Lyons
	created 4/17/91

	4/28/91 version 2 of files window
	5/1/91  version 3: made file objects
	5/3/91	added format dialog routines
	5/29/91 EMP removed lib headers
	6/25/91 removed icon definitions
	7/22/91 added Read Data Files menu item
	8/8/91	guarded all strcpy and strcat calls
	9/9/91	added help strings
	9/16/91	redesigned file window
	9/19/91 added file info window
	10/12/91 removed Set File Type stuff
	10/28/91 changed interface to File Readers (with EMP)
	1/20/92	changed order of file type identification in ReadDir
	2/5/92	added DropInContents method to file window corral
	4/8/92	made all chdirs temporary, never change dirs on user
	4/17/92	changed Open to use Object Function
	4/23/92 changed Set Format and Show Info to use Object Functions
	5/26/92 EMP fixed OpenParent not to fail at root
	5/26/92 EMP made GetFileWindow public
	6/13/92 EMP changed to use method declarations, not prototypes
	10/15/92 Modified file info window to make two-line directory name
	2/3/93	Fixed bug in file info routine
	9/21/93 Eliminated multiple file lists
	9/24/93 Changed browser window to be "modal"
	9/30/93	Redesigned Browser and Save windows layout
	3/30/94 Made file icon class and cleaned up icon vs. file vars
	5/3/94  Added ChangeDir routine; updated OpenFile to read scripts
	5/5/94  Added ReadAFile
*/

#include "Scian.h"

/*** move to Scian.h ***/
#ifdef GETPWNAM
#include <pwd.h>
#endif

#include "ScianStyle.h"
#include "ScianTypes.h"
#include "ScianMethods.h"
#include "ScianArrays.h"
#include "ScianFontSystem.h"
#include "ScianWindows.h"
#include "ScianTextBoxes.h"
#include "ScianSliders.h"
#include "ScianButtons.h"
#include "ScianTitleBoxes.h"
#include "ScianObjWindows.h"
#include "ScianObjFunctions.h"
#include "ScianDatabase.h"
#include "ScianSymbols.h"
#include "ScianColors.h"
#include "ScianControls.h"
#include "ScianTimers.h"
#include "ScianLists.h"
#include "ScianIDs.h"
#include "ScianEvents.h"
#include "ScianErrors.h"
#include "ScianDialogs.h"
#include "ScianIcons.h"
#include "ScianFiles.h"
#include "ScianParser.h"
#include "ScianDatasets.h"
#include "ScianScripts.h"
#include "ScianPreferences.h"
#include "ScianFixedClasses.h"
#include "ScianFileSystem.h"
#include "ScianCalculator.h"

#ifndef MAKEBACKUP
#define MAKEBACKUP 750
#endif

#define MAXUSERNAME	20	/*Max # of chars in user name*/

/* defs for file window parameters */

#define FILESTOP	\
	(MINORBORDER + 3*SMALLGAP + 2*OKBTNHT + CHECKBOXHEIGHT)

#define MINBTNWID	120
#define FILTERBTNWID	220
#define VIEWBTNWID	160
#define BACKBTNWID	150
#define FILENAMEWID	240
#define FILENAMELABEL	100

#define FILESWIDTH	450
#define CORRALHEIGHT	300	/* default height of the corral */
#define BROWSERHEIGHT	\
	(FILESTOP + CORRALHEIGHT + MINORBORDER + OKBTNHT + SMALLGAP)
#define SAVEWINHEIGHT	\
	(FILESTOP + CORRALHEIGHT + 2*MINORBORDER + SMALLGAP + OKBTNHT \
		+ EDITBOXHEIGHT)


/* defs for file format dialog box */
#define RADIOWID	84
#define BOTPART		(BOTMARGIN + OKBTNHT + INTERSPACE)

#define STICKYALL	(STICKYRIGHT + STICKYLEFT + STICKYBOTTOM + STICKYTOP)

/* defs for file info window */
#define NINFOITEMS	5
#define INFOITEMHT	40
#define INFOITEMWID	120
#define INFONAMEWIDTH	76
#define INFOFIELDWIDTH	200
#define INFOFIELDHEIGHT	(NINFOITEMS*INFOITEMHT)
#define INFOWIDTH	(3*MINORBORDER + INFONAMEWIDTH + INFOFIELDWIDTH)
#define INFOHEIGHT	\
	(MINORBORDER + MAJORBORDER + GAP + 2*INFOFONTSIZE + 2*DEFLINESPACE + BARWIDTH + INFOFIELDHEIGHT)
#define INFOFONT	"Helvetica-Bold"
#define INFOFONTSIZE	12


/* EMP static method declarations */

static ObjPtr UpdateTitle();
static ObjPtr CancelDialog();
static ObjPtr OKDialog();
static ObjPtr DoParent();
static ObjPtr DropInFileWin();
static ObjPtr OpenFile();
static ObjPtr OpenFolder();
static ObjPtr MakeFileList();
static ObjPtr DoEnter();
static ObjPtr Collect();
static ObjPtr DoSave();
static ObjPtr DoCancel();
static ObjPtr ShowBrowseFiles();
static ObjPtr ShowSaveFiles();
static ObjPtr ReShowFiles();
static ObjPtr ChangeView();
static ObjPtr ChangeBack();
static ObjPtr MakeWhichIcon();

/* static function prototypes */
#ifdef HAVE_PROTOTYPES

static ObjPtr SetFormat(ObjPtr);
static ObjPtr ShowInfo(ObjPtr);
static char *GetFileInfo(char *, char *);
static void FileFormatDialog(FuncTyp);
static ObjPtr OKFormat(int, ObjPtr);
static ObjPtr SelFileList(ObjPtr);
static ObjPtr SelDirList(ObjPtr);
static ObjPtr NewFileWindow(WinInfoPtr, char *);
static ObjPtr GetDirectory(char *);
static int FileType(char *);

#else

static ObjPtr SetFormat();
static ObjPtr ShowInfo();
static char *GetFileInfo();
static void FileFormatDialog();
static ObjPtr OKFormat();
static ObjPtr SelFileList();
static ObjPtr SelDirList();
static ObjPtr NewFileWindow();
static ObjPtr GetDirectory();
static int FileType();

#endif

#ifndef ALLFILES
#define ALLFILES 812	/* file filter switch for browser, save windows */
#endif

#ifndef UPDATE
#define UPDATE	823	/* Update message to text box */
#endif

#ifndef PROMPT
#define PROMPT 837	/* Prompt text string for a text box */
#endif

/********************************************************* GLOBALS */

ObjPtr dirClass;	/* class of directories containing file list */
ObjPtr unixFileClass;	/* class of UNIX files */
ObjPtr unixDirClass;	/* class of UNIX directories */
ObjPtr fileIconClass;	/* class of file (and directory) icons */

char dirSave[MAXPATHLEN+1];	/* save original working directory name */
char curDir[MAXPATHLEN+1];	/* buffer for current directory name */
char replySave[MAXPATHLEN+1];	/* keep dialog reply text for reediting */

ObjPtr aList;	/* list for collecting selected files */
int fileBrowserNumber, saveWindowNumber;


/********************************************************** INIT FILE SYSTEM */
void InitFileSystem(void)
{
    /*Make UNIX file class*/
    unixFileClass = NewObject(fileClass, 0L);
    AddToReferenceList(unixFileClass);
    DefineFixedClass(unixFileClass, CLASS_UNIXFILE, "unixFile");
    SetVar(unixFileClass, DOUBLECLICK, NewString(OF_READ));
    SetMethod(unixFileClass, READFILE, OpenFile);
    SetMethod(unixFileClass, COLLECT, Collect);

    /*Make UNIX directory class*/
    unixDirClass = NewObject(fileClass, 0L);
    AddToReferenceList(unixDirClass);
    DefineFixedClass(unixDirClass, CLASS_UNIXDIR, "unixDir");
    SetVar(unixDirClass, DOUBLECLICK, NewString(OF_OPEN));
    SetMethod(unixDirClass, OPEN, OpenFolder);

    /* create class for directory objects */
    dirClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(dirClass);
    DefineFixedClass(dirClass, CLASS_DIR, "folder");
    SetMethod(dirClass, FILELIST, MakeFileList);
    DeclareDependency(dirClass, FILELIST, CHANGED);
    DeclareDependency(dirClass, FILELIST, NAME);

    /* create class for file system icons */
    fileIconClass = NewObject(iconClass, 0L);
    AddToReferenceList(fileIconClass);
    SetMethod(fileIconClass, WHICHICON, MakeWhichIcon);
    DeclareIndirectDependency(fileIconClass, WHICHICON, REPOBJ, FILETYPE);

    getwd(dirSave);		/* save original working directory name */
    *replySave = '\0';	/* clear saved path */
    
    fileBrowserNumber = 0;	/* last used number for browser window */
    saveWindowNumber = 0;	/* last used number for save window */
}

void KillFileSystem(void)
{
    RemoveFromReferenceList(dirClass);
    RemoveFromReferenceList(unixFileClass);
    RemoveFromReferenceList(unixDirClass);
    RemoveFromReferenceList(fileIconClass);
}

/******************************************************* OBJ FUNCTION ROUTINES */

static ObjPtr Collect(theFile)
ObjPtr theFile;
/* used as method for collecting files to Set Format or Show Info */
{
    PrefixList(aList, theFile);
    return ObjTrue;
}

void SetUpCollect()
/* setup routine for Set Format and Show Info */
{
    aList = NewList();
}

Bool ChangeDir(path)
char *path;
/* changes to directory specified by path; returns true iff success */
{
	if (!path || chdir(path) == 0)
	{
		return true;
	}
	else
	{
		return false;
	}
}

void RestoreDir()
/* restores original directory after opening files */
{
    chdir(dirSave);
}

char *GetCurDir(void)
{
    getwd(curDir);
    return curDir;
}

#ifdef HAVE_PROTOTYPES
Bool FileExists(char *fileName)
#else
Bool FileExists(fileName)
char *fileName;
#endif
/* returns true iff file fileName exists in the current directory */
{
    struct stat stbuf;

    if (stat(fileName, &stbuf) == 0)
    {
	return true;
    }
    else
    {
	return false;
    }
}

/******************************************************* DO NEW FILE WINDOW */
void DoNewFileWindow(void)
/* New File Window menu routine */
{
    WinInfoPtr askWindow;

    Log("Show FilesWindow\n");
    InhibitLogging(true);

    if (*replySave)
    {
	strncpy(tempStr, replySave, TEMPSTRSIZE);
	tempStr[TEMPSTRSIZE] = '\0';
    }
    else if (PrefExists(PREF_DEFDIR))
    {
	strncpy(tempStr, GetPrefString(PREF_DEFDIR), TEMPSTRSIZE);
	tempStr[TEMPSTRSIZE-1] = '\0';
    }
    else getwd(tempStr);
    askWindow = AskUser((WinInfoPtr) NULLOBJ, "Enter directory name:",
	     (FuncTyp) NewFileWindow, tempStr);
    if (askWindow) SetVar((ObjPtr) askWindow, HELPSTRING, NewString("This dialog window \
allows you to enter the name of the directory for which you want to open a new File Window. \
The default directory name, showing when the window first comes up, may be set in the \
Preferences window."));
    InhibitLogging(false);
}

static ObjPtr NewFileAgain(WinInfoPtr win, int n)
/* called if error in directory name from New File Window dialog */
{
    if (n == 1) DoNewFileWindow();
    else *replySave = '\0'; /* clear error string */
    return NULLOBJ;
}

#ifdef HAVE_PROTOTYPES
static ObjPtr NewFileWindow(WinInfoPtr owner, char *replyText)
#else
static ObjPtr NewFileWindow(owner, replyText)
WinInfoPtr owner;
char *replyText;
#endif
/* puts up the New File Window */
{
    char *s, t[MAXPATHLEN + 1];
    WinInfoPtr alert;
    DIR *dirp;

    SKIPBLANKS(replyText);
    if (*replyText == '~')
    {
	/*See if next character could be a name*/
	if (isdigit(replyText[1]) || isalpha(replyText[1]))
	{
	    /*Must be some other user*/
#ifdef GETPWNAM
	    int k;
	    struct passwd *pwdEntry;
	    char userName[MAXUSERNAME + 1];

	    for (k = 1;
		 k < MAXUSERNAME && (isdigit(replyText[k]) || isalpha(replyText[k]));
		 ++k)
	    {
		userName[k-1] = replyText[k];
	    }
	    userName[k-1] = '\0';

	    pwdEntry = getpwnam(userName);
	    if (pwdEntry)
	    {
		/*Entry's OK*/
		strcpy(t, pwdEntry -> pw_dir);
		strcat(t, replyText + k);
	    }
	    else
	    {
		/* unknown user */
		strcpy(replySave, replyText); /* save text for user reediting */
		sprintf(t, "Unknown user, %s.", userName);
		alert = AlertUser(UIERRORALERT, owner, t,
			    NewFileAgain, 2, "Cancel", "Re-enter");
		if (alert) SetVar((ObjPtr) alert, HELPSTRING, NewString("You have used the \
'~user' construct in a directory name to find another user's home directory.  \
No user with the specified name could be found.  Press \
the Re-enter button to correct the entry, or press the Cancel key to cancel the request."));
		return ObjTrue;
	    }
#else
	    /* save text for user reediting */
	    strcpy(replySave, replyText);

	    alert = AlertUser(UIERRORALERT, owner, "Sorry, this version of SciAn does not \
support the '~user' construct in a directory name.", NewFileAgain,
			    2, "Cancel", "Re-enter");
	    if (alert) SetVar((ObjPtr) alert, HELPSTRING, NewString("You have used the \
'~user' construct in a directory name to find a user's home directory.  This \
version of SciAn has not been compiled to accept this construct.  Press \
the Re-enter button to correct the entry, or press the Cancel key to cancel the request."));
	    return ObjTrue;
#endif
	}
	else
	{
	    if (s = getenv("HOME")) /* substitute for squiggle */
	    {
		strcpy(t, s);
		strcat(t, replyText + 1);
	    }
	}
    }
    else if (*replyText)
    {
	strcpy(t, replyText);
    }
    else
    {
	t[0] = '/';
	t[1] = '\0';
    }

    if (ChangeDir(t))	/* check if valid directory */
    {
	getwd(t);	/* get expanded name back */
	
	GetFileWindow(t);
	*replySave = '\0';  /* clear error string */
	RestoreDir();	    /* return to original directory */
	return ObjTrue;
    }
    else /* error */
    {
#define PROMPT1 "Could not open directory "
#define PROMPT2 ". Do you want to enter another directory or cancel the request?"

	/* save text for user reediting */
	strcpy(replySave, replyText);

	/* set up prompt in tempStr */
	strcpy(tempStr, PROMPT1);
	strncat(tempStr, (t ? t : replyText), 
		TEMPSTRSIZE - strlen(PROMPT1) - strlen(PROMPT2) - 1);
	strcat(tempStr, PROMPT2);
	alert = AlertUser(UIERRORALERT, owner, tempStr, NewFileAgain,
			2, "Cancel", "Re-enter");
	if (alert) SetVar((ObjPtr) alert, HELPSTRING, NewString("This dialog window \
came up because the directory entered in the New File Window dialog could not be opened. Press \
the Re-enter button to correct that entry, or press the Cancel key to cancel the request."));
	    
    }
    return ObjTrue;
}

/******************************************************* SET FORMAT */
void ProcessSetFormat(void)
{
    ObjPtr corral;
    int fmt, i,n;

    if (!selWinInfo || !(corral = GetVar((ObjPtr) selWinInfo, CORRAL))
	    || ListCount(aList) == 0) return;

    /* now make dialog */
    FileFormatDialog(OKFormat);

    /* selWinInfo is now the new dialog */
    SetVar((ObjPtr) selWinInfo, FILELIST, aList); /* stash file list with dialog */

    if (logging) InhibitLogging(false);
    return;
}

static ObjPtr OKFormat(int fmt, ObjPtr list)
{
    ObjPtr theFile, theRdr, theIcon, keyList, fileReaders;
    int i, n = ListCount(list);

    /* make a list of all the file readers */
    keyList = NewList();
    PostfixList(keyList, NewSymbol(CLASSID));
    PostfixList(keyList, NewInt(CLASS_FILEREADER));
    fileReaders = SearchDatabase(keyList);

    /* fmt is the number of the button that was selected in the format group */
    if (fmt < 0) return ObjFalse; /* no choice made */
    else if (fmt >= ListCount(fileReaders)) /* choice was "none" */
    {
	for (i=0; i<n; ++i) /* clear FORMAT and READER, leave filetype same */
	{
	    theFile = GetListElem(list, i);
	    theIcon = GetVar(theFile, REPICON);
	    SetVar(theFile, FORMAT, NULLOBJ);
	    ImInvalid(theIcon);	/* cause redraw */
	}
    }
    else
    {
	ObjPtr p, q, r;
	
	theRdr = GetListElem(fileReaders, fmt);
	p = NewInt(DATAFILE);
	q = GetVar(theRdr, NAME);
	r = NewInt(ICONDATAFILE);
	for (i=0; i<n; ++i)
	{
	    theFile = GetListElem(list, i);
	    theIcon = GetVar(theFile, REPICON);
	    SetVar(theFile, FORMAT, GetVar(theRdr, NAME));
	    SetVar(theFile, FILETYPE, p);
	    SetVar(theIcon, WHICHICON, r);
	    ImInvalid(theIcon);	/* cause redraw */
	}
    }
    return ObjTrue;
}

static ObjPtr MakeWhichIcon(icon)
ObjPtr icon;
{
    /* makes a file icon's WHICHICON var */
    ObjPtr theFile;
    int fileType;
    
    theFile = GetVar(icon, REPOBJ);
    if (theFile)
    {
	fileType = GetInt(GetVar(theFile, FILETYPE));
	switch (fileType)
	{
	    case DIRFILE:
		SetVar(icon, WHICHICON, NewInt(ICONFOLDER));
		break;
		
	    case OBJFILE:
		SetVar(icon, WHICHICON, NewInt(ICONOBJFILE));
		break;
		
	    case TEXTFILE:
	        SetVar(icon, WHICHICON, NewInt(ICONTEXTFILE));
		break;
		
	    case DATAFILE:
		SetVar(icon, WHICHICON, NewInt(ICONDATAFILE));
		break;
		
	    case SCIANFILE: /*** miscType -- look up ***/
	    case LOGFILE:
		SetVar(icon, WHICHICON, NewInt(ICONSCIANDOC));
		break;
		
	    case BINFILE:
		SetVar(icon, WHICHICON, NewInt(ICONBINFILE));
		break;
		
	    default:
		SetVar(icon, WHICHICON, NewInt(ICONUNKNOWN));
		break;
	}
    }
    return ObjTrue;
}

/******************************************************************** SHOW INFO */
void ProcessShowInfo(void)
{
    WinInfoPtr infoWin;
    ObjPtr textbox, thePath, list, theString;
    char *path;
    int i;

    path = GetString(thePath = GetVar(GetListElem(aList, 0), PATH));

    if (!ChangeDir(path))
    {
	ReportError("ProcessShowInfo", "Could not change directory");
	return;
}

    /* get the info window */
    theString = NewString("File Info");
    infoWin = GetDialog((WinInfoPtr) NULLOBJ, theString, "File Info",
	    INFOWIDTH, INFOHEIGHT, INFOWIDTH, INFOHEIGHT, WINUI + WINFIXEDSIZE);
    if (textbox = GetVar((ObjPtr) infoWin, REPOBJ)) /* dialog exists */
    {
	ThingListPtr p;
	ObjPtr theFile, field, contents;
	char *fileName;
	int x, itemWid;

	/* pop info window */
	PopWindow(infoWin);
	SetVar(textbox, VALUE, thePath);
	ImInvalid(textbox);

	/* put in file info textbox for each file on the list */
	field = GetVar((ObjPtr) infoWin, FIELD);
	SetVar(field, CONTENTS, contents = NewList());
	p = LISTOF(aList);
	x = MINORBORDER;
	SetupFont(INFOFONT, INFOFONTSIZE); /* so we can check filename length */
	while (p)
	{
	    theFile = p -> thing;
	    fileName = GetString(GetVar(theFile, NAME));
	    itemWid = StrWidth(fileName);
	    if (itemWid < INFOITEMWID) itemWid = INFOITEMWID;
	    textbox = NewTextBox(x, x + itemWid, 
		     -INFOFIELDHEIGHT + MINORBORDER, -MINORBORDER,
		    PLAIN, "File Info", GetFileInfo(path, fileName));
	    PrefixList(contents, textbox);
	    SetVar(textbox, PARENT, field);
	    SetTextFont(textbox, INFOFONT);
	    SetTextSize(textbox, INFOFONTSIZE);
	    SetTextLineSpace(textbox, INFOITEMHT - INFOFONTSIZE);
	    p = p -> next;
	    x = x + itemWid + 2*MINORBORDER;
	}
	ScrollHome(field);
	RecalcScroll(field);
    }
    else /* create info window */
    {
	ObjPtr field, panel, contents, button;
	ThingListPtr p;
	ObjPtr theFile;
	char *fileName;
	int x, itemWid;

	contents = GetVar((ObjPtr) infoWin, CONTENTS); /* contents of window */

	/* put in help string */
	SetVar((ObjPtr) infoWin, HELPSTRING, NewString("This window shows directory \
information for the selected file(s) in the file window. If more than one file was selected, \
you can use the scrollbar at the bottom of the window to view information for each of them."));

	/* put in panel */
	panel = NewPanel(greyPanelClass, 0, INFOWIDTH, 0, INFOHEIGHT);
	PrefixList(contents, panel);
	SetVar(panel, PARENT, (ObjPtr) infoWin);

	contents = GetVar(panel, CONTENTS); /* contents of panel */

	/* put path textbox at top */
	textbox = NewTextBox(MINORBORDER, INFOWIDTH - MINORBORDER,
			INFOHEIGHT - MAJORBORDER - 2*INFOFONTSIZE - 2*DEFLINESPACE,
			INFOHEIGHT - MAJORBORDER, PLAIN, "Pathname", GetString(thePath));
	SetVar(textbox, PARENT, panel);
	PrefixList(contents, textbox);
	SetTextFont(textbox, INFOFONT);
	SetTextSize(textbox, INFOFONTSIZE);
	SetTextAlign(textbox, RIGHTALIGN);
	SetVar((ObjPtr) infoWin, REPOBJ, textbox); /* ref to by this routine */

	/* put in "fieldnames" textbox */
	textbox = NewTextBox(MINORBORDER, MINORBORDER + INFONAMEWIDTH, 
		MINORBORDER,
		INFOHEIGHT-MAJORBORDER-2*INFOFONTSIZE-2*DEFLINESPACE-GAP-MINORBORDER-1,
		PLAIN, "Fieldnames", 
		"Name\nSize\nOwner\nMode\nModified");
	SetVar(textbox, PARENT, panel);
	PrefixList(contents, textbox);
	SetTextFont(textbox, INFOFONT);
	SetTextSize(textbox, INFOFONTSIZE);
	SetTextAlign(textbox, RIGHTALIGN);
	SetTextLineSpace(textbox, INFOITEMHT-INFOFONTSIZE);

	/* put in field for info text boxes */
	field = NewControlField(2*MINORBORDER + INFONAMEWIDTH, INFOWIDTH - MINORBORDER,
		MINORBORDER, MINORBORDER + INFOFIELDHEIGHT + BARWIDTH, 
		"Info Field", OBJECTSFROMTOP + BARBOTTOM);
	SetVar(field, BACKGROUND, NewInt(UIGRAY75));
	SetVar(field, BORDERTYPE, NewInt(1));
	SetVar(field, PARENT, panel);
	PrefixList(contents, field);
	SetVar(field, CONTENTS, contents = NewList());
	SetVar((ObjPtr) infoWin, FIELD, field);

	/* put in file info textbox for each file on the list */
	p = LISTOF(aList);
	x = MINORBORDER;
	SetupFont(INFOFONT, INFOFONTSIZE); /* so we can check filename length */
	while (p)
	{
	    theFile = p -> thing;
	    fileName = GetString(GetVar(theFile, NAME));
	    itemWid = StrWidth(fileName);
	    if (itemWid < INFOITEMWID) itemWid = INFOITEMWID;
	    textbox = NewTextBox(x, x + itemWid, 
		     -INFOFIELDHEIGHT + MINORBORDER, -MINORBORDER,
		    PLAIN, "File Info", GetFileInfo(path, fileName));
	    PrefixList(contents, textbox);
	    SetVar(textbox, PARENT, field);
	    SetTextFont(textbox, INFOFONT);
	    SetTextSize(textbox, INFOFONTSIZE);
	    SetTextLineSpace(textbox, INFOITEMHT - INFOFONTSIZE);
	    p = p -> next;
	    x = x + itemWid + 2*MINORBORDER;
	}
	RecalcScroll(field);
    }
    if (logging) InhibitLogging(false);
    RestoreDir(); /* return to original working directory */
    return;
}

#ifdef HAVE_PROTOTYPES
static char *GetFileInfo(char *path, char *fileName)
#else
static char *GetFileInfo(path, fileName)
char *path, *fileName;
#endif
{
    char *p = tempStr, *s, buf[20];
    struct stat stbuf;
    struct passwd *pwbuf;
    struct tm *tmbuf;
    short mode;

    s = fileName;
    while (*p++ = *s++) ;
    *(p - 1) = '\n';

    /* get file info */
    if (stat(fileName, &stbuf) == -1) /* This happens! (eg, link to a bin, dead link) */
    {
	    strcpy(p, "?\n?\n?\n?");
    }
    else
    {
	/* get size in K */
	sprintf(buf,"%dK",(stbuf.st_size + 1023)/1024);
	s = buf;
	while (*p++ = *s++) ;
	*(p - 1) = '\n';

	/* get owner name */
	pwbuf = getpwuid(stbuf.st_uid);
	if (pwbuf)
	{
		s = pwbuf -> pw_name;
	}
	else
	{
		s = "(unavailable)";
	}
	while (*p++ = *s++) ; /* copy name */
	*(p - 1) = '\n';

	/* get mode bits */
	mode = stbuf.st_mode;
	*p++ = mode & S_IRUSR ? 'r' : '-';
	*p++ = mode & S_IWUSR ? 'w' : '-';
	*p++ = mode & S_IXUSR ? 'x' : '-';
	*p++ = ' ';
	*p++ = mode & S_IRGRP ? 'r' : '-';
	*p++ = mode & S_IWGRP ? 'w' : '-';
	*p++ = mode & S_IXGRP ? 'x' : '-';
	*p++ = ' ';
	*p++ = mode & S_IROTH ? 'r' : '-';
	*p++ = mode & S_IWOTH ? 'w' : '-';
	*p++ = mode & S_IXOTH ? 'x' : '-';
	*p++ = '\n';

	/* get modified date */
	tmbuf = localtime(&stbuf.st_mtime);
	sprintf(buf,"%02d/%02d/%02d %02d:%02d\n", tmbuf->tm_mon + 1, tmbuf->tm_mday,
		tmbuf->tm_year, tmbuf->tm_hour, tmbuf->tm_min);
	s = buf;
	while (*p = *s++) ++p;
    }
    return tempStr;
}

/********************************************************************* FILE FORMAT DIALOG
*	Puts up a dialog for selecting data file format.
*	The OKRoutine is called with the selected format and list of files:
*
*		void OKRoutine(int fmt, ObjPtr list)
*/
#ifdef HAVE_PROTOTYPES
static void FileFormatDialog(FuncTyp OKRoutine)
#else
static void FileFormatDialog(OKRoutine)
FuncTyp OKRoutine;
#endif
{
    WinInfoPtr theDialog;
    ObjPtr panel, alertPanel, contents, promptBox, radioBtn, cancelBtn, okBtn;
    ObjPtr thePrompt, radioGroup, keyList, fileReaders;
    int row, col, nRows, nCols, totWid, totHt;
    int n, nBtns;

    /* make a list of all the file readers */
    keyList = NewList();
    PostfixList(keyList, NewSymbol(CLASSID));
    PostfixList(keyList, NewInt(CLASS_FILEREADER));
    fileReaders = SearchDatabase(keyList);

    nBtns = ListCount(fileReaders) + 1;

    /* first compute sizes from the number of buttons */
    /* number of buttons is one more than number of formats to allow for "none" */
    if (nBtns < 11) nCols = 2, nRows = (nBtns + 1)/2;
    else if (nBtns < 25) nCols = 3, nRows = (nBtns + 2)/3;
    else nCols = 4, nRows = (nBtns + 3)/4;
    totWid = 2*SIDEMARGIN + nCols*RADIOWID + (nCols - 1)*GAP;
    totHt = BOTMARGIN + OKBTNHT + 2*INTERSPACE + nRows*CHECKBOXHEIGHT + (nRows - 1)*GAP
	    + PROMPTTEXTSIZE + TOPMARGIN;

    /* see if dialog exists */
    if (!selWinInfo) return;
    thePrompt = NewString("Set file format to:");
    theDialog = GetDialog(selWinInfo, thePrompt, "Set Data Format",
	    totWid, totHt, totWid, totHt, 
	    WINUI + WINFIXEDSIZE + WINCENTERED + WINNOFRAME);
    if(!theDialog) return;

    if (GetVar((ObjPtr) theDialog, REPOBJ)) /* dialog exists */
    {
	/* pop the dialog */
	PopWindow(theDialog);
}
    else /* make the dialog box */
    {
	ThingListPtr runner;

	SetVar((ObjPtr) theDialog, REPOBJ, thePrompt); /* mark dialog */
	contents = GetVar((ObjPtr) theDialog, CONTENTS); /* contents of window */
	SetVar((ObjPtr) theDialog, HELPSTRING, NewString("This window shows \
the data file formats recognized by this version of SciAn. Click on the proper format \
for the selected files and press the OK button. For more information, see the section \
on data file formats in the SciAn User's manual."));

	/* put in panel */
	panel = NewPanel(greyPanelClass, 0, totWid, 0, totHt);
	PrefixList(contents, panel);
	SetVar(panel, PARENT, (ObjPtr) theDialog);

	contents = GetVar(panel, CONTENTS); /* contents of panel */

	/* put in pretty panel */
	alertPanel = NewAlertPanel(0, totWid, 0, totHt, "Fmt Panel");
	PrefixList(contents, alertPanel);
	SetVar(alertPanel, PARENT, panel);

	/* put in prompt */
	promptBox = NewTextBox(SIDEMARGIN, totWid - SIDEMARGIN,
			totHt - TOPMARGIN - PROMPTTEXTSIZE, totHt - TOPMARGIN,
			PLAIN, "Prompt Box", GetString(thePrompt));
	PrefixList(contents, promptBox);
	SetVar(promptBox, PARENT, panel);
	SetVar(promptBox, UIFONT, NewInt(PROMPTTEXTFONT));
	SetTextColor(promptBox, PROMPTTEXTCOLOR);

	/* put in the bottom buttons */
	cancelBtn = NewButton(totWid - SIDEMARGIN - 2*OKBTNWID - GAP,
			totWid - SIDEMARGIN - OKBTNWID - GAP,
			BOTMARGIN, BOTMARGIN + OKBTNHT, "Cancel");
	PrefixList(contents, cancelBtn);
	SetVar(cancelBtn, PARENT, panel);
	SetMethod(cancelBtn, CHANGEDVALUE, CancelDialog);
	SetVar(cancelBtn, HELPSTRING, NewString("Pressing this button \
will cancel this dialog window without changing the format of the selected file(s)."));

	okBtn = NewButton(totWid - SIDEMARGIN - OKBTNWID,
			totWid - SIDEMARGIN,
			BOTMARGIN, BOTMARGIN + OKBTNHT, "OK");
	PrefixList(contents, okBtn);
	SetVar(okBtn, PARENT, panel);
	SetMethod(okBtn, OKMETHOD, OKRoutine); /* for ref by OKDialog */
	SetMethod(okBtn, CHANGEDVALUE, OKDialog);
	SetVar(okBtn, HELPSTRING, NewString("Pressing this button \
will dismiss this dialog window and set the format of the selected file(s)."));
	SetVar(okBtn, REPOBJ, (ObjPtr) theDialog);

	/* oh, yeah -- guess we need a few radio buttons */
	radioGroup = NewRadioButtonGroup("Data Format");
	PrefixList(contents, radioGroup);
	SetVar(radioGroup, PARENT, panel);
	runner = LISTOF(fileReaders);
	for (n=0, row=nRows-1; row>=0 && runner; --row)
	{
	    for (col = 0; col<nCols && runner; ++col)
	    {
		ObjPtr theName;
		theName = GetVar(runner -> thing, NAME);
		radioBtn = NewRadioButton(
			SIDEMARGIN + col*(RADIOWID + GAP),
			SIDEMARGIN + (col+1)*RADIOWID + col*GAP,
			BOTPART + row*(CHECKBOXHEIGHT + GAP),
			BOTPART + (row+1)*CHECKBOXHEIGHT + row*GAP,
			theName ? GetString(theName) : "");
		AddRadioButton(radioGroup, radioBtn);
		runner = runner -> next; 
	    }
	}
	/* now add "none" button */
	row = 0; col = nCols - 1;
	radioBtn = NewRadioButton(
		SIDEMARGIN + col*(RADIOWID + GAP),
		SIDEMARGIN + (col+1)*RADIOWID + col*GAP,
		BOTPART + row*(CHECKBOXHEIGHT + GAP),
		BOTPART + (row+1)*CHECKBOXHEIGHT + row*GAP,
		"NONE");
	AddRadioButton(radioGroup, radioBtn);
	
	SetVar(okBtn, DATA, radioGroup); /* ref by OKDialog */
    }
    return;
}

static ObjPtr OKDialog(btn)
ObjPtr btn;
{
    FuncTyp OKRtn;
    ObjPtr list, dialog;
    int val;

    list = GetVar(GetVar(btn, REPOBJ), FILELIST); /* btn's REPOBJ is dialog */
    val = GetInt(GetVar(GetVar(btn, DATA), VALUE));
    OKRtn = GetMethod(btn, OKMETHOD);

    DeferClose();
    OKRtn(val, list);
    return ObjTrue;
}

static ObjPtr CancelDialog(btn)
ObjPtr btn;
{
    DeferClose();
    return ObjFalse;
}

/******************************************************************* GET FILE WINDOW */
#ifdef HAVE_PROTOTYPES
void GetFileWindow(char *dirName)
#else
void GetFileWindow(dirName)
char *dirName;
#endif

/* Create a(nother) file browser */
{
    WinInfoPtr fileWin;
    ObjPtr panel, contents, corral, thePath;
    ObjPtr tb, btn, theDir;
    int btnWid;
    FuncTyp method;

    /* make name for window */
    sprintf(tempStr, "File Browser %d", ++fileBrowserNumber);
    
    /* get the dialog window */
    fileWin = NewObjWindow(dialogClass, tempStr, WINUI, FILESWIDTH, BROWSERHEIGHT, SCRWIDTH, SCRHEIGHT);
    if (fileWin)
    {
	SetVar((ObjPtr) fileWin, CONTENTS, contents = NewList());
	SetVar((ObjPtr) fileWin, OWNERWINDOW, NULLOBJ);
	SetMethod((ObjPtr) fileWin, SHOWFILES, ShowBrowseFiles);
    }
    else return;

    /* create file window */
    InhibitLogging(true);
    thePath = NewString(dirName);

    /* put in help string */
    SetVar((ObjPtr) fileWin, HELPSTRING, NewString("This window shows the files \
and sub-directories located in the directory given at the top of the window. The files appear as \
icons corresponding to file type; sub-directories appear as folders. To open a \
folder, select it and press the Open button, or simply double click on it."));

    /* put in panel */
    panel = NewPanel(greyPanelClass, 0, FILESWIDTH, 0, BROWSERHEIGHT);
    PrefixList(contents, panel);
    SetVar(panel, PARENT, (ObjPtr) fileWin);
    SetVar(panel, STICKINESS, NewInt(STICKYALL));

    contents = GetVar(panel, CONTENTS); /* contents of panel */
    
    /* put in directory name text box */
    tb = NewTextBox(MAJORBORDER, FILESWIDTH - MINORBORDER - MINBTNWID - SMALLGAP,
	BROWSERHEIGHT - (MAJORBORDER + 3*PROMPTTEXTSIZE + 2*DEFLINESPACE),
	BROWSERHEIGHT - MAJORBORDER, PLAIN, "Directory", "");
    SetVar(tb, STICKINESS, NewInt(STICKYLEFT + STICKYRIGHT + STICKYTOP));
    PrefixList(contents, tb);
    SetVar(tb, PARENT, panel);
    SetVar(tb, UIFONT, NewInt(UIBOLDNORMALFONT));
    SetVar(tb, OWNERWINDOW, (ObjPtr) fileWin);
    SetVar((ObjPtr)fileWin, TITLETEXT, tb);
    SetVar(tb, PROMPT, NewString("Files in directory "));
    SetMethod(tb, UPDATE, UpdateTitle);
    DeferMessage(tb, UPDATE);
    
    /* put in file filter checkbox */
    btn = NewCheckBox(MINORBORDER, MINORBORDER + FILTERBTNWID,
	BROWSERHEIGHT - FILESTOP + SMALLGAP, BROWSERHEIGHT - FILESTOP + SMALLGAP + CHECKBOXHEIGHT, 
	"Show SciAn files only", true);
    PrefixList(contents, btn);
    SetVar(btn, PARENT, panel);
    SetVar(btn, OWNERWINDOW, (ObjPtr) fileWin);
    SetVar(btn, STICKINESS, NewInt(STICKYTOP + STICKYLEFT));
    SetVar(btn, HELPSTRING, NewString("When this box is checked, only SciAn \
data, log, and object files are shown; the files must have the proper filename \
extensions to be recognized by SciAn. If it is not checked, all files are shown. \
Folders are always shown."));
    SetMethod(btn, CHANGEDVALUE, ReShowFiles);

#ifndef RELEASE
    /* put in view mode checkbox */
    btn = NewCheckBox(FILESWIDTH - MINORBORDER - VIEWBTNWID, FILESWIDTH - MINORBORDER,
	BROWSERHEIGHT - FILESTOP + SMALLGAP, BROWSERHEIGHT - FILESTOP + SMALLGAP + CHECKBOXHEIGHT, 
	"Show List View", false);
    SetVar(btn, PARENT, panel);
    PrefixList(contents, btn);
    SetVar(btn, STICKINESS, NewInt(STICKYTOP + STICKYRIGHT));
    SetVar(btn, OWNERWINDOW, (ObjPtr) fileWin);
    SetVar(btn, HELPSTRING, NewString("When this box is checked, files are shown in \
a list; if it is not checked, files are shown as icons. Only view by icons is implemented in this version."));

    SetVar(btn, ACTIVATED, ObjFalse); /*** list view not implemented -- grey out btn ***/
    SetMethod(btn, CHANGEDVALUE, ChangeView);
#endif

    /* put in Open Parent button */
    btn = NewButton(FILESWIDTH - MINORBORDER - MINBTNWID, FILESWIDTH - MINORBORDER,
	BROWSERHEIGHT - MINORBORDER - OKBTNHT, BROWSERHEIGHT - MINORBORDER, 
	    "Open Parent");
    PrefixList(contents, btn);
    SetVar(btn, STICKINESS, NewInt(STICKYTOP + STICKYRIGHT));
    SetVar(btn, PARENT, panel);
    SetVar(btn, OWNERWINDOW, (ObjPtr) fileWin);
    SetVar(btn, HELPSTRING, NewString("Pressing this button will \
change directories to the next higher directory level. The current directory \
will appear as a folder in the new directory."));
    SetMethod(btn, CHANGEDVALUE, DoParent);

    /* put in Open Folder button */
    btn = NewButton(FILESWIDTH - MINORBORDER - MINBTNWID, FILESWIDTH - MINORBORDER,
	BROWSERHEIGHT - MINORBORDER - 2*OKBTNHT - SMALLGAP,
	BROWSERHEIGHT - MINORBORDER - OKBTNHT - SMALLGAP, 
	    OF_OPEN);
    AssocButtonWithObjFunction(btn, OF_OPEN);
    PrefixList(contents, btn);
    SetVar(btn, STICKINESS, NewInt(STICKYTOP + STICKYRIGHT));
    SetVar(btn, PARENT, panel);

    /* put in icon corral */
    corral = NewIconCorral(NULLOBJ, MINORBORDER, FILESWIDTH - MINORBORDER,
	BROWSERHEIGHT - FILESTOP - CORRALHEIGHT, BROWSERHEIGHT - FILESTOP,
	OBJECTSFROMTOP + BARRIGHT + BARBOTTOM);
    SetVar(corral, STICKINESS, NewInt(STICKYALL));
    PrefixList(contents, corral);
    SetVar(corral, PARENT, panel);
    SetVar((ObjPtr) fileWin, CORRAL, corral);
    SetVar(corral, OWNERWINDOW, (ObjPtr) fileWin);
    SetVar(corral, ALLFILES, ObjFalse);	/* file filter switch = SciAn files only */
    SetMethod(corral, DROPINCONTENTS, DropInFileWin);

    /* put in action buttons */
    btnWid = (FILESWIDTH - 2*MINORBORDER - 2*SMALLGAP)/3;
    btn = NewButton(MINORBORDER, MINORBORDER + btnWid,
		    MINORBORDER, MINORBORDER + OKBTNHT, OF_READ);
    AssocButtonWithObjFunction(btn, OF_READ);
    if (btn)
    {
	    SetVar(btn, PARENT, panel);
	    PrefixList(contents, btn);
	    SetVar(btn, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM + FLOATINGRIGHT));
    }

    btn = NewButton(MINORBORDER + btnWid + SMALLGAP, 
    MINORBORDER + SMALLGAP + 2*btnWid,
		    MINORBORDER, MINORBORDER + OKBTNHT,
		    OF_SETFORMAT);
    AssocButtonWithObjFunction(btn, OF_SETFORMAT);
    if (btn)
    {
	    SetVar(btn, PARENT, panel);
	    PrefixList(contents, btn);
	    SetVar(btn, STICKINESS, NewInt(STICKYBOTTOM + FLOATINGRIGHT + FLOATINGLEFT));
    }

    btn = NewButton(FILESWIDTH - MINORBORDER - btnWid, FILESWIDTH - MINORBORDER,
		    MINORBORDER, MINORBORDER + OKBTNHT,
		    OF_SHOWINFO);
    AssocButtonWithObjFunction(btn, OF_SHOWINFO);
    if (btn)
    {
	    SetVar(btn, PARENT, panel);
	    PrefixList(contents, btn);
	    SetVar(btn, STICKINESS, NewInt(FLOATINGLEFT + STICKYRIGHT + STICKYBOTTOM));
    }

    theDir = GetDirectory(dirName);
    if (theDir)
    {
	SetVar((ObjPtr) fileWin, REPOBJ, theDir); /* REPOBJ of file window is the directory */
    }
    else
    {
	ReportError("GetFileWindow", "GetDirectory failed");
	return;
    }

    /* now put in the file icons */
    method = GetMethod((ObjPtr) fileWin, SHOWFILES);
    if (method)
    {
	(*method)(fileWin);
    }
    InhibitLogging(false);
}

static ObjPtr UpdateTitle(ObjPtr tb)
{
    ObjPtr theDir, thePrompt, theTitle;
    WinInfoPtr theWin;
    
    theWin = (WinInfoPtr) GetVar(tb, OWNERWINDOW);
    theDir = GetVar((ObjPtr) theWin, REPOBJ);
    theTitle = GetVar(theDir, NAME);
    thePrompt = GetVar(tb, PROMPT);
    strcpy(tempStr, GetString(thePrompt));
    strcat(tempStr, GetString(theTitle));
    SetTextBox(tb, tempStr);
}

#if 1
/******************************************************************* GET SAVE WINDOW */
#ifdef HAVE_PROTOTYPES
void GetSaveWindow(ObjPtr saveObj, char *dirName, char *initName, char *kind)
#else
void GetSaveWindow(saveObj, dirName, initName, kind)
ObjPtr saveObj;	/* object to be saved */
char *dirName;	/* default directory pathname */
char *initName;	/* initial name */
char *kind;	/* kind of file being saved */
#endif

/* Create Save dialog */
{
    WinInfoPtr saveWin;
    ObjPtr panel, contents, corral, thePath;
    ObjPtr tb, btn, btnGroup, theDir;
    int btnWid;
    FuncTyp method;

    /* make name for window */
    sprintf(tempStr, "Save File %d", ++saveWindowNumber);
    
    /* get the dialog window */
    saveWin = NewObjWindow(dialogClass, tempStr, WINUI, 
	    FILESWIDTH, SAVEWINHEIGHT, SCRWIDTH, SCRHEIGHT); /* sic */
    if(saveWin)
    {
	SetVar((ObjPtr) saveWin, CONTENTS, contents = NewList());
	SetVar((ObjPtr) saveWin, OWNERWINDOW, NULLOBJ);
	SetMethod((ObjPtr) saveWin, SHOWFILES, ShowSaveFiles);
    }
    else return;

    /* create save window */
    InhibitLogging(true);
    thePath = NewString(dirName);

    /* put in help string */
    SetVar((ObjPtr) saveWin, HELPSTRING,
	    NewString("This window allows you to name and save SciAn files. \
The file will be saved in the directory shown at the top of the window. \
You can change directories by using the Open Parent button and by opening \
folders, exactly as in the File Browser window. The file will be given the \
name in the text box below the icon corral. The window will close automatically \
when you click the Save or Cancel button."));

    /* put in panel */
    panel = NewPanel(greyPanelClass, 0, FILESWIDTH, 0, SAVEWINHEIGHT); /* sic */
    PrefixList(contents, panel);
    SetVar(panel, PARENT, (ObjPtr) saveWin);
    SetVar(panel, STICKINESS, NewInt(STICKYALL));

    contents = GetVar(panel, CONTENTS); /* contents of panel */

    /* put in text box for prompt and directory */
    tb = NewTextBox(MAJORBORDER, FILESWIDTH - MINORBORDER - MINBTNWID - SMALLGAP,
	SAVEWINHEIGHT - (MAJORBORDER + 2*OKBTNHT + SMALLGAP),
	SAVEWINHEIGHT - MAJORBORDER, PLAIN, "Directory", "");
    SetVar(tb, STICKINESS, NewInt(STICKYLEFT + STICKYRIGHT + STICKYTOP));
    PrefixList(contents, tb);
    SetVar(tb, PARENT, panel);
    SetVar(tb, UIFONT, NewInt(UIBOLDNORMALFONT));
    SetVar(tb, OWNERWINDOW, (ObjPtr) saveWin);
    SetVar((ObjPtr)saveWin, TITLETEXT, tb);
    sprintf(tempStr, "Save %s file in directory ", kind);
    SetVar(tb, PROMPT, NewString(tempStr));
    SetMethod(tb, UPDATE, UpdateTitle);
    DeferMessage(tb, UPDATE);
    
    /* put in file filter checkbox */
    btn = NewCheckBox(MINORBORDER, MINORBORDER + FILTERBTNWID,
	SAVEWINHEIGHT - FILESTOP + SMALLGAP, SAVEWINHEIGHT - FILESTOP + SMALLGAP + CHECKBOXHEIGHT, 
	"Show SciAn files only", true);
    PrefixList(contents, btn);
    SetVar(btn, PARENT, panel);
    SetVar(btn, OWNERWINDOW, (ObjPtr) saveWin);
    SetVar(btn, STICKINESS, NewInt(STICKYTOP + STICKYLEFT));
    SetVar(btn, HELPSTRING, NewString("When this box is checked, only SciAn \
data, log, and object files are shown; the files must have the proper filename \
extensions to be recognized by SciAn. If it is not checked, all files are shown. \
Folders are always shown."));
    SetMethod(btn, CHANGEDVALUE, ReShowFiles);

#ifndef RELEASE
    /* put in view mode checkbox */
    btn = NewCheckBox(FILESWIDTH - MINORBORDER - VIEWBTNWID, FILESWIDTH - MINORBORDER,
	SAVEWINHEIGHT - FILESTOP + SMALLGAP, SAVEWINHEIGHT - FILESTOP + SMALLGAP + CHECKBOXHEIGHT, 
	"Show List View", false);
    SetVar(btn, PARENT, panel);
    PrefixList(contents, btn);
    SetVar(btn, STICKINESS, NewInt(STICKYTOP + STICKYRIGHT));
    SetVar(btn, OWNERWINDOW, (ObjPtr) saveWin);
    SetVar(btn, HELPSTRING, NewString("When this box is checked, files are shown in \
a list; if it is not checked, files are shown as icons. Only view by icons is implemented in this version."));

    SetVar(btn, ACTIVATED, ObjFalse); /*** list view not implemented -- grey out btn ***/
    SetMethod(btn, CHANGEDVALUE, ChangeView);
    
    /* put in Make Backup checkbox */
    btn = NewCheckBox(MINORBORDER, MINORBORDER + BACKBTNWID,
	MINORBORDER, MINORBORDER + CHECKBOXHEIGHT, "Make Backup", true);
    SetVar(btn, PARENT, panel);
    PrefixList(contents, btn);
    SetVar(btn, STICKINESS, NewInt(STICKYBOTTOM + STICKYLEFT));
    SetVar(btn, OWNERWINDOW, (ObjPtr) saveWin);
    SetVar(btn, HELPSTRING, NewString("When this box is checked, if the saved file has the same \
name as an existing file, that file is renamed by appending .bak before the new file is written."));
    SetMethod(btn, CHANGEDVALUE, ChangeBack);
    SetVar((ObjPtr) saveWin, MAKEBACKUP, ObjTrue);
#endif

    /* put in Open Parent button */
    btn = NewButton(FILESWIDTH - MINORBORDER - MINBTNWID, FILESWIDTH - MINORBORDER,
	SAVEWINHEIGHT - MINORBORDER - OKBTNHT, SAVEWINHEIGHT - MINORBORDER, 
	    "Open Parent");
    PrefixList(contents, btn);
    SetVar(btn, STICKINESS, NewInt(STICKYTOP + STICKYRIGHT));
    SetVar(btn, PARENT, panel);
    SetVar(btn, OWNERWINDOW, (ObjPtr) saveWin);
    SetVar(btn, HELPSTRING, NewString("Pressing this button will \
change directories to the next higher directory level. The current directory \
will appear as a folder in the new directory."));
    SetMethod(btn, CHANGEDVALUE, DoParent);

    /* put in Open Folder button */
    btn = NewButton(FILESWIDTH - MINORBORDER - MINBTNWID, FILESWIDTH - MINORBORDER,
	SAVEWINHEIGHT - MINORBORDER - 2*OKBTNHT - SMALLGAP,
	SAVEWINHEIGHT - MINORBORDER - OKBTNHT - SMALLGAP, 
	    OF_OPEN);
    AssocButtonWithObjFunction(btn, OF_OPEN);
    PrefixList(contents, btn);
    SetVar(btn, STICKINESS, NewInt(STICKYTOP + STICKYRIGHT));
    SetVar(btn, PARENT, panel);

    /* put in icon corral */
    corral = NewIconCorral(NULLOBJ, MINORBORDER, FILESWIDTH - MINORBORDER,
		    SAVEWINHEIGHT - FILESTOP - CORRALHEIGHT, SAVEWINHEIGHT - FILESTOP,
		    OBJECTSFROMTOP + BARRIGHT + BARBOTTOM);
    SetVar(corral, STICKINESS, NewInt(STICKYALL));
    PrefixList(contents, corral);
    SetVar(corral, PARENT, panel);
    SetVar((ObjPtr) saveWin, CORRAL, corral);
    SetVar(corral, OWNERWINDOW, (ObjPtr) saveWin);
    SetVar(corral, ALLFILES, ObjFalse);	/* file filter switch = SciAn files only */
    SetMethod(corral, DROPINCONTENTS, DropInFileWin);

    /* put in label for filename textbox */
    tb = NewTextBox(FILESWIDTH - MINORBORDER - FILENAMEWID - TEXTBOXSEP - FILENAMELABEL, 
	FILESWIDTH - MINORBORDER - FILENAMEWID - TEXTBOXSEP,
	MINORBORDER + OKBTNHT + SMALLGAP + EDITBOXHEIGHT - PROMPTTEXTSIZE - EDGE + EDITBOXDOWN, 
	MINORBORDER + OKBTNHT + SMALLGAP + EDITBOXHEIGHT + EDITBOXDOWN - EDGE, ONE_LINE,
	"Filename Label", "Filename:");
    PrefixList(contents, tb);
    SetVar(tb,  PARENT, panel);
    SetVar(tb, STICKINESS, NewInt(STICKYBOTTOM + STICKYRIGHT));
    SetVar(tb, UIFONT, NewInt(UIBOLDNORMALFONT));
    SetTextAlign(tb, RIGHTALIGN);
    
    /* put in textbox for filename */
    tb = NewTextBox(FILESWIDTH - MINORBORDER - FILENAMEWID, FILESWIDTH - MINORBORDER,
	MINORBORDER + OKBTNHT + SMALLGAP, MINORBORDER + OKBTNHT + SMALLGAP + EDITBOXHEIGHT,
	EDITABLE + WITH_PIT + ONE_LINE, "Filename", initName);
    SetVar(tb, STICKINESS, NewInt(STICKYRIGHT + STICKYBOTTOM));
    PrefixList(contents, tb);
    SetVar(tb, PARENT, panel);
    SelectAll(tb);
    SetMethod(tb, ENTERMETHOD, DoEnter);
    SetVar((ObjPtr) saveWin, FILENAME, tb);
    
    /* put in Save and Cancel buttons */
    btn = NewButton(FILESWIDTH - MINORBORDER - OKBTNWID, FILESWIDTH - MINORBORDER,
	MINORBORDER, MINORBORDER + OKBTNHT, "Save"); /***/
    if (btn)
    {
	SetVar(btn, PARENT, panel);
	PrefixList(contents, btn);
	SetVar(btn, STICKINESS, NewInt(STICKYRIGHT + STICKYBOTTOM));
	SetVar(btn, HELPSTRING, NewString("Press this button to \
save the file and close this window."));
	SetVar(tb, BUTTON, btn);    /* referenced by DoEnter routine of filename tb */
	SetMethod(btn, CHANGEDVALUE, DoSave);
	SetVar(btn, OWNERWINDOW, (ObjPtr) saveWin);
	SetVar((ObjPtr) saveWin, SAVEOBJ, saveObj);

	/* put border around Save button to indicate default action */
	tb = NewTitleBox(FILESWIDTH - MINORBORDER - OKBTNWID - EDGE,
	    FILESWIDTH - MINORBORDER + EDGE - 1,
	    MINORBORDER - EDGE, MINORBORDER + OKBTNHT + EDGE - 1, NIL);
	PrefixList(contents, tb);
	SetVar(tb, PARENT, panel);
	SetVar(tb, STICKINESS, NewInt(STICKYRIGHT + STICKYBOTTOM));
    }

    
    btn = NewButton(FILESWIDTH - MINORBORDER - FILENAMEWID,
	FILESWIDTH - MINORBORDER - FILENAMEWID + OKBTNWID,
	MINORBORDER, MINORBORDER + OKBTNHT, "Cancel"); /***/
    if (btn)
    {
	SetVar(btn, PARENT, panel);
	PrefixList(contents, btn);
	SetVar(btn, STICKINESS, NewInt(STICKYRIGHT + STICKYBOTTOM));
	SetVar(btn, HELPSTRING, NewString("Press this button to \
close this window without saving the file."));
	SetVar(btn, OWNERWINDOW, (ObjPtr) saveWin);
	SetMethod(btn, CHANGEDVALUE, DoCancel);
    }
    
    theDir = GetDirectory(dirName);
    if (theDir)
    {
	SetVar((ObjPtr) saveWin, REPOBJ, theDir); /* REPOBJ of file window is the directory */
    }
    else
    {
	ReportError("GetSaveWindow", "GetDirectory failed");
	return;
    }

    /* now put in the file icons */
    method = GetMethod((ObjPtr) saveWin, SHOWFILES);
    if (method)
    {
	(*method)(saveWin);
    }
    InhibitLogging(false);
}

static ObjPtr DoEnter(replyBox)
ObjPtr replyBox;
{
    ObjPtr btn;

    btn = GetVar(replyBox, BUTTON); /* Save button */
    SetValue(btn, ObjTrue);
    return ObjTrue;
}

static ObjPtr SaveOrNot(win, n)
WinInfoPtr win;
int n;
/* callback for Replace File alert in DoSave: 0=Cancel, 1=Replace */
{
    if (n > 0)
    {
	ObjPtr var;
	FuncTyp method;
	char *fileName;
	
        fileName = GetString(GetValue(GetVar((ObjPtr) win, FILENAME)));
	if (GetPredicate((ObjPtr) win, MAKEBACKUP))
	{
	    sprintf(tempStr, "%s.bak", fileName);
	    rename(fileName, tempStr);
	}
	var = GetVar((ObjPtr) win, SAVEOBJ);
	if (var)
	{
	    method = GetMethod(var, SAVEFILE);
	    if (method)
	    {
		(*method)(var, fileName);
	    }
	}
	/* mark directory as changed */
	var = GetVar((ObjPtr) win, REPOBJ);
	if (var)
	{
	    SetVar(var, CHANGED, ObjTrue);
	}
	DeferMessage((ObjPtr) win, DISPOSE); /* close window */
	
    }
    RestoreDir();
    return ObjTrue;
}

static ObjPtr DoSave(btn)
ObjPtr btn;
{
    WinInfoPtr w;
    ObjPtr tb, var;
    ObjPtr repObj;
    ObjPtr saveObj;
    FuncTyp method;
    char *fileName;

    w = (WinInfoPtr) GetVar(btn, OWNERWINDOW);
    repObj = GetObjectVar("DoSave", (ObjPtr) w, REPOBJ);
    if (!repObj) return NULLOBJ;
    var = GetStringVar("DoSave", repObj, NAME);
    if (!var) return NULLOBJ;

    ChangeDir(GetString(var));

    tb = GetVar((ObjPtr) w, FILENAME);
    fileName = GetString(GetVar(tb, VALUE));
    if (FileExists(fileName))
    {
	/* ask before overwriting or renaming existing file */
	WinInfoPtr alert;
	
	sprintf(tempStr, "Replace existing file %s?", fileName);
	alert = AlertUser(UICAUTIONALERT, w, tempStr, SaveOrNot, 2, "Cancel", "Replace");
	if (alert)
	{
	    SetVar((ObjPtr) alert, HELPSTRING, NewString("There is already a file with \
the same name as the file you are saving. Press Cancel to go back to the Save \
Window and change the filename. Press Replace to write the new file with the \
same name; in this case, if the Make Backup checkbox in the Save Window is on, \
the existing file will be renamed by appending .bak to its name before writing \
the new file."));
	}
	return ObjTrue; 
    }
    saveObj = GetVar((ObjPtr) w, SAVEOBJ);
    if (saveObj)
    {
	method = GetMethod(saveObj, SAVEFILE);
	if (method)
	{
	    (*method)(saveObj, fileName);
	}
    }
    /* mark directory as changed */
    SetVar(repObj, CHANGED, ObjTrue);
    DeferMessage((ObjPtr) w, DISPOSE); /* close window */

    RestoreDir();

    return ObjTrue;
}

static ObjPtr DoCancel(btn)
ObjPtr btn;
{
    WinInfoPtr w;
    
    w = (WinInfoPtr) GetVar(btn, OWNERWINDOW);
    DeferMessage((ObjPtr) w, DISPOSE); /* close window */
    return NULLOBJ;
}


#endif

#ifdef HAVE_PROTOTYPES
ObjPtr IconHelp(ObjPtr icon, ObjPtr class)
#else
ObjPtr IconHelp(icon, class)
ObjPtr icon, class;
#endif
{
    int type = GetInt(GetVar(GetVar(icon, REPOBJ), FILETYPE));

    switch (type)
    {
	case DIRFILE:
	    SetVar(class, HELPSTRING, NewString("This icon represents a sub-directory \
within the directory of this file window. Double clicking on the folder will open a file window \
for the sub-directory."));
	    break;

	case DATAFILE:
	    SetVar(class, HELPSTRING, NewString("\
This icon represents a data file with a format recognized by this version of SciAn. The \
name of the format is shown under the file name. \
Double clicking on this icon will read the file into the Datasets window."));
	    break;

	case OTHFILE:
	    SetVar(class, HELPSTRING, NewString("This icon represents a file whose type \
SciAn could not determine from the filename extension. If the file is a data file, you must \
identify the data format before SciAn can read the file. Select the file, press the Set \
Format button, and then select the proper format."));
	    break;

	default:
	    SetVar(class, HELPSTRING, NewString("This icon represents a file in the \
directory of this file window."));
	    break;
    }
    return NULLOBJ;
}

/************************************************************* GET DIRECTORY
*	Looks for directory of given name in the data base.
*	If found, returns it, else makes a new directory object,
*	and adds the directory to the database.
*/
#ifdef HAVE_PROTOTYPES
static ObjPtr GetDirectory(char *dirName)
#else
static ObjPtr GetDirectory(dirName)
char *dirName;
#endif
{
    ObjPtr keyList, dirList;
    ThingListPtr p;

    /* search database for this directory */
    keyList = NewList();
    PostfixList(keyList, NewSymbol(CLASSID));
    PostfixList(keyList, NewInt(CLASS_DIR));
    PostfixList(keyList, NewSymbol(NAME));
    PostfixList(keyList, NewString(dirName));
    dirList = SearchDatabase(keyList);
    if (dirList && (p = LISTOF(dirList)))
    {
	if (p -> next) /* whoops! should only be one... */
	{
	    ReportError("GetDirectory", "Duplicate directory in database!");
	    return NULLOBJ;
	}
	else
	{
	    return p -> thing;	/* directory found */
	}
    }
    else
    {
	ObjPtr theDir;
	
	/* new directory */

	if (!ChangeDir(dirName))
	{
	    ReportError("GetDirectory", "Error changing directory");
	    return NULLOBJ;
	}

	/* now create a directory object */
	theDir = NewObject(dirClass, 0);
	SetVar(theDir, NAME, NewString(dirName));

	/* add directory to database */
	AddObjToDatabase(theDir);

	return theDir;
    }
}

#ifdef HAVE_PROTOTYPES
static ObjPtr FindMiscType(char *ext)
#else
static ObjPtr FindMiscType(ext)
char *ext;
#endif
/*Finds a miscellaneous SciAn type corresponding to ext*/
{
    ObjPtr keyList;
    ObjPtr result;

    keyList = NewList();
    PostfixList(keyList, NewSymbol(CLASSID));
    PostfixList(keyList, NewInt(CLASS_FILETYPE));
    PostfixList(keyList, NewSymbol(EXTENSION));
    PostfixList(keyList, NewString(ext));
    result = SearchDatabase(keyList);

    if (result && LISTOF(result))
    {
	return (LISTOF(result)) -> thing;
    }
    else
    {
	return NULLOBJ;
    }
}

#ifdef HAVE_PROTOTYPES
static ObjPtr MakeFileList(ObjPtr theDir)
#else
static ObjPtr MakeFileList()
ObjPtr theDir;
#endif
{
    struct stat stbuf;
#ifdef DIRENT
    struct dirent *dp;
#else
    struct direct *dp;
#endif
    DIR *dirp;
    ObjPtr fileList, thePath, miscType;
    char *dirName;

    LongOperation();

    /* Make FILELIST for given directory */
    fileList = NewList();

    /* read the directory */
    thePath = GetVar(theDir, NAME);
    dirName = GetString(thePath);
    if (!ChangeDir(dirName))
    {
	ReportError("MakeFileList", "Error changing directory");
	return ObjFalse;
    }
    dirp = opendir(dirName);
    while ((dp = readdir(dirp)) != NULL)
    {
	char *ext;
	ObjPtr theFile, theRdr;

	if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0) continue;
	stat(dp->d_name, &stbuf);
	if((stbuf.st_mode & S_IFMT) == S_IFLNK) continue; /* must be dead link */
	if ((stbuf.st_mode & S_IFMT) == S_IFSOCK) continue; /* socket file */

	if (*dp->d_name == '.') continue; /* ignore dot files */
	ext = strrchr(dp->d_name, '.'); /* point to filename extension */


	/* figure out file type */
	if((stbuf.st_mode & S_IFMT) == S_IFDIR)
	{
	    /* entry is a directory */
	    theFile = NewObject(unixDirClass, 0);
	    SetVar(theFile, NAME, NewString(dp->d_name));
	    SetVar(theFile, PATH, thePath);
	    SetVar(theFile, FILETYPE, NewInt(DIRFILE));
	    PrefixList(fileList, theFile);
	}
	else
	{
	    theFile = NewObject(unixFileClass, 0);
	    SetVar(theFile, NAME, NewString(dp->d_name));
	    SetVar(theFile, PATH, thePath);
	    
	    /* check file extension to identify file type */
	    if (ext++) /* skip dot */
	    {
		if (strcmp(ext, "o") == 0)
		{
		    /* object file */
		    SetVar(theFile, FILETYPE, NewInt(OBJFILE));
		    PrefixList(fileList, theFile);
		}
		else if (strcmp(ext, "txt") == 0
		    || strcmp(ext, "c") == 0
		    || strcmp(ext, "h") == 0
		    || strcmp(ext, "f") == 0 )
		{
		    /* text file */
		    SetVar(theFile, FILETYPE, NewInt(TEXTFILE));
		    PrefixList(fileList, theFile);
		}
		else if ( miscType = FindMiscType(ext))
		{
		    SetVar(theFile, FILETYPE, NewInt(SCIANFILE));
		    SetVar(theFile, FORMAT, GetVar(miscType, NAME));
		    PrefixList(fileList, theFile);
		}
		else if (strcmp(ext, "log") == 0)
		{
		    /* script or log file */
		    SetVar(theFile, FILETYPE, NewInt(LOGFILE));
		    SetVar(theFile, FORMAT, NewString("LOG"));
		    PrefixList(fileList, theFile);
		}
		else if ( theRdr = FindExtensionReader(ext) )
		{
		    /* known data file extension */
		    SetVar(theFile, FILETYPE, NewInt(DATAFILE));
		    SetVar(theFile, FORMAT, GetVar(theRdr, NAME));
		    PrefixList(fileList, theFile);
		}
		else /* other file type */
		{
		    SetVar(theFile, FILETYPE, NewInt(OTHFILE));
		    PrefixList(fileList, theFile);
		}
	    }
	    else if (stbuf.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))
	    {
		/* entry is (probably) executable file */
		SetVar(theFile, FILETYPE, NewInt(BINFILE));
		PrefixList(fileList, theFile);
	    }
	    else /* no dot, not executable: other file type */
	    {
		SetVar(theFile, FILETYPE, NewInt(OTHFILE));
		PrefixList(fileList, theFile);
	    }
	}
    }
    closedir(dirp);
    RestoreDir();
    
    /* sort file list and save on directory */
    SetVar(theDir, FILELIST, SortListByStringVar(fileList, NAME, false));
    return ObjTrue;
}

/*********************************************************************** SHOW FILES
*	Puts icons for the selected file types into the corral
*	of the given file browser
*/
static ObjPtr ShowBrowseFiles(fileWin)
WinInfoPtr fileWin;
{
    ObjPtr theDir, theFile, theIcon, theRdr, iconList;
    ObjPtr fileList, theType, thePath, var;
    ObjPtr corral, contents;
    int fileType, left, right, bottom, top, width;
    ThingListPtr p;
    FuncTyp ReScroll;
    Bool allFiles;
    int whichIcon;

    /* this may take a while */
    LongOperation();

    theDir = GetVar((ObjPtr) fileWin, REPOBJ);	/* get directory */
    if (!theDir || GetInt(GetVar(theDir, CLASSID)) != CLASS_DIR)
    {
	/* coding error */
	ReportError("ShowFiles", "Argument not a file window!");
	return ObjFalse;
    }
    MakeVar(theDir, FILELIST);
    fileList = GetVar(theDir, FILELIST);
    p = LISTOF(fileList);
    thePath = GetVar(theDir, NAME);
    
    corral = GetVar((ObjPtr) fileWin, CORRAL);
    Get2DIntBounds(corral, &left, &right, &bottom, &top);
    width = right - left;

    SetVar(corral, CONTENTS, contents = NewList());
    allFiles = GetPredicate(corral, ALLFILES);

    /* now put the selected filetype icons in the corral */
    while(p)
    {
	theFile = p -> thing;
	theType = GetVar(theFile, FILETYPE);
	fileType = GetInt(theType);
	if (allFiles || fileType == DIRFILE || fileType == DATAFILE
	     || fileType == LOGFILE || fileType == SCIANFILE)
	{
	    /* make icon for this file */
	    theIcon = NewObject(fileIconClass, 0L); 
	    SetVar(theIcon, NAME, GetVar(theFile, NAME));
	    SetVar(theIcon, REPOBJ, theFile);
	    SetVar(theIcon, PARENT, corral);
	    MakeVar(theIcon, WHICHICON);
	    SetVar(theIcon, PATH, thePath);
	    SetMethod(theIcon, MAKE1HELPSTRING, IconHelp);
	    SetVar(theIcon, ICONLOC, NULLOBJ);
	    PrefixList(contents, theIcon);
	}
	p = p -> next;
    }
    ScrollHome(corral);
    RecalcScroll(corral);
    ImInvalid(corral);	/* redraw if necessary */
    return ObjTrue; /* function is used as method */
}

/*********************************************************************** SHOW FILES
*	Puts icons for the selected file types into the corral
*	of the given file saver
*/
static ObjPtr ShowSaveFiles(fileWin)
WinInfoPtr fileWin;
{
    ObjPtr theDir, theFile, theIcon, theRdr, iconList;
    ObjPtr fileList, theType, thePath, var;
    ObjPtr corral, contents;
    int fileType, left, right, bottom, top, width;
    ThingListPtr p;
    FuncTyp ReScroll;
    Bool allFiles;
    int whichIcon;

    /* this may take a while */
    LongOperation();

    theDir = GetVar((ObjPtr) fileWin, REPOBJ);	/* get directory */
    if (!theDir || GetInt(GetVar(theDir, CLASSID)) != CLASS_DIR)
    {
	/* coding error */
	ReportError("ShowFiles", "Argument not a file window!");
	return ObjFalse;
    }
    MakeVar(theDir, FILELIST);
    fileList = GetVar(theDir, FILELIST);
    p = LISTOF(fileList);
    thePath = GetVar(theDir, NAME);
    
    corral = GetVar((ObjPtr) fileWin, CORRAL);
    Get2DIntBounds(corral, &left, &right, &bottom, &top);
    width = right - left;

    SetVar(corral, CONTENTS, contents = NewList());
    allFiles = GetPredicate(corral, ALLFILES);

    /* now put the selected filetype icons in the corral */
    while(p)
    {
	theFile = p -> thing;
	theType = GetVar(theFile, FILETYPE);
	fileType = GetInt(theType);
	if (allFiles || fileType == DIRFILE || fileType == DATAFILE
	     || fileType == LOGFILE || fileType == SCIANFILE)
	{
	    /* make icon for this file */
	    theIcon = NewObject(fileIconClass, 0L); 
	    SetVar(theIcon, NAME, GetVar(theFile, NAME));
	    SetVar(theIcon, REPOBJ, theFile);
	    if (fileType != DIRFILE)
	    {
		SetVar(theIcon, REPOBJONLY, ObjFalse);
		SetVar(theIcon, ICONGREYED, ObjTrue);
		SetMethod(theIcon, ICONGREYED, (FuncTyp) 0);
	    }
	    SetVar(theIcon, PARENT, corral);
	    MakeVar(theIcon, WHICHICON);
	    SetVar(theIcon, PATH, thePath);
	    SetMethod(theIcon, MAKE1HELPSTRING, IconHelp);
	    SetVar(theIcon, ICONLOC, NULLOBJ);
	    PrefixList(contents, theIcon);
	}
	p = p -> next;
    }
    ScrollHome(corral);
    RecalcScroll(corral);
    ImInvalid(corral);	/* redraw if necessary */
    return ObjTrue; /* function is used as method */
}

#ifdef HAVE_PROTOTYPES
static ObjPtr ReShowFiles(ObjPtr btn)
#else
static ObjPtr ReShowFiles(btn)
ObjPtr btn;
#endif
{
    WinInfoPtr w;
    ObjPtr corral;
    Bool allFiles;
    FuncTyp method;
    
    w = (WinInfoPtr) GetVar(btn, OWNERWINDOW);
    corral = GetVar((ObjPtr) w, CORRAL);
    allFiles = GetPredicate(corral, ALLFILES);
    SetVar(corral, ALLFILES, allFiles ? ObjFalse : ObjTrue);
    method = GetMethod((ObjPtr) w, SHOWFILES);
    if (method)
    {
	(*method)(w);
    }
    return NULLOBJ;
}

#ifdef HAVE_PROTOTYPES
static ObjPtr ChangeView(ObjPtr btn)
#else
static ObjPtr ChangeView(btn)
ObjPtr btn;
#endif
{
    return NULLOBJ; /*** stub ***/
}

#ifdef HAVE_PROTOTYPES
static ObjPtr ChangeBack(ObjPtr btn)
#else
static ObjPtr ChangeBack(btn)
ObjPtr btn;
#endif
{
    ObjPtr w;
    
    w = GetVar(btn, OWNERWINDOW);
    SetVar(w, MAKEBACKUP, (IsTrue(GetValue(btn)) ? ObjTrue : ObjFalse));
    return ObjTrue;
}

static ObjPtr DropInFileWin(corral, icon, x, y)
ObjPtr corral, icon;
int x,y;
{
#if 0
	ObjPtr theFile, theFormat, theReader;
	int fileType;
	theFile = GetVar(icon, REPOBJ);
	if (theFile && IsFile(theFile)
	{
		fileType = GetInt(GetVar(theFile, FILETYPE));
		if (fileType == DIRFILE)
		{
			DoUniqueTask(ReportCantDoIt);
			return ObjFalse;
		}
		/* ok: to copy file need to make new icon and file objects, copy vars,
			and update fileList, corral contents */
		
	}
#else
	printf("Drop in file window not implemented.\n");
#endif
	return ObjFalse;
}

/*********************************************************************** DO PARENT
*   CHANGEDVALUE method for Open Parent button 
*/
static ObjPtr DoParent(button)
ObjPtr button;
{

    char *s, dirName[MAXPATHLEN+1];
    ObjPtr theDir;
    WinInfoPtr fileWin;
    FuncTyp method;

    fileWin = (WinInfoPtr) GetVar(button, OWNERWINDOW);  /* get file window */
    theDir = GetVar((ObjPtr) fileWin, REPOBJ);	/* get directory */
    
    strncpy(dirName, GetString(GetVar(theDir, NAME)), MAXPATHLEN);
    dirName[MAXPATHLEN] = '\0'; /* ensure termination */
    s = strrchr(dirName, '/');
    if (s)
    {
	if (s > dirName)
	{
	    *s = '\0';
	}
	else
	{
	    *(s + 1) = '\0';
	}
	theDir = GetDirectory(dirName);
	SetVar((ObjPtr) fileWin, REPOBJ, theDir);
	DeferMessage(GetVar((ObjPtr) fileWin, TITLETEXT), UPDATE);
	method = GetMethod((ObjPtr) fileWin, SHOWFILES);
	if (method)
	{
	    (*method)(fileWin);
	}
    }
    DoTask(DeselectAll);
    return ObjTrue;
}

/**************************************************************** READAFILE
*   Reads a log or data file without depending on browser
*   Returns true if success, false otherwise
*/
#ifdef HAVE_PROTOTYPES
Bool ReadAFile(char *path, char *name, char *format)
#else
Bool ReadAFile(path, name, format)
char *path, *name, *format;
#endif
{
    ObjPtr retVal;
    ObjPtr stmt;

    if (!path)
    {
	path = GetCurDir();
    }

    stmt = NewFileRead(NewString(path), NewString(name), NewString(format));

    retVal = Eval(stmt);

    return IsTrue(retVal);
}

/**************************************************************** OPEN FILE
*	Reads a data or script file; returns ObjTrue if success
*	or ObjFalse if failure
*/
static ObjPtr OpenFile(theFile)
ObjPtr theFile;
{
    ObjPtr thePath, theName, theFormat;
    
    thePath = GetStringVar("OpenFile", theFile, PATH);
    theName = GetStringVar("OpenFile", theFile, NAME);
    theFormat = GetStringVar("OpenFile", theFile, FORMAT);
    if(thePath && theName && theFormat
      && ReadAFile(GetString(thePath), GetString(theName), GetString(theFormat)))
    {
	return ObjTrue;
    }
    else
    {
	return ObjFalse;
    }
}

/*************************************************************** OPEN FOLDER */
static ObjPtr OpenFolder(theFile)
ObjPtr theFile;
{
    char *path, *name, dirName[MAXPATHLEN+1];
    DIR *dirp;
    ObjPtr newDir;
    FuncTyp method;

    path = GetString(GetVar(theFile, PATH));
    name = GetString(GetVar(theFile, NAME));
    if (strlen(path) + 1 + strlen(name) > MAXPATHLEN)
    {
	/* Oh, no! The dang pathname is too long! */
	WarnUser(CW_PATHTOOLONGERROR);
	return ObjFalse;
    }
    strcpy(dirName, path);
    if (dirName[strlen(dirName) - 1] != '/')
    {
	strcat(dirName, "/");
    }
    strcat(dirName, name);
    dirName[MAXPATHLEN] = '\0';

    if(dirp = opendir(dirName)) /* good path? */
    {
	closedir(dirp);
	newDir = GetDirectory(dirName);
	SetVar((ObjPtr) selWinInfo, REPOBJ, newDir);
	DeferMessage(GetVar((ObjPtr) selWinInfo, TITLETEXT), UPDATE);
	method = GetMethod((ObjPtr) selWinInfo, SHOWFILES);
	if (method)
	{
	    (*method)(selWinInfo);
	}
	DoTask(DeselectAll);
	return ObjTrue;
    }
    else
    {
	/* Hey! There shouldn't be an error here! */
	WarnUser(CW_CHDIRERROR);
	return ObjFalse;
    }
}
