/*ScianMain.c
  Eric Pepke
  Main routines for SciAn
*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianArrays.h"
#include "ScianLists.h"
#include "ScianEvents.h"
#include "ScianWindows.h"
#include "ScianObjWindows.h"
#include "ScianObjFunctions.h"
#include "ScianGlobalFunctions.h"
#include "ScianWindowFunctions.h"
#include "ScianCurrentFunctions.h"
#include "ScianVisWindows.h"
#include "ScianDialogs.h"
#include "ScianDatasets.h"
#include "ScianColors.h"
#include "ScianControls.h"
#include "ScianDrawings.h"
#include "ScianIcons.h"
#include "ScianRecorders.h"
#include "ScianFiles.h"
#include "ScianFileSystem.h"
#include "ScianGarbageMan.h"
#include "ScianScripts.h"
#include "ScianTimers.h"
#include "ScianErrors.h"
#include "ScianIDs.h"
#include "ScianSpaces.h"
#include "ScianPreferences.h"
#include "ScianSciences.h"
#include "ScianSnap.h"
#include "ScianAxes.h"
#include "ScianTextFiles.h"
#include "ScianMenus.h"
#include "ScianFixedClasses.h"
#include "ScianArenas.h"
#include "ScianVisWindows.h"

Bool stuffInited = false;
Bool runningRemote = false;
real missingData = 1.1E37;
real plusInf = 1E37;
real minusInf = -1E37;
Bool scriptSelectP = false;
Bool abortScriptP = true;
Bool graphicsInited = false;

CaveSharedMemory *caveSharedMemory = 0;
void *caveArena = 0;

#define CAVEARENASIZE	10000

extern Bool runningScript;	/*True iff running script*/

#ifdef HAVE_PROTOTYPES
void InitStuff(void);
#else
void InitStuff();
#endif

struct
    {
	char *name;		/*Name of window system*/
	char *longName;		/*Long name of graphics system*/
    } windowSystemNames[N_WINDOW_SYSTEMS] =
    {
#ifdef WINDOWS4D
	{"gl",		"SGI Graphics Library windows"},
#else
	{NULL, NULL},
#endif
#ifdef WINDOWSCAVE
	{"cave",	"CAVE virtual environment"},
#else
	{NULL, NULL},
#endif
#ifdef WINDOWS4D
	{"arena",	"SGI Graphics Library 3-D arena"},
#else
	{NULL, NULL},
#endif
    };

#ifdef GRAPHICS
#ifdef CURSORS4D
/*Cursors*/
Cursor watchCursor = 
    {
	0x07e0,
	0x07e0,
	0x07e0,
	0x07e0,
	0x0810,
	0x1008,
	0x1008,
	0x11cc,
	0x110c,
	0x1108,
	0x1108,
	0x0810,
	0x07e0,
	0x07e0,
	0x07e0,
	0x07e0
    };
Cursor questionCursor = 
    {
	0x0300,
	0x0300,
	0x0000,
	0x0300,
	0x0300,
	0x0300,
	0x0380,
	0x0180,
	0x0070,
	0x0038,
	0x0018,
	0x1818,
	0x1818,
	0x1c38,
	0x0ff0,
	0x07e0
    };
Cursor iBeamCursor = 
    {
	0x3C78,
	0x0280,
	0x0100,
	0x0100,
	0x0380,
	0x0100,
	0x0100,
	0x0100,
	0x0100,
	0x0100,
	0x0100,
	0x0100,
	0x0100,
	0x0280,
	0x3C78
    };
Cursor iconCursor = 
    {
	0xFFFF,
	0xFFFF,
	0xFFFF,
	0xFFFF,
	0xFFFF,
	0xFFFF,
	0xFFFF,
	0xFFFF,
	0xFFFF,
	0xFFFF,
	0xFFFF,
	0xFFFF,
	0xFFFF,
	0xFFFF,
	0xFFFF,
	0xFFFF
    };
#endif
#endif


#ifdef PARANOID
/*Define tracking memory allocations*/

#define NMEMCHUNKS	257

static long memStats[NMEMCHUNKS][2];
#ifdef CACHEMEMORY
static void *memCache[NMEMCHUNKS];
#endif

void *Alloc(long n)
/*Allocates n bytes*/
{
    register void *p;
    if (n < NMEMCHUNKS)
    {
	++memStats[n][0];
#ifdef CACHEMEMORY
	if (memCache[n])
	{
	    p = memCache[n];
	    memCache[n] = *((void **) p);
	    *((long *) p) = n;
	    return (void *) (((long *) p) + 1);
	}
#endif
    }
    else
    {
	++memStats[NMEMCHUNKS - 1][0];
    }
    p = malloc(n + sizeof(long));
    *((long *) p) = n;
    return (void *) (((long *) p) + 1);
}

void *Realloc(void *p, long n)
/*Reallocates p to be n bytes*/
{
    p = (void *) (((long *) p) - 1);
    if (*((long *) p) < NMEMCHUNKS)
    {
	++memStats[*((long *) p)][1];
    }
    else
    {
	++memStats[NMEMCHUNKS - 1][1];
    }
    if (n < NMEMCHUNKS)
    {
	++memStats[n][0];
    }
    else
    {
	++memStats[NMEMCHUNKS - 1][0];
    }
    p = realloc(p, n + sizeof(long));
    *((long *) p) = n;
    return (void *) (((long *) p) + 1);
}

void Free(void *p)
/*Frees p*/
{
    register int n;
    p = (void *) (((long *) p) - 1);
    n = *((long *) p);
    if (n < NMEMCHUNKS)
    {
	++memStats[n][1];

#ifdef CACHEMEMORY
	*((void **) p) = memCache[n];
	memCache[n] = p;
#else
	free(p);
#endif
    }
    else
    {
	++memStats[NMEMCHUNKS - 1][1];
	free(p);
    }
}

void PrintMemStats(void)
/*Prints memory stats*/
{
#ifdef MALLOCH
    int k;
	struct mallinfo mi;

    printf("Size    Allocations  Deallocations     Current\n");
    for (k = 0; k < NMEMCHUNKS; ++k)
    {
	printf("%4d%c   %11d    %11d    %11d\n",
	    k, k == NMEMCHUNKS - 1 ? '+' : ' ', memStats[k][0], memStats[k][1],  memStats[k][0] - memStats[k][1]);
    }

	mi = mallinfo();
	fprintf(stderr,"arena=%d, ordblks=%d, smblks=%d, hblkhd=%d, hblks=%d\n",
			mi.arena, mi.ordblks, mi.smblks, mi.hblkhd, mi.hblks);
	fprintf(stderr,"usmblks=%d, fsmblks=%d, uordblks=%d, fordblks=%d, keepcost=%d\n",
		mi.usmblks, mi.fsmblks, mi.uordblks, mi.fordblks, mi.keepcost);
#endif
}
#endif

Bool graphicsInitDone = false;

#ifdef HAVE_PROTOTYPES
static void CAVEGraphicsInit(void)
#else
static void CAVEGraphicsInit()
#endif
/*Routine to set initialized graphics for cave*/
{
}

#ifdef HAVE_PROTOTYPES
void InitGraphics(void)
#else
void InitGraphics()
#endif
/*Initializes graphics*/
{
    if (!graphicsInited)
    {
	graphicsInited = true;

	switch(windowSystem)
	{
	    case WS_CAVE:
		/*Initialize cave system and initstuff*/
#ifdef WINDOWSCAVE
		{
		    caveArena = CAVEUserSharedMemory(CAVEARENASIZE);
		    if (caveArena)
		    {
			caveSharedMemory = amalloc(sizeof(CaveSharedMemory), caveArena);
	    
			caveSharedMemory -> runningP = true;
			caveSharedMemory -> confirmExitTime = 0;
			caveSharedMemory -> lastEvent = 0;
		    }
		    CAVEInit();
		    
		    CAVEInitApplication(InitStuff, 0);
		    CAVEDisplay(CAVEGraphicsInit, 0);
		}
#endif
		break;
	    case WS_ARENA:
		NewArenaWindow("Arena");
		MainWindow();
		break;
	    default:
		MainWindow();
		break;
	}
    }
}

/*Read states*/
#define RS_FILE		0	/*Read names as file names*/
#define RS_FORMAT	1	/*Read names as formats*/
#define RS_LOG		2	/*Read names as log files*/
#define RS_SCRIPT	3	/*Read names as script files*/
#define RS_DIRECTORY	4	/*Read names as directory files*/
#define RS_WINDOWSYSTEM	5	/*Read window system*/

static char **globalArgv;
static int globalArgc;

#ifdef HAVE_PROTOTYPES
void ParsePreInitArgs(void)
#else
void ParsePreInitArgs()
#endif
/*Parses the arguments in globalArgc and globalArgv, only looking at pre-init args*/
{
    int k, gs;
    char format[256];		/*Current file format*/
    ObjPtr corral;		/*Corral in the datasets window*/
    char *a;
    Bool scriptSet;
    int readState = RS_FILE;

    scriptSet = false;

    strcpy(format, "");

    for (k = 1; k < globalArgc; ++k)
    {
	/*Go through arguments, parsing them*/
	if (readState != RS_FILE)
	{
	    switch(readState)
	    {
		case RS_WINDOWSYSTEM:
		    /*Read in a graphics system*/
		    for (gs = 0; gs < N_WINDOW_SYSTEMS; ++gs)
		    {
			if (windowSystemNames[gs] . name &&
			    0 == strcmp2(globalArgv[k], windowSystemNames[gs] . name))
			{
			    windowSystem = gs;
			    if (IMMERSIVEP) oneVisWindow = true;
			    break;
			}
		    }
		    if (gs >= N_WINDOW_SYSTEMS)
		    {
			fprintf(stderr, "%s: Undefined window system '%s'\n", globalArgv[0], globalArgv[k]);
			fprintf(stderr, "Valid options are:\n");
			for (gs = 0; gs < N_WINDOW_SYSTEMS; ++gs)
			{
			    if (windowSystemNames[gs] . name)
			    {
				fprintf(stderr, "%-12s%s\n",
				    windowSystemNames[gs] . name, windowSystemNames[gs] . longName);
			    }
			}
			exit(-1);
		    }
		    break;
	    }
	    readState = RS_FILE;
	}
	else if (globalArgv[k][0] == '-')
	{
	    /*It's a switch*/
	    readState = RS_FILE;
	    
	    for (a = &(globalArgv[k][1]); *a; ++a)
	    {
		switch (*a)
		{
		    case 'w':
			/*Select a graphics system*/
			if (graphicsInited)
			{
			    fprintf(stderr, "%s: Window system selector -w must appear before graphics initialization.\n", globalArgv[0]);
			    exit(-1);
			}
			else
			{
			    readState = RS_WINDOWSYSTEM;
			}
			break;
		    default:
			break;
		}
	    }
	}
    }

    if (windowSystem < -1)
    {
	for (gs = 0; gs < N_WINDOW_SYSTEMS; ++gs)
	{
	    if (windowSystemNames[gs] . name)
	    {
		windowSystem = gs;
		break;
	    }
	}
	if (gs >= N_WINDOW_SYSTEMS)
	{
	    fprintf(stderr, "%s: No window systems defined\n", globalArgv[0]);
	    exit(-1);
	}
    }
}

#ifdef HAVE_PROTOTYPES
void ParseArgs(void)
#else
void ParseArgs()
#endif
/*Parses the arguments in globalArgc and globalArgv*/
{
    int k, gs;
    char format[256];		/*Current file format*/
    ObjPtr corral;		/*Corral in the datasets window*/
    char *a;
    Bool scriptSet;
    int readState = RS_FILE;

    scriptSet = false;

    strcpy(format, "");

    for (k = 1; k < globalArgc; ++k)
    {
	/*Go through arguments, parsing them*/
	if (readState != RS_FILE)
	{
	    /*Read this argument as a whatever*/
	    switch(readState)
	    {
		case RS_FORMAT:
		    strncpy(format, globalArgv[k], 255);
		    format[255] = 0;
		    break;
		case RS_SCRIPT:
		    if (scriptSet)
		    {
			fprintf(stderr, "%s: Only one script can be given on the command line.\n", globalArgv[0]);
			exit(-1);
		    }
		    scriptSet = true;
		    BeginScript(globalArgv[k]);
		    break;
		case RS_LOG:
		    if (strcmp(globalArgv[k], "-") == 0)
		    {
			OpenLogFile((char *) 0);
		    }
		    else
		    {
			OpenLogFile(globalArgv[k]);
		    }
		    break;
		case RS_DIRECTORY:
		    {
			char t[MAXPATHLEN + 1];
			if (chdir(globalArgv[k]) == 0) /* check if valid directory */
			{
			    getwd(t);	/* get expanded name back */
			    GetFileWindow(t);
			    chdir(dirSave);	/* return to original directory */
			}
			else
			{
			    fprintf(stderr, "%s: '%s' is an invalid directory.\n", globalArgv[0], globalArgv[k]);
			    exit(-1);
			}
		    }
		    break;
		case RS_WINDOWSYSTEM:
		    /*Ignore a graphics system*/
		    break;
	    }
	    readState = RS_FILE;
	}
	else if (globalArgv[k][0] == '-')
	{
	    /*It's a switch*/
	    readState = RS_FILE;
	    
	    for (a = &(globalArgv[k][1]); *a; ++a)
	    {
		switch (*a)
		{
		    case 'f':
			/*New file format*/
			readState = RS_FORMAT;
			break;
		    case '1':
			/*One observer, clock, whatever*/
			++a;
			switch (*a)
			{
			    case 'c':
			    case 'C':
				oneClock = true;
				break;
			    case 'o':
			    case 'O':
				oneObserver = true;
				break;
			    case 'l':
			    case 'L':
				oneLights = true;
				break;
			    case 'r':
			    case 'R':
				oneRenderer = true;
				break;
			    case 'p':
			    case 'P':
				onePalette = true;
				break;
			    case 'v':
			    case 'V':
				oneVisWindow = true;
				break;
			    default:
				fprintf(stderr, "%s: Unrecognized object type for -1 flag\n", globalArgv[0]);
				exit(-1);
			}
			break;
		    case 's':
			/*Script to run*/
			readState = RS_SCRIPT;
			break;
		    case 'S':
			/*Select in scripts*/
			scriptSelectP = true;
			break;
		    case 'C':
			/*Continue in scripts*/
			abortScriptP = false;
			break;
		    case 'l':
			/*Open log file*/
			readState = RS_LOG;
			break;
		    case 'v':
			/*Enable recording*/
    #ifdef GRAPHICS
    #ifdef IRIS
			blankscreen(FALSE);
			blanktime(0);
    #endif
    #endif
			recordEnabled = true;
			break;
		    case 'd':
			/*Demo mode*/
			SetDemoFonts();
			break;
		    case 'o':
			/*Open a directory*/
			readState = RS_DIRECTORY;
			break;
		    case 'w':
			/*Ignore a graphics system*/
			readState = RS_WINDOWSYSTEM;
			break;
		    default:
			/*Undefined switch*/
			fprintf(stderr, "%s: Switch %s is undefined.\n", globalArgv[0], globalArgv[k]);
			exit(-1);
		}
	    }
	}
	else if (globalArgv[k][0] == '+')
	{
	    if (0 == strcmp(globalArgv[k], "+mdoc"))
	    {
		EmitMenuDoc();
	    }
	    if (0 == strcmp(globalArgv[k], "+hdoc"))
	    {
		EmitHelpDoc();
	    }
	}
	else
	{
	    WinInfoPtr datasetsWindow;
	    ObjPtr reader;

	    /*It must be a file name*/

	    datasetsWindow = DatasetsWindow();

	    SelWindow(datasetsWindow);
	    IdleAllWindows();
	    LongOperation();

	    if (format && strlen(format))
	    {
		ReadAFile((char *) 0, globalArgv[k], format);
	    }
	    else
	    {
		int i;
		ObjPtr var;

		for (i = strlen(globalArgv[k]);
		     i >= 0 && (globalArgv[k][i] != '.');
		     --i);

		reader = FindExtensionReader(&(globalArgv[k][i + 1]));
		if (reader)
		{
		    var = GetStringVar("ParseArgs", reader, NAME);
		    if (var)
		    {
			ReadAFile((char *) 0, globalArgv[k], GetString(var));
		    }
		}
		else
		{
		    WarnUser(CW_CLFORMATERROR);
		}
	    }
	}
    }
}

Bool handlingSignals = true;

#ifdef SIGNALS
#ifdef HAVE_PROTOTYPES
void SignalHandler(int sig, int code, struct sigcontext *context)
#else
void SignalHandler(sig, code, context)
int sig;
int code;
struct sigcontext *context;
#endif
/*Handles signals*/
{
    if (0 == handlingSignals)
    {
	/*Just expire*/
	return;
    }
    switch(sig)
    {
#ifdef SIGALRM
	case SIGALRM:
	    /*Get some new events*/
	    NewEvents();
	    {
		struct itimerval value, ovalue;
	
		value . it_interval . tv_sec = 0;
		value . it_interval . tv_usec = 0;
		value . it_value . tv_sec = 0;
		value . it_value . tv_usec = 100000L;
		setitimer(ITIMER_REAL, &value, &ovalue);
	    }
	    signal(SIGALRM, SignalHandler);
	    break;
#endif
#ifdef SIGDANGER
	case SIGDANGER:
	    signal(SIGDANGER, SignalHandler);
	    break;
#endif
    }
}
#endif

TestTextFile()
{
    TextFilePtr file;

    file = OpenTextFile("testo.txt", TF_READ + TF_CONTINUATION + TF_HASHCOMMENTS);
    if (file)
    {
	char *curLine;
	while (curLine = GetNextLine(file))
	{
	    char *chickenWhere;
	    chickenWhere = strstr(curLine, "chicken");
	    if (chickenWhere)
	    {
		CurLineError(file, "'chicken' may not appear", chickenWhere,
		    chickenWhere + strlen("chicken"));
	    }
	}
	CloseTextFile(file);
    }
    else
    {
	perror("testo.txt");
    }
}

main(argc, argv)
int argc;
char *argv[];
{
    int k;
#ifdef MALLOCH
#ifdef M_MXFAST
    mallopt(M_MXFAST, 1024);
#endif
#endif

    globalArgv = argv;
    globalArgc = argc;

#ifdef PARANOID
    for (k = 0; k < NMEMCHUNKS; ++k)
    {
	memStats[k][0] = 0;
	memStats[k][1] = 0;
#ifdef CACHEMEMORY
	memCache[k] = NULL;
#endif
    }
#endif

#ifdef GRAPHICS
#ifdef IRIS
    foreground();
#endif
#endif

#ifdef SIGNALS
    /*Set up signals*/
#ifdef SIGDANGER
    signal(SIGDANGER, SignalHandler);
#endif
#endif

    ParsePreInitArgs();

#ifdef WINDOWSCAVE
    if (CAVEP)
    {
	signal(SIGINT, CAVEExit);
    }
#endif

#ifndef GETGDESCAFTER
    GetConfiguration();
#endif
    InitNames();
    InitObjects();
    InitFixedClasses();
    InitTextFiles();
    InitArrays();
    InitLists();
    InitSound();
    InitControls();
    InitDatabase();
    InitSciences();
    InitAxes();
    InitIcons();
    InitDatasets();
    InitFilters();
    InitParser();
    InitCalculator();
    InitFiles();
    InitScripts();

    InitGraphics();

    if (!CAVEP)
    {
	ParseArgs();
    }

    MainLoop();

    CloseAllLogFiles();
#ifdef GRAPHICS
#ifdef IRIS
    if (recordEnabled)
    {
	blanktime(100000);
    }
#endif
#endif
    handlingSignals = false;
    DisposeAllWindows();
    KillProjects();
    KillSnapshots();
    KillEvents();
    KillWindowFunctions();
    KillRepobjFunctions();
    KillGlobalFunctions();
    KillRecorders();
    KillPreferences();
    KillColors();
    KillDialogs();
    KillObjFunctions();
    KillCurrentFunctions();
    KillMenus();
    KillVisWindows();
    KillObjWindows();
    KillArenas();
    KillWindows();
    KillPictures();
    KillVisObjects();
    KillTimers();
    KillSpaces();
    KillFilters();
    KillDatasets();
    KillGeometry();
    KillAnimation();
    KillDrawings();
#if 0
    KillFonts();
#endif
    KillIcons();
    KillSockets();
    KillDraw();
    KillAxes();
    KillSciences();
    KillDatabase();
    KillSound();
    KillControls();
    KillCalculator();
    KillParser();
    KillScripts();
    KillFiles();
    KillLists();
    KillArrays();
    TrashDay();
    KillTextFiles();
    KillFixedClasses();
    KillObjects();
    KillNames();
    switch (windowSystem)
    {
	case WS_CAVE:
	    /*Exit cave system*/
#ifdef WINDOWSCAVE
	    {
		if (caveSharedMemory)
		{
		    afree(caveSharedMemory, caveArena);
		}
		CAVEExit();
	    }
#endif
	    break;
    }
    exit(0);
}

#ifdef HAVE_PROTOTYPES
static void DefineFunctionKeys(void)
#else
static void DefineFunctionKeys()
#endif
/*Defines a bunch of function keys*/
{
#ifndef RELEASE
    DefineFunctionKey(FK_1, 0, GF_UNDO);
#endif
    DefineFunctionKey(FK_2, 0, CF_CUT);
    DefineFunctionKey(FK_3, 0, CF_COPY);
    DefineFunctionKey(FK_4, 0, CF_PASTE);
    DefineFunctionKey(FK_5, 0, OF_SHOW_CONTROLS);
    DefineFunctionKey(FK_6, 0, WF_HIDEPANEL);
    DefineFunctionKey(FK_7, 0, OF_PUSHTOBOTTOM);
    DefineFunctionKey(FK_8, 0, OF_BRINGTOTOP);
    DefineFunctionKey(FK_9, 0, OF_PICK_UP);
#ifdef IRISNTSC
    DefineFunctionKey(FK_10, 0, GF_TOGGLENTSC);
    DefineFunctionKey(FK_11, 0, GF_TOGGLESTEREO);
#endif
    DefineFunctionKey(FK_12, 0, GF_CONTEXTHELP);
    DefineFunctionKey(FK_PRINT_SCREEN, 0, WF_SAVEWINDOW);
    DefineFunctionKey(FK_PRINT_SCREEN, F_SHIFTDOWN, WF_SAVEFWINDOW);
    DefineFunctionKey(FK_PRINT_SCREEN, F_OPTIONDOWN, WF_SAVESCREEN);
    DefineFunctionKey('q', F_OPTIONDOWN, GF_QUIT);
    DefineFunctionKey('p', F_OPTIONDOWN, GF_NEWPROJECT);
}

#ifdef HAVE_PROTOTYPES
void InitStuff(void)
#else
void InitStuff()
#endif
/*Initializes stuff.  Has to be called after one window has been opened.*/
{
    if (!stuffInited)
    {

#ifdef GRAPHICS
#ifdef GL4D
#ifdef IRIS
	defaultMonitor = getmonitor();
#else
#ifdef HZ60
	defaultMonitor = HZ60;
#else
	defaultMonitor = 1;
#endif
#endif
#endif
#endif
#ifdef GETGDESCAFTER
	GetConfiguration();
#endif
	InitDraw();
	InitSockets();
	InitColors();
	InitDrawings();
	InitAnimation();
	InitSpaces();
	InitTimers();
	InitGeometry();
	InitPictures();
	InitFonts();
	DefineIconFonts();
	InitWindows();
	InitObjWindows();
	InitArenas();
	InitMenus();
	InitCurrentFunctions();
	InitObjFunctions();
	InitVisWindows();
	InitVisObjects();
	InitDialogs();
	InitPreferences();
	InitRecorders();
	InitGlobalFunctions();
	InitRepobjFunctions();
	InitWindowFunctions();
	InitEvents();
	InitSnapshots();
	InitProjects();

	DefineFunctionKeys();

	/*Set up cursors*/
#ifdef GRAPHICS
#ifdef CURSORS4D
#ifdef GL4D
	curstype(C16X1);
#endif
	defcursor(WATCHCURSOR, watchCursor);
	curorigin(WATCHCURSOR, 8, 8);
	defcursor(QUESTIONCURSOR, questionCursor);
	curorigin(QUESTIONCURSOR, 8, 8);
	defcursor(IBEAMCURSOR, iBeamCursor);
	curorigin(IBEAMCURSOR, 8, 4);
	defcursor(ICONCURSOR, iconCursor);
	curorigin(ICONCURSOR, 8, 8);
#endif
#endif
	stuffInited = true;
#ifdef SIGNALS
#ifdef SIGALRM
	signal(SIGALRM, SignalHandler);
	{
	    struct itimerval value, ovalue;
    
	    value . it_interval . tv_sec = 0;
	    value . it_interval . tv_usec = 0;
	    value . it_value . tv_sec = 0;
	    value . it_value . tv_usec = 100000L;
    
	    setitimer(ITIMER_REAL, &value, &ovalue);
	}
#endif
#endif

	if (CAVEP)
	{
	    NewArenaWindow("Arena");
	    MainWindow();
	    rgbp = true;
	    curDisplayMode = DMDOUBLE;

	    ParseArgs();
	}
    }
}

