/*ScianParser.c
  Eric Pepke
  March 17, 1994

  Parser for text expressions in SciAn

  This is basically a rather boring top-down parser, which used calls to the 
  calculator to make expressions.  There is one instance of semantic-driven syntax,
  which is used to reduce memory needs when creating arrays.

  <array> ::= [ {<expression>} ]
*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianArrays.h"
#include "ScianIDs.h"
#include "ScianEvents.h"
#include "ScianRecorders.h"
#include "ScianColors.h"
#include "ScianWindows.h"
#include "ScianObjWindows.h"
#include "ScianDialogs.h"
#include "ScianVisWindows.h"
#include "ScianSliders.h"
#include "ScianTimers.h"
#include "ScianMethods.h"
#include "ScianLists.h"
#include "ScianErrors.h"
#include "ScianGarbageMan.h"
#include "ScianSpaces.h"
#include "ScianIcons.h"
#include "ScianSockets.h"
#include "ScianObjFunctions.h"
#include "ScianGlobalFunctions.h"
#include "ScianCurrentFunctions.h"
#include "ScianWindowFunctions.h"
#include "ScianRepobjFunctions.h"
#include "ScianHelp.h"
#include "ScianStyle.h"
#include "ScianNames.h"
#include "ScianTextFiles.h"
#include "ScianDatabase.h"
#include "ScianSymbols.h"
#include "ScianControls.h"
#include "ScianScripts.h"
#include "ScianFixedClasses.h"
#include "ScianParser.h"
#include "ScianCalculator.h"

char *begToken, *endToken;	/*Beginning and ending of current token*/

TextFilePtr curFile = 0;	/*Current file used to parse*/
int tc = TC_NULL;		/*Current type cast.  Used for semantic direction of array parsing.*/

#ifdef HAVE_PROTOTYPES
void ParseFromFile(TextFilePtr file)
#else
void ParseFromFile(file)
TextFilePtr file;
#endif
/*Parses from a file*/
{
    curFile = file;
}

#ifdef HAVE_PROTOTYPES
void ParseError(char *errMes)
#else
void ParseError(errMes)
char *errMes;
#endif
/*Generates an error message*/
{
    if (curFile)
    {
	CurLineError(curFile, errMes, begToken, endToken <= endToken ? begToken + 1 : endToken);
    }
}

#define ARRAYREADCHUNK	200

#ifdef HAVE_PROTOTYPES
char *ParseArray(char *s, ObjPtr *o, int expType)
#else
char *ParseArray(s, o, type)
char *s;
ObjPtr *o;
char expType;
#endif
/*
    Parses an array at s
    Returns
	s if the object was syntactically bad or null
	the first character after the object if it was OK.
	Puts in o the object if it was semantically valid
        Leavs o alone if it was not
    Generates a script error if syntactically or semantically invalid.
    expType is the expected type of the array, if TC_NULL, assume real

    <array> ::= [ {<expression>} ]
*/
{
    char *r;
    char arg[200];
    int rank;
    int rankTest;
    long k;
    long *dimensions = NULL;
    long *dimensionTest = NULL;

    short *shortData = NULL;
    real *realData = NULL;
    void **pointerData = NULL;
    char *byteData = NULL;
    ObjPtr *objectData = NULL;
    long nAlloc;

    r = s;

    SKIPBLANKS(r);
    if (*r != '[')
    {
	ParseError("This is not a valid array");
	return s;
    }

    ++r;

    /*Make room for rank of 1*/
    rank = 1;
    dimensions = (long *) Alloc(sizeof(long) * rank);
    dimensionTest = (long *) Alloc(sizeof(long) * rank);
    dimensions[0] = -1;
    dimensionTest[0] = 0;

    /*Make room for the first group of data*/
    nAlloc = ARRAYREADCHUNK;
    if (expType == TC_SHORTARRAY)
    {
	shortData = (short *) Alloc(sizeof(short) * nAlloc);
    }
    else if (expType == TC_REALARRAY)
    {
	realData = (real *) Alloc(sizeof(real) * nAlloc);
    }
    else if (expType == TC_BYTEARRAY)
    {
	byteData = (char *) Alloc(sizeof(char) * nAlloc);
    }
    else if (expType == TC_POINTERARRAY)
    {
	pointerData = (void **) Alloc(sizeof(void *) * nAlloc);
    }
    else if (expType == TC_OBJECTARRAY)
    {
	objectData = (ObjPtr *) Alloc(sizeof(ObjPtr) * nAlloc);
    }
    else
    {
	realData = (real *) Alloc(sizeof(real) * nAlloc);
    }
    k = 0;
    rankTest = 0;

    while (rankTest >= 0)
    {
	SKIPBLANKS(r);

	if (*r == '[')
	{
	    /*One more rank*/
	    ++dimensionTest[rankTest];

	    ++rankTest;

	    if (rankTest >= rank)
	    {
		/*Have to crank up the rank*/
		rank = rankTest + 1;

		dimensions = (long *) Realloc(dimensions, rank * sizeof(long));
		dimensionTest = (long *) Realloc(dimensionTest, rank * sizeof(long));
		dimensions[rank - 1] = -1;	/*Start with unknown rank*/
	    }
	    dimensionTest[rankTest] = 0;	/*Start with 0 dimension test*/

	    ++r;
	}
	else if (*r == ']')
	{
	    /*Come to the end of this row.  Check the dimension*/
	    if (dimensions[rankTest] < 0)
	    {
		dimensions[rankTest] = dimensionTest[rankTest];
	    }
	    else if (dimensions[rankTest] > dimensionTest[rankTest])
	    {
		begToken = r;
		endToken = r + 1;
		ParseError("Too few elements");
	    }
	    else if (dimensions[rankTest] < dimensionTest[rankTest])
	    {
		begToken = r;
		endToken = r + 1;
		ParseError("Too many elements");
	    }

	    --rankTest;

	    ++r;
	}
	else if (expType == TC_SHORTARRAY)
	{
	    /*Expect a short value here*/
	    int val;

	    SHIFTNUM(arg, r);
	    if (ParseInteger(&val, arg))
	    {
		if (k >= nAlloc)
		{
		    nAlloc += ARRAYREADCHUNK;
		    shortData = (short *) Realloc(shortData, sizeof(short) * nAlloc);
		}
		shortData[k] = val;
		++k;     

		++dimensionTest[rankTest];
	    }
	    else
	    {
		ParseError("A short integer was expected here");
	    }
	}
	else if (expType == TC_BYTEARRAY)
	{
	    /*Expect a byte value here*/
	    int val;

	    SHIFTNUM(arg, r);
	    if (ParseInteger(&val, arg))
	    {
		if (k >= nAlloc)
		{
		    nAlloc += ARRAYREADCHUNK;
		    byteData = (char *) Realloc(byteData, sizeof(char) * nAlloc);
		}
		byteData[k] = val;
		++k;     

		++dimensionTest[rankTest];
	    }
	    else
	    {
		ParseError("A byte integer was expected here");
	    }
	}
	else if (expType == TC_POINTERARRAY)
	{
	    /*Expect a pointer value here*/
	    void *val;

	    SHIFTKW(arg, r);
	    if (1 == sscanf(arg, "%lx", &val))
	    {
		if (k >= nAlloc)
		{
		    nAlloc += ARRAYREADCHUNK;
		    pointerData = (void **) Realloc(pointerData, sizeof(char) * nAlloc);
		}
		pointerData[k] = val;
		++k;     

		++dimensionTest[rankTest];
	    }
	    else
	    {
		ParseError("A pointer was expected here");
	    }
	}
	else if (expType == TC_OBJECTARRAY)
	{
	    /*Expect a byte value here*/
	    ObjPtr val = ERROBJ;

	    r = EvalExpressionArg(r, &val);
	    if (val == ERROBJ) val = NULLOBJ;

	    if (k >= nAlloc)
	    {
		nAlloc += ARRAYREADCHUNK;
		objectData = (ObjPtr *) Realloc(objectData, sizeof(char) * nAlloc);
	    }
	    objectData[k] = val;
	    ++k;     

	    ++dimensionTest[rankTest];
	}
	else
	{
	    /*Expect a real value here*/
	    real val;
	    char *t;

	    t = ParseReal(&val, r);
	    if (t)
	    {
		r = t;
		if (k >= nAlloc)
		{
		    nAlloc += ARRAYREADCHUNK;
		    realData = (real *) Realloc(realData, sizeof(real) * nAlloc);
		}
		realData[k] = val;
		++k;     

		++dimensionTest[rankTest];
	    }
	    else
	    {
		begToken = r;
		endToken = r;
		ParseError("A real number was expected here");
	    }
	}
    }

    if (expType == TC_SHORTARRAY)
    {
	short *elements;
	register long l;

	*o = NewArray(AT_SHORT, rank, dimensions);
	elements = ELEMENTS(*o);

	for (l = 0; l < k; ++l)
	{
	    elements[l] = shortData[l];
	}
    }
    else if (expType == TC_BYTEARRAY)
    {
	char *elements;
	register long l;

	*o = NewArray(AT_BYTE, rank, dimensions);
	elements = ELEMENTS(*o);

	for (l = 0; l < k; ++l)
	{
	    elements[l] = byteData[l];
	}
    }
    else if (expType == TC_POINTERARRAY)
    {
	void **elements;
	register long l;

	*o = NewArray(AT_POINTER, rank, dimensions);
	elements = ELEMENTS(*o);

	for (l = 0; l < k; ++l)
	{
	    elements[l] = pointerData[l];
	}
    }
    else if (expType == TC_OBJECTARRAY)
    {
	ObjPtr *elements;
	register long l;

	*o = NewArray(AT_OBJECT, rank, dimensions);
	elements = ELEMENTS(*o);

	for (l = 0; l < k; ++l)
	{
	    elements[l] = objectData[l];
	}
    }
    else
    {
	*o = NewArray(AT_REAL, rank, dimensions);
	CArray2Array(*o, realData);
    }

    SAFEFREE(dimensions);
    SAFEFREE(dimensionTest);
    SAFEFREE(shortData);
    SAFEFREE(realData);
    SAFEFREE(byteData);
    SAFEFREE(pointerData);
    SAFEFREE(objectData);

    return r;
}

#ifdef HAVE_PROTOTYPES
char *ParseObject(char *s, ObjPtr *o)
#else
char *ParseObject(s, o)
char *s;
ObjPtr *o;
#endif
/*
    Parses an object at s
    Returns
	s if the object was syntactically or semantically bad or null
	the first character after the object if it was OK.
	Puts in o the object if it was syntactically valid
        Leavs o alone if it was not
    Generates a parse error if syntactically invalid.

    <object> ::= <number> | <array> | <string> | 
	<compound name> {<variable selector>} | <class name>  {<variable selector>} |  
	<simple name>  {<variable selector>} | <script variable>  {<variable selector>} |
	<function call> | <special function>
    <simple name> ::= <identifier>
    <compound name> ::= <window name> : <simple name>
    <window name> ::= <identifier>
    <class name> ::= <class> @ <simple name>
    <class> ::= <identifier>
    <script variable> ::= $ <identifier> {<variable selector>}
    <variable selector> ::= . <variable name>
    <variable name> ::= <identifier>
    <function call> ::= <simple name> ( [<expression>] {, <expression>}])
*/
{
    char arg[256];
    char *r;

    r = s;

    /*Get to the first character*/
    SKIPBLANKS(r);

    begToken = r;

    /*Based on one-char lookahead, parse the argument*/
    if (*r == '-' || *r == '+' || *r == '.' || isdigit(*r) || *r == '#')
    {
	real val;
	char *t;

	/*It's a number*/
	t = ParseReal(&val, r);
	if (t)
	{
	    r = t;
	    *o = NewReal((real) val);
	}
	else
	{
	    endToken = r;
	    ParseError("This number has bad syntax");
	    return s;
	}
    }
    else if (*r == '$')
    {
	/*It's a script variable*/
	++r;
	SHIFTKW(arg, r);
	if (strlen(arg))
	{
	    ScriptVarPtr varPtr;
	    varPtr = FindReadScriptVar(arg);
	    if (!varPtr)
	    {
		ScriptError("That variable is not defined");
		endToken = r;
		return s;
	    }
	    *o = NewScriptVariable(varPtr);
	}
	else
	{
	    endToken = r;
	    ScriptError("The variable name is missing");
	}
    }
    else if (isalpha(*r) || (*r == '\\'))
    {
	/*It's an object name or NULLOBJ*/
	SHIFTKW(arg, r);

	if (0 == strcmp2(arg, "NULLOBJ"))
	{
	    *o = NULLOBJ;
	}
	else
	{
	    ObjPtr list = NULLOBJ;

	    if (*r == ':')
	    {
		char arg2[256];
		WinInfoPtr window;

		/*Compound name, first part window*/
		++r;
		SHIFTKW(arg2, r);
		if (strlen(arg2))
		{
		    window = GetWinFromTitle(arg);
		    if (window)
		    {
			*o = NewWindowVariable(window, arg2);
		    }
		    else
		    {
			endToken = r;
			ParseError("Cannot find a unique window by that name.");
			return s;
		    }
		}
		else
		{
		    endToken = r;
		    ParseError("An object name is expected after the window name.");
		    return s;
		}
	    }
	    else if (*r == '@')
	    {
		/*Class name*/
		int k;
		ObjPtr keyList;
		char arg2[256];
		++r;
		SHIFTKW(arg2, r);
		if (!strlen(arg2))
		{
		    endToken = r;
		    ParseError("An object name is expected after the object class.");
		    return r;
		}

		k = ClassNameToID(arg);

		if (k < 0)
		{
		    sprintf(arg2, "Class %s is not recognized", arg);
		    endToken = r;
		    ParseError(arg2);
		    return r;
		}

		*o = NewClassVariable(k, arg2);
	    }
	    else
	    {
		SKIPBLANKS(r);

		if (*r == '(')
		{
		    ObjPtr argsList;
		    ObjPtr oneArg;
		    int nArgs;		/*# of args*/
		    ObjPtr var;

		    ++r;
		    argsList = NewList();
		    nArgs = 0;

		    SKIPBLANKS(r);

		    if (*r != ')')
		    {
			oneArg = ERROBJ;
			r = ParseExpression(r, &oneArg);
			if (oneArg == ERROBJ)
			{
			    return r;
			}
			++nArgs;
			PostfixList(argsList, oneArg);

			SKIPBLANKS(r);
	
			while (*r == ',')
			{
			    ++r;
			    oneArg = ERROBJ;
			    r = ParseExpression(r, &oneArg);
			    if (oneArg == ERROBJ)
			    {
				return r;
			    }
			    ++nArgs;
			    PostfixList(argsList, oneArg);
			    SKIPBLANKS(r);
			}
		    }

		    if (*r != ')')
		    {
			ParseError("Close parenthesis ')' expected");
			return r;
		    }

		    var = NewCalcFunction(arg, argsList);
		    if (var) *o = var;

		    begToken = r + 1;
		    endToken = r + 1;
		    return r + 1;
		}
		else
		{
		    if (scriptWindow && IsValidWindow(scriptWindow))
		    {
			*o = NewWindowVariable(scriptWindow, arg);
		    }
		    else
		    {
			endToken = r;
			ParseError("There is no currently selected window in which to find the object.");
			return s;
		    }
		}
	    }

	    /*Now create variable references, if there are any*/

	    SKIPBLANKS(r);

	    while (*r == '.')
	    {
		NameTyp internalID;

		++r;

		SHIFTKW(arg, r);
		internalID = GetInternalID(arg);
		
		if (*o)
		{
		    *o = NewVarReference(*o, internalID); 
		}
	    }
	}
    }
    else if (*r == '[')
    {
	return ParseArray(r, o, tc);
    }
    else if (*r == '"')
    {
	int k;
	/*It's a string*/
	begToken = r;
	++r;
	k = 0;
	while (*r && *r != '"')
	{
	    if (*r == '\\')
	    {
		++r;
		if (isdigit(*r))
		{
		    int nDigits, value;
		    value = 0;
		    nDigits = 0;
		    do
		    {
			value *= 8;
			value += *r - '0';
			++nDigits;
			++r;
		    } while (isdigit(*r) && nDigits < 3);
		    tempStr[k++] = value;
		}
		else 
		switch (*r)
		{
		    case 'n':
		    case 'N':
			tempStr[k++] = '\n';
			++r;
			break;
		    case 't':
		    case 'T':
			tempStr[k++] = '\t';
			++r;
			break;
		    case 'r':
		    case 'R':
			tempStr[k++] = '\r';
			++r;
			break;
		    default: tempStr[k++] = *r++;
			
		}
	    }
	    else
	    {
		tempStr[k++] = *r++;
	    }
	}
	if (*r == '"') ++r;
	endToken = r;
	tempStr[k] = 0;
	*o = NewString(tempStr);
    }
    else
    {
	endToken = r;
	ParseError("This is an unrecognized character");
	return s;
    }

    return r;
}  

#ifdef HAVE_PROTOTYPES
char *ParseExpression(char *s, ObjPtr *o)
#else
char *ParseExpression(s, o)
char *s;
ObjPtr *o;
#endif
/*
    Parses an expression

    <expression> ::= <type cast> <expression> | 
	<object>
    <type cast> ::= (<type name>)
*/
{
    char *r;

    r = s;

    /*Get to the first character*/
    SKIPBLANKS(r);

    if (*r == '(')
    {
	char msg[200];
	char arg[256];
	char *oldR;
	int newTC = TC_NULL;

	/*Maybe it's a type cast*/

	oldR = r;

	++r;
	SHIFTKW(arg, r);
	SKIPBLANKS(r);
	if (*r != ')')
	{
	    /*Must not be a type cast*/
	    r = oldR;
	}
	else
	{
	    ++r;

	    if (0 == strcmp2(arg, "int"))
	    {
		newTC = TC_INTEGER;
	    }
	    else if (0 == strcmp2(arg, "shortarray"))
	    {
		newTC = TC_SHORTARRAY;
	    }
	    else if (0 == strcmp2(arg, "bytearray"))
	    {
		newTC = TC_BYTEARRAY;
	    }
	    else if (0 == strcmp2(arg, "pointerarray"))
	    {
		newTC = TC_POINTERARRAY;
	    }
	    else if (0 == strcmp2(arg, "objectarray"))
	    {
		newTC = TC_OBJECTARRAY;
	    }
	    else
	    {
		/*Must not be a type cast*/
		r = oldR;
	    }

	    if (newTC != TC_NULL)
	    {
		/*It is a type cast!*/
		ObjPtr exp;
		int oldTC;

		exp = ERROBJ;

		oldTC = tc;
		tc = newTC;
		r = ParseExpression(r, &exp);
		tc = oldTC;

		if (exp == ERROBJ)
		{
		    /*No good.  Just return*/
		    return s;
		}
		else
		{
		    /*Make a new type cast*/
		    *o = NewTypeCast(exp, newTC);
		    return r;
		}
	    }
	}
    }

    /*Just parse an object*/
    return ParseObject(s, o);
}

#ifdef HAVE_PROTOTYPES
void InitParser(void)
#else
void InitParser()
#endif
/*Initializes the parser*/
{
}


#ifdef HAVE_PROTOTYPES
void KillParser(void)
#else
void KillParser()
#endif
/*Kills the parser*/
{
}
