/*ScianPick.c
  Picking system for spaces in SciAn
  Eric Pepke
  May 30, 1991
*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianLists.h"
#include "ScianWindows.h"
#include "ScianSpaces.h"
#include "ScianDraw.h"
#include "ScianIDs.h"
#include "ScianErrors.h"
#include "ScianPick.h"
#include "ScianPictures.h"

#define MAXNPICKS		400	/*Maximum # of pick tuples*/

#ifdef GRAPHICS
static Matrix tempMatrix, pickMatrix;
#endif

short pickBuffer[MAXNPICKS * 4 + 1];	/*Pick buffer*/

void StartPick()
/*Starts picking mode*/
{
#ifdef GRAPHICS
    getmatrix(tempMatrix);
    pick(pickBuffer, MAXNPICKS * 4);
    initnames();

    /*Load NULLOBJ*/
    loadname(0);
    pushname(0);

    /*Load null vertex*/
    pushname(0);
    pushname(0);

    MATCOPY(pickMatrix, Identity);
#endif
}

#ifdef HAVE_PROTOTYPES
void PickObject(ObjPtr object, int resolveType)
#else
void PickObject(object, resolveType)
ObjPtr object;
int resolveType;
#endif
/*Declares that object is about to be drawn for picking*/
{
#ifdef GRAPHICS

    /*Get rid of old vertex*/
    popname();
    popname();

    /*Get rid of old object*/
    popname();

    /*Push this object*/
    loadname((((long) object) >> 16) & 0xFFFF);
    pushname(((long) object) & 0xFFFF);

    /*Load null vertex*/
    pushname(0);
    pushname(0);
#endif
}

#ifdef HAVE_PROTOTYPES
void PickVertex(VertexPtr vertex)
#else
void PickVertex(vertex)
VertexPtr vertex;
#endif
/*Identifies a vertex for picking*/
{
    /*Get rid of old vertex*/
    popname();
    popname();

    /*Load this vertex*/
    pushname((((long) vertex) >> 16) & 0xFFFF);
    pushname(((long) vertex) & 0xFFFF);
}

#ifdef HAVE_PROTOTYPES
void EndPickObject(void)
#else
void EndPickObject()
#endif
/*Ends the picking for object*/
{
    /*Get rid of old vertex*/
    popname();
    popname();

    /*Get rid of old object*/
    popname();

    /*Load NULLOBJ*/
    loadname(0);
    pushname(0);

    /*Load null vertex*/
    pushname(0);
    pushname(0);
}

#ifdef HAVE_PROTOTYPES
ObjPtr StopPick(VertexPtr *bestVertex)
#else
ObjPtr StopPick(bestVertex)
VertexPtr *bestVertex;
#endif
/*Stops picking mode and returns a list of objects*/
{
#ifdef GRAPHICS
    int nHits, k, d;
    ObjPtr retVal;
    ObjPtr anObject = NULLOBJ;
    ObjPtr whichObject;
    VertexPtr whichVertex;
    real px, py, pz;
    FuncTyp method;
    Matrix projMatrix, viewMatrix, curMatrix;
    real xnew, ynew, znew, wnew;
    real bestDist;
    real r2;

    /*Get the current matrices for transformation*/
    mmode(MPROJECTION);
    getmatrix(projMatrix);
    mmode(MVIEWING);
    getmatrix(viewMatrix);
    MULTMATRIX(viewMatrix, projMatrix, curMatrix);

    /*Restore the old matrix*/
    loadmatrix(tempMatrix);
    nHits = endpick(pickBuffer);
    if (nHits < 0)
    {
	ReportError("StopPick", "Pick buffer was too small");
	nHits = -nHits;
    }

    /*Go through the hits and distribute through the buffer*/
    retVal = NULLOBJ;
    bestDist = PLUSINF;

    k = 0;
    while (nHits)
    {
	/*Go through the hits and see which one is the closest*/

	if (pickBuffer[k] != 4)
	{
	    char err[200];
	    sprintf(err, "Bad pick size %d shorts\n", pickBuffer[k]);
	    ReportError("StopPick", err);
	    return NULLOBJ;
	}
	++k;
	--nHits;

	whichObject = (ObjPtr) ((pickBuffer[k] << 16) | (0xFFFF & pickBuffer[k + 1]));
	k += 2;
	whichVertex = (VertexPtr) ((pickBuffer[k] << 16) | (0xFFFF & pickBuffer[k + 1]));
	k += 2;

	anObject = whichObject;

	if (whichVertex)
	{
	    /*See if this is a better pick*/

	    px = whichVertex -> position[0];
	    py = whichVertex -> position[1];
	    pz = whichVertex -> position[2];
    
	    xnew = px * curMatrix[0][0] + py * curMatrix[1][0] +
		   pz * curMatrix[2][0] +      curMatrix[3][0];
    
	    ynew = px * curMatrix[0][1] + py * curMatrix[1][1] +
		   pz * curMatrix[2][1] +      curMatrix[3][1];
    
	    znew = px * curMatrix[0][2] + py * curMatrix[1][2] +
		   pz * curMatrix[2][2] +      curMatrix[3][2];
    
	    wnew = px * curMatrix[0][3] + py * curMatrix[1][3] +
		   pz * curMatrix[2][3] +      curMatrix[3][3];
    
	    xnew /= wnew;
	    ynew /= wnew;
	    znew /= wnew;
    
	    r2 = xnew * xnew + ynew * ynew;
    
	    if (r2 < bestDist)
	    {
		retVal = whichObject;
		if (bestVertex)
		{
		    *bestVertex = whichVertex;
		}
		bestDist = r2;
	    }
	}
    }

    return retVal ? retVal : anObject;
#endif
}
