/*ScianArenas.c
  Eric Pepke
  June 20, 1994
  Arena system in SciAn
*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianLists.h"
#include "ScianWindows.h"
#include "ScianColors.h"
#include "ScianIDs.h"
#include "ScianObjWindows.h"
#include "ScianErrors.h"
#include "ScianDraw.h"
#include "ScianControls.h"
#include "ScianArrays.h"
#include "ScianScripts.h"
#include "ScianVisWindows.h"
#include "ScianDialogs.h"
#include "ScianIcons.h"
#include "ScianEvents.h"
#include "ScianStyle.h"
#include "ScianObjFunctions.h"
#include "ScianPictures.h"
#include "ScianSpaces.h"
#include "ScianLights.h"
#include "ScianArenas.h"
#include "ScianPerspec.h"
#include "ScianPreferences.h"
#include "ScianDatasets.h"
#include "ScianSound.h"
#include "ScianVisStreams.h"

/*Line for mkmk*/
/*$cc -Olimit 2000 ScianArenas.c -c $(CFLAGS)
*/

Bool drawCabinets = true;
#define KEYROTFACTOR	0.05
#define KEYMOVEFACTOR	0.1

real arenaTime;

WinInfoPtr arenaWindowClass;
WinInfoPtr curArenaWindow = 0;		/*Current arena window*/
ObjPtr arenaClass;			/*Class of all arenas*/
ObjPtr curArena;			/*The current arena open*/
ObjPtr curArenaLights;
ObjPtr openArena = NULLOBJ;		/*The open arena*/
ObjPtr wandClass = NULLOBJ;		/*Class of wands*/
ObjPtr curArenaDraw = NULLOBJ;		/*Current obj being drawn in arena*/
ObjPtr curRayIntersect = NULLOBJ;	/*Current object with which ray intersects*/
ObjPtr wandToolClass = NULLOBJ;		/*Class of wand tools*/
ObjPtr fingerWandToolClass = NULLOBJ;	/*Class of finger tools*/
ObjPtr spaceMoveToolClass = NULLOBJ;	/*Space move tool*/
ObjPtr rakeWandToolClass = NULLOBJ;

ObjPtr wallClass = NULLOBJ;		/*Class of walls*/

static float material[30];

#define NARENALIGHTS		2		/*Number of lights available by default*/

lightSpec defArenaLights[] =
    {
	{(M_PI_4 * 1.2), M_PI_4, {1.0, 1.0, 1.0}, 1.0, false},
	/*{M_PI_2, M_PI_4, {0.7, 1.0, 1.0}, 0.5, false},*/
	{-M_PI_4, - 0.7 * M_PI_4, {0.7, 1.0, 1.0}, 0.5, false},
	{M_PI, M_PI_4, {1.0, 0.0, 0.5}, 0.2, true},
	{0.0, M_PI_2, {1.0, 1.0, 0.6}, 0.3, true}
    };

real arenaWindowAngle = AW_STARTANGLE;

#ifdef HAVE_PROTOTYPES
static void DrawArenaObject(ObjPtr object);
#else
static void DrawArenaObject();
#endif

#ifdef HAVE_PROTOTYPES
static ObjPtr NewWall(char *name)
#else
static ObjPtr NewWall(name)
char *name;
#endif
/*Returns a new wall*/
{
    ObjPtr retVal;

    retVal = NewObject(wallClass, 0L);
    SetVar(retVal, NAME, NewString(name));

    SetVar(retVal, CONTENTS, NewList());

    return retVal;
}

static MakeMethod MakeWallAppearance(wall)
ObjPtr wall;
/*Makes a wall's appearance*/
{
    SetVar(wall, APPEARANCE, ObjTrue);
    ImInvalid(wall);

    return ObjTrue;
}

/*Windows to keep for demo*/
char *demoWindows[] = 
    {
	"Arena",
	"SciAn",
	"Datasets"
    };

static void DisposeIfNotDemo(window)
WinInfoPtr window;
/*Disposes window if it's not a demo window*/
{
    ObjPtr name;

    MakeVar(window, NAME);
    name = GetVar((ObjPtr) window, NAME);
    if (name)
    {
	int k;

	for (k = 0; k < sizeof(demoWindows) / sizeof(char *); ++k)
	{
	    if (0 == strcmp2(demoWindows[k], GetString(name)))
	    {
		return;
	    }
	}
    }

    DeferMessage((ObjPtr) window, DISPOSE);
}

#ifdef HAVE_PROTOTYPES
void ResetDemoArenaWindows(void)
#else
void ResetDemoArenaWindows();
#endif
/*Resets the windows for the demo*/
{
    ForAllWindows(DisposeIfNotDemo);
}

static ObjPtr WallWandInteract(wall, intType, loc, dir, flags)
ObjPtr wall;
int intType;
real dir[3];
real loc[3];
/*Interacts a wand with a wall*/
{
    ObjPtr var;
    Quaternion rot, rotInv;
    real wloc[3];
    real newLoc[3];
    real newDir[3];
    real distance;
    FuncTyp method;
    ObjPtr contents;

    /*Get the location*/
    var = GetVar(wall, ARENALOCATION);
    if (!var) return ObjFalse;
    Array2CArray(wloc, var);

    /*Get the rotation*/
    var = GetVar(wall, ARENAROTATION);
    if (!var) return ObjFalse;
    Array2CArray(rot, var);

    /*Transform wand to be in coordinates of wand*/
    newLoc[0] = loc[0] - wloc[0];
    newLoc[1] = loc[1] - wloc[1];
    newLoc[2] = loc[2] - wloc[2];

    QInvert(rot, rotInv);

    QRot(newLoc, rot, newLoc);
    QRot(dir, rot, newDir);

    contents = GetVar(wall, CONTENTS);
    if (contents)
    {
	method = GetMethod(contents, WANDINTERACT);
	if (method)
	{
	    return (*method)(contents, intType, newLoc, newDir, flags);
	}
    }

    return NULLOBJ;
}

static ObjPtr DrawWallContentsInArena(wall)
ObjPtr wall;
/*Draws a wall's contents in the arena*/
{
    ObjPtr var;
    ObjPtr contents;

    contents = GetVar(wall, CONTENTS);
    if (contents)
    {
	ThingListPtr runner;
	for (runner = LISTOF(contents); runner; runner = runner -> next)
	{
	    DrawArenaObject(runner -> thing);
	}
    }

    return ObjTrue;
}

static ObjPtr DrawWallInArena(wall)
ObjPtr wall;
/*Draws a wall in an arena*/
{
    ObjPtr var;
    Quaternion rot;
    real loc[3];
    real m[3][3];
    int i, j;
    Matrix xform;
    FuncTyp method;

    /*Get the location*/
    var = GetVar(wall, ARENALOCATION);
    if (!var) return ObjFalse;
    Array2CArray(loc, var);

    /*Get the rotation*/
    var = GetVar(wall, ARENAROTATION);
    if (!var) return ObjFalse;
    Array2CArray(rot, var);

#ifdef GRAPHICS
    /*Draw the wall*/
    pushmatrix();

    translate(loc[0], loc[1], loc[2]);

    QuaternionToMatrix(rot, m);
    MATCOPY(xform, Identity);
    for (i = 0; i < 3; ++i)
    {
	for (j = 0; j < 3; ++j)
	{
	    xform[i][j] = m[i][j];
	}
    }
    multmatrix(xform);


    lmcolor(LMC_DIFFUSE);

    method = GetMethod(wall, DRAWARENACONTENTS);
    if (method)
    {
	(*method)(wall);
    }

    popmatrix();
#endif

    return ObjTrue;
}

static ObjPtr NewArenaLight(k)
int k;
/*Returns a new light with default definition k*/
{
    ObjPtr light;
    ObjPtr baseColor;
    ObjPtr name;
    ObjPtr loc;
    real *locPtr;

    light = NewObject(lightClass, 0);

    /*Set light position*/
    loc = NewRealArray(1, 3L);
    locPtr = ArrayMeat(loc);
    *locPtr++ = rcos(defArenaLights[k] . theta) * rcos(defArenaLights[k] . phi);
    *locPtr++ = rsin(defArenaLights[k] . theta) * rcos(defArenaLights[k] . phi);
    *locPtr++ = rsin(defArenaLights[k] . phi);
    SetVar(light, LOCATION, loc);

    /*Set light color*/
    baseColor = NewRealArray(1, 3L);
    CArray2Array(baseColor, defArenaLights[k] . baseColor);
    SetVar(light, BASECOLOR, baseColor);

    /*Set light brightness*/
    SetVar(light, BRIGHTNESS, NewReal(defArenaLights[k] . brightness));

    /*Set whether the light is enabled or not*/
    SetVar(light, HIDDEN, NewInt(defArenaLights[k] . enabled));
    
    return light;
}

#ifdef HAVE_PROTOTYPES
void AddWindowToArena(WinInfoPtr window)
#else
void AddWindowToArena(window)
WinInfoPtr window;
#endif
/*Adds window to the arena if it isn't already there*/
{
    real loc[3], tloc[3];
    real axis[3];
    Quaternion rot;
    ObjPtr var;
    int k;
    int d;
    ThingListPtr runner;

    if (GetPredicate((ObjPtr) window, HIDDEN))
    {
	return;
    }

    if (openArena)
    {
	/*Stick it in the arena*/
	ObjPtr contents;
	char *prefWallName;
	ObjPtr prefWall = NULLOBJ;
	ThingListPtr runner;

	contents = GetListVar("AddWindowToArena", openArena, CONTENTS);

	var = GetVar((ObjPtr) window, PREFWALL);
	if (var)
	{
	    prefWallName = GetString(var);
 	}
	else
	{
	    prefWallName = (char *) 0;
	}

	/*Search for a wall in the arena*/
	prefWall = NULLOBJ;

	if (prefWallName)
	{
	    /*Search for a wall with the name given*/
	    for (runner = LISTOF(contents); runner; runner = runner -> next)
	    {
		if (IsWall(runner -> thing))
		{
		    var = GetVar(runner -> thing, NAME);
		    if (var && 0 == strcmp2(prefWallName, GetString(var)))
		    {
			prefWall = runner -> thing;
		    }
		}
	    }
	}

	if (!prefWall)
	{
	    for (runner = LISTOF(contents); runner; runner = runner -> next)
	    {
		if (IsWall(runner -> thing))
		{
		    prefWall = runner -> thing;
		    break;
		}
	    }
	}

	if (!prefWall)
	{
	    ReportError("AddWindowToArena", "No wall found");
 	}
	else
	{
	    /*There's a preferred wall.  Add it.*/

	    contents = GetListVar("AddWindowToArena", prefWall, CONTENTS);

	    if (!contents) return;

	    if (IsInSet(contents, (ObjPtr) window))
	    {
		/*Already there; don't do anything*/
		return;
	    }

	    /*Find a location*/
	    for (k = 0; ; ++k)
	    {
		for (d = 1; d >= -1; d -= 2)
		{
		    loc[0] = (k * d) * AW_WINXSPACE;
		    loc[1] = AW_DEFHEIGHT;
		    loc[2] = 0.0;

		    for (runner = LISTOF(contents); runner; runner = runner -> next)
		    {
			var = GetFixedArrayVar("AddWindowToArena", runner -> thing, ARENALOCATION, 1, 3L);
			if (!var) return;

			Array2CArray(tloc, var);

			if (ABS(loc[0] - tloc[0]) < AW_WINXMINSPACE)
			{
			    goto nogood;
			}
		    }
		    goto fine;
nogood:;
		}
	    }
fine:
	    /*Give the window a location*/
	    var = NewRealArray(1, 3L);
	    CArray2Array(var, loc);
	    SetVar((ObjPtr) window, ARENALOCATION, var);
    
	    rot[0] = 1.0;
	    rot[1] = 0.0;
	    rot[2] = 0.0;
	    rot[3] = 0.0;
	    var = NewRealArray(1, 4L);
	    CArray2Array(var, rot);
	    SetVar((ObjPtr) window, ARENAROTATION, var);
    
	    /*Give it a default width, height, thickness*/
	    SetVar((ObjPtr) window, ARENALENGTH, NewReal(AW_DEFOPENLENGTH));
	    SetVar((ObjPtr) window, ARENAWIDTH, NewReal(AW_DEFWIDTH));

	    SetVar(prefWall, CONTENTS, contents = AddToSet(contents, (ObjPtr) window));

	}
    }
}

#ifdef HAVE_PROTOTYPES
void RemoveWindowFromArena(WinInfoPtr window)
#else
void RemoveWindowFromArena(window)
ObjPtr window;
#endif
/*Removes window from the arena*/
{
    ObjPtr name;
    ThingListPtr runner;

    if (openArena)
    {
	ObjPtr windows;

	windows = GetListVar("RemoveWindowFromArena", openArena, CONTENTS);
	if (windows)
	{
	    for (runner = LISTOF(windows); runner; runner = runner -> next)
	    {
		ObjPtr contents;
		contents = GetListVar("RemoveWindowFromArena", runner -> thing, CONTENTS);
		SetVar(runner -> thing, CONTENTS, RemoveFromList(contents, (ObjPtr) window));
	    }
	}
    }
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewArena(int left, int right, int bottom, int top)
#else
ObjPtr NewArena(left, right, bottom, top)
int left, right, bottom, top;
#endif
/*Makes a new arena with bounds left, right, bottom, top.
  There can be only one.*/
{
    ObjPtr wall, contents;
    real loc[3];
    real axis[3];
    Quaternion rot1, rot2, rrot;
    ObjPtr var;

    if (openArena) return openArena;

    openArena = NewObject(arenaClass, 0);
	
    Set2DIntBounds(openArena, left, right, bottom, top);
    SetVar(openArena, CONTENTS, contents = NewList());

    /*Put in the walls*/
    wall = NewWall("Left");
    PostfixList(contents, wall);
 
    loc[0] = -3.0;
    loc[1] = 0.0;
    loc[2] = 0.0;
    var = NewRealArray(1, 3L);
    CArray2Array(var, loc);
    SetVar(wall, ARENALOCATION, var);

    /*Give it a rotation*/
    axis[0] = 1.0;
    axis[1] = 0.0;
    axis[2] = 0.0;
    RotToQuaternion(axis, M_PI_2, rot1);

    axis[0] = 0.0;
    axis[1] = 1.0;
    axis[2] = 0.0;
    RotToQuaternion(axis, M_PI_2, rot2);

    QMult(rot1, rot2, rrot);

    var = NewRealArray(1, 4L);
    CArray2Array(var, rrot);
    SetVar(wall, ARENAROTATION, var);

    wall = NewWall("Front");
    PostfixList(contents, wall);

    loc[0] = 0.0;
    loc[1] = 3.0;
    loc[2] = 0.0;

    var = NewRealArray(1, 3L);
    CArray2Array(var, loc);
    SetVar(wall, ARENALOCATION, var);

    /*Give it a rotation*/
    axis[0] = 1.0;
    axis[1] = 0.0;
    axis[2] = 0.0;
    RotToQuaternion(axis, M_PI_2, rrot);

    var = NewRealArray(1, 4L);
    CArray2Array(var, rrot);
    SetVar(wall, ARENAROTATION, var);

    return openArena;
}

#ifdef HAVE_PROTOTYPES
void StartArena(ObjPtr arena, int left, int right, int bottom, int top,
	int action, int whichView)
#else
void StartArena(arena, left, right, bottom, top, action, whichView)
ObjPtr arena;
int left, right, bottom, top;
int action;
int whichView;
#endif
/*Start a drawing, rotation, or press in an arena.  Action is the action
  that is being performed*/
{
#ifdef GRAPHICS
    real eyePosn[3];		/*Coords of the eye*/
    real focusPoint[3];		/*Coords of the focus point*/
    real upVector[3];		/*The up vector of the observer*/
    real forwardVector[3];
    ObjPtr object;
    ObjPtr psStuff;
    ObjPtr boundsArray;
    real bounds[4];
    real bigBounds[6];		/*Bounds of the objects*/
    Coord center[3];		/*The center of the data set*/
    ObjPtr contents;
    real scaleFactor[3];	/*Factor to scale*/
    real maxSize;		/*Maximum size of the objects this draw*/
    ObjPtr observer;		/*Observer of the arena*/
    real eyeOffset;
    ObjPtr var;
    Matrix proj, model;
    float aspect;		/*Aspect ratio*/
    real sideVector[3];

    curWhichView = whichView;

    /*Change bounds and aspect ratio depending on whichView*/
    aspect = ASPECT;
    switch(VIEWOF(whichView))
    {
	case VIEW_XLEFT:
	case VIEW_WRIGHT:
	case VIEW_CSRIGHT:
	    left = (left + right) / 2 + 1;
	    aspect = ASPECT;
	    break;
	case VIEW_XRIGHT:
	case VIEW_WLEFT:
	case VIEW_CSLEFT:
	    right = (left + right) / 2;
	    aspect = ASPECT;
	    break;
	case VIEW_CLEFT:
	    bottom = YSTEREO + YBLANK;
	    top = bottom + YSTEREO;
	    break;
	case VIEW_CRIGHT:
	    bottom = 0;
	    top = bottom + YSTEREO;
	    break;
    }

    if (!CAVEP)
    {
	/*Set a new viewport to the current area*/
	SetSubPort(left, right, bottom, top);
    }

    MakeVar(arena, OBSERVER);
    observer = GetObjectVar("StartArena", arena, OBSERVER);
    if (!observer) return;

    if (ISLEFT(whichView))
    {
	    var = GetVar(observer, BINOCULARITY);
	    eyeOffset = -GetReal(var) * 0.5;
    }
    else if (ISRIGHT(whichView))
    {
	    var = GetVar(observer, BINOCULARITY);
	    eyeOffset = GetReal(var) * 0.5;
    }
    else
    {
	    eyeOffset = 0.0;
    }

    /*Set up current viewing parameters*/
    var = GetVar(observer, FOCUSDIST);
    if (var)
    {
	curFocusDist = GetReal(var);
    }
    else
    {
	curFocusDist = INITEYEDIST;
    }

    var = GetVar(observer, VIEWFIELD);
    if (var)
    {
	curViewField = GetReal(var);
    }
    else
    {
	curViewField = INITAOV;
    }

    var = GetVar(observer, VIEWCLIP);
    if (var)
    {
	Array2CArray(curViewClip, var);
    }
    else
    {
	curViewClip[0] = INITNEARCLIP;
	curViewClip[1] = INITFARCLIP;
    }
    if (curViewClip[0] < MINCLIP) curViewClip[0] = MINCLIP;
    if (curViewClip[1] < MINCLIP) curViewClip[1] = MINCLIP;

    /*Get information about the eye*/
    GetObserverLocation(eyePosn, observer);
    GetFocusPoint(focusPoint, observer);
    GetAdjustedVectors(forwardVector, upVector, observer);

    if (ISSTEREO(whichView))
    {
	real temp;

	/*Have to slide eyePosn and focusPoint over somewhat*/
	CROSS(forwardVector, upVector, sideVector);

	/*Move eye over*/
	eyePosn[0] += sideVector[0] * eyeOffset;
	eyePosn[1] += sideVector[1] * eyeOffset;
	eyePosn[2] += sideVector[2] * eyeOffset;
    }

    /*Get information about the clock*/
    MakeVar(arena, TIME);
    var = GetVar(arena, TIME);
    if (var)
    {
	arenaTime = GetReal(var);
    }
    else
    {
	arenaTime = 0.0;
    }

    /*Set the current arena*/
    curArena = arena;

    if (action == PICKSPACE)
    {
	StartPick();
    }

    if (!CAVEP)
    {
#ifdef GL4D
	mmode(MPROJECTION);
#endif
	if (whichView & VIEW_MOD_ORTHO)
	{
	    /*Plain old orthographic projection.*/
	    ortho(-aspect * curViewField, aspect * curViewField,
		  -curViewField, curViewField, curViewClip[0], curViewClip[1]);
	}
	else
	{
	    /*Exciting perspective projection!!!*/
	    perspective((long) (curViewField * 10.0), aspect,
			curViewClip[0], curViewClip[1]);
	}
    
#ifdef GL4D
	mmode(MVIEWING);
	loadmatrix(Identity);
#endif
    }
    else
    {
	mmode(MVIEWING);
    }

    switch(VIEWOF(whichView))
    {
	case VIEW_CSRIGHT:
	    upat(eyePosn[0], eyePosn[1], eyePosn[2],
		focusPoint[0], focusPoint[1], focusPoint[2], 
		-sideVector[0], -sideVector[1], -sideVector[2]);
	    break;
	case VIEW_CSLEFT:
	    upat(eyePosn[0], eyePosn[1], eyePosn[2],
		focusPoint[0], focusPoint[1], focusPoint[2], 
		sideVector[0], sideVector[1], sideVector[2]);
	    break;	
	default:
	    upat(eyePosn[0], eyePosn[1], eyePosn[2],
		focusPoint[0], focusPoint[1], focusPoint[2], 
		upVector[0], upVector[1], upVector[2]);	
	    break;
    }

    if (CAVEP)
    {
	translate(0.0, 0.0, CAVEORIGINOFFSET);
	scale(CAVESCALE, CAVESCALE, CAVESCALE);
    }

    getmatrix(viewerCoordsMatrix);

#if 0
    printf("Matrix from arena:\n");
    PrintMatrix(viewerCoordsMatrix);
#endif

    /*Set up the lights*/
    MakeVar(arena, LIGHTS);
    curArenaLights = GetVar(arena, LIGHTS);

    if (action == DRAWSPACE)
    {
	int i;
	real radius;

	/*Shift filter for 3-D glasses*/
	if (rgbp && (VIEWOF(whichView) == VIEW_CRLEFT || VIEWOF(whichView) == VIEW_RCRIGHT))
	{
	    BeginMask(false, true, true, true);
	}
	else if (rgbp && (VIEWOF(whichView) == VIEW_CRRIGHT || VIEWOF(whichView) == VIEW_RCLEFT))
	{
	    BeginMask(true, false, false, true);
	}
	else if (rgbp && (VIEWOF(whichView) == VIEW_GRLEFT || VIEWOF(whichView) == VIEW_RGRIGHT))
	{
	    BeginMask(false, true, false, true);
	}
	else if (rgbp && (VIEWOF(whichView) == VIEW_GRRIGHT || VIEWOF(whichView) == VIEW_RGLEFT))
	{
	    BeginMask(true, false, false, true);
	}
	else
	{
	    BeginMask(true, true, true, true);
	}

	if (VIEWOF(whichView) == VIEW_EOLEFT || VIEWOF(whichView) == VIEW_OERIGHT)
	{
	    BeginEvenOnly();
	}
	if (VIEWOF(whichView) == VIEW_OELEFT || VIEWOF(whichView) == VIEW_EORIGHT)
	{
	    BeginOddOnly();
	}

	/*Go to Gouraud shading*/
	shademodel(GOURAUD);

	/*Set up the Z-buffer*/
	if (action == DRAWSPACE)
	{
#ifdef GC_ZMIN
		curZMin = getgconfig(GC_ZMIN);
		curZMax = getgconfig(GC_ZMAX);
#else
		curZMin = 0;
		curZMax = 0x7FFFFF;
#endif
	    nudgeSize = (long) ((((real) curZMax) - ((real) curZMin)) * NUDGEFACTOR);

	    curZMin += nudgeSize * MAXNNUDGES;
	    curZMax -= nudgeSize * MAXNNUDGES;

	    if (!CAVEP)
	    {
	        lsetdepth(curZMin, curZMax);
	    }
	    zbuffer(TRUE);
	    zclear();
	}

	/*Draw rotation and motion guides*/
	if (whichView & VIEW_MOD_ORTHO)
	{
	    radius = SPACESPHERESIZE * 1.85 * curViewField;
	}
	else
	{
	    radius = SPACESPHERESIZE * curFocusDist * rsin(curViewField * M_PI / 180.0);
	}

	if (GetPredicate(curArena, SHOWROTGUIDES))
	{
	    SetUIColor(UIWHITE);

	    DrawWFSphere(focusPoint[0], focusPoint[1], focusPoint[2], radius);
	    DrawSpaceLine(focusPoint[0] - radius, focusPoint[1], focusPoint[2],
			  focusPoint[0] + radius, focusPoint[1], focusPoint[2]);
	    DrawSpaceLine(focusPoint[0], focusPoint[1] - radius, focusPoint[2],
			  focusPoint[0], focusPoint[1] + radius, focusPoint[2]);
	    DrawSpaceLine(focusPoint[0], focusPoint[1], focusPoint[2] - radius,
			  focusPoint[0], focusPoint[1], focusPoint[2] + radius);
	}
	if (GetPredicate(curArena, SHOWMOTGUIDES))
	{
	    real x, y;
	    SetUIColor(UIGRAY50);
#if 0
	    for (x = -10.0; x <= 10.0; x += 1.0)
	    {
		for (y = -10.0; y <= 10.0; y += 1.0)
		{
		    DrawSpaceLine(x, y, -10.0, x, y, 10.0);
		    DrawSpaceLine(x, -10.0, y, x, 10.0, y);
		    DrawSpaceLine(-10.0, x, y, 10.0, x, y);
		}
	    }
#endif
	    BeginMask(true, true, true, false);
	    for (x = -10.0; x <= 10.0; x += 1.0)
	    {
		DrawSpaceLine(x, -10.0, -1.0, x, 10.0, -1.0);
		DrawSpaceLine(x, -10.0, 1.0, x, 10.0, 1.0);
	    }
	    for (y = -10.0; y <= 10.0; y += 1.0)
	    {
		DrawSpaceLine(-10.0, y, -1.0, 10.0, y, -1.0);
		DrawSpaceLine(-10.0, y, 1.0, 10.0, y, 1.0);
	    }
	    EndMask();
#if 0
	    DrawSpaceLine(-0.1, 0.0, 0.0,  0.0, 0.1, 0.0);
	    DrawSpaceLine(0.0, 0.1, 0.0,  0.1, 0.0, 0.0);
	    DrawSpaceLine(0.1, 0.0, 0.0,  0.0, -0.1, 0.0);
	    DrawSpaceLine(0.0, -0.1, 0.0, -0.1, 0.0, 0.0);
#endif
	}
    }

    /*Save observer position*/
    observerPosition[0] = eyePosn[0] * scaleFactor[0] + center[0];
    observerPosition[1] = eyePosn[1] * scaleFactor[1] + center[1];
    observerPosition[2] = eyePosn[2] * scaleFactor[2] + center[2];

#endif
}

#ifdef HAVE_PROTOTYPES
void StopArena(int action)
#else
void StopArena(action)
int action;
#endif
/*Stops drawing, rotating, or pressing in a arena*/
{
#ifdef GRAPHICS
    if (action == DRAWSPACE)
    {
	real radius;

	OptimizeSharpness();

	if (curWhichView & VIEW_MOD_ORTHO)
	{
	    radius = SPACESPHERESIZE * 1.85 * curViewField;
	}
	else
	{
	    radius = SPACESPHERESIZE * curFocusDist * rsin(curViewField * M_PI / 180.0);
	}

	loadmatrix(viewerCoordsMatrix);

	EndMask();

	if (VIEWOF(curWhichView) == VIEW_EOLEFT || VIEWOF(curWhichView) == VIEW_OERIGHT)
	{
	    EndEvenOnly();
	}
	if (VIEWOF(curWhichView) == VIEW_OELEFT || VIEWOF(curWhichView) == VIEW_EORIGHT)
	{
	    EndOddOnly();
	}
    }

    if (!CAVEP)
    {
	if (action == DRAWSPACE)
	{
	    zbuffer(FALSE);
	}
    
	RestoreSubPort();
    }

#endif
    curArena = NULLOBJ;
}

#ifdef HAVE_PROTOTYPES
static void DrawArenaObject(ObjPtr object)
#else
static void DrawArenaObject(object)
ObjPtr object;
#endif
/*Draws an object within an arena*/
{
    FuncTyp DrawFunction;			/*Method*/

    if (GetPredicate(object, INHIBITDRAWING))
    {
	return;
    }

    if (GetPredicate(object, HIDDEN))
    {
	return;
    }

    if (curArenaLights)
    {
	StartLights(curArena, curArenaLights, false);
    }

#ifdef GRAPHICS
    lmbind(MATERIAL, ARENAMATERIAL);
    lmcolor(LMC_DIFFUSE);
#endif

    OptimizeColor();

    curArenaDraw = object;

    BoundsInvalid(object, GetVarChangeCount(object, BUFFER0));
    SetVar(object, BUFFER0, ObjTrue);

    DrawFunction = GetMethod(object, DRAWARENA);
    if (DrawFunction)
    {
	(*DrawFunction)(object);
    }
}

#ifdef HAVE_PROTOTYPES
void DrawArenaContents(ObjPtr arena, int left, int right, int bottom, int top, int viewType)
#else
void DrawArenaContents(arena, left, right, bottom, top, viewType)
ObjPtr arena;
int left, right, bottom, top;
int viewType;
#endif
{
#ifdef GRAPHICS
    ObjPtr contents;
    Bool anyTransparent;		/*True iff any objects are transparent*/
    ThingListPtr drawList;		/*List of objects to draw*/
    ObjPtr frontPanel;
    int prefDrawMoving;			/*Preferred draw moving style*/
    ObjPtr var;
    Bool isCached;
    ObjPtr wand, tool;

    /*Set a new viewport to the current area*/
    StartArena(arena, left, right, bottom, top, DRAWSPACE, viewType);

    /*Figure out the range of drawing qualities to use.*/
    if (interactiveMoving)
    {
	/*It's interactive.  Figure out what quality to use*/
	prefDrawMoving = GetPrefInteger(PREF_DRAWMOVING);

	if (prefDrawMoving == DM_SKELETON)
	{
	    minDrawingQuality = DQ_MIN;
	    maxDrawingQuality = DQ_OUTLINE;
	}
	else if (prefDrawMoving == DM_FULL)
	{
	    minDrawingQuality = DQ_MIN;
	    maxDrawingQuality = DQ_MAX;
	}
	else if (prefDrawMoving == DM_AUTO)
	{
	    var = GetVar(arena, DRAWINGTIME);
	    if (var)
	    {
		long drawingTime;
		/*There is a drawing time recorded.  Get it*/
		drawingTime = GetInt(var);

		if (drawingTime < MAXMOVETIME)
		{
		    /*It's fast enough, might as well do it fully*/
		    minDrawingQuality = DQ_MIN;
		    maxDrawingQuality = DQ_MAX;
		}
		else
		{
		    /*Do outline*/
		    minDrawingQuality = DQ_MIN;
		    maxDrawingQuality = DQ_OUTLINE;
		}
	    }
	    else
	    {
		/*Assume we have to do full drawing*/
		minDrawingQuality = DQ_MIN;
		maxDrawingQuality = DQ_MAX;
	    }
	}
    }
    else
    {
	/*Not interactive.  Just use full quality*/
	minDrawingQuality = DQ_MIN;
	maxDrawingQuality = DQ_MAX;
    }

    /*See if the wand is a fingertool*/
    wand = GetObjectVar("DrawArenaContents", arena, WAND);
    if (wand)
    {
	tool = GetObjectVar("DrawArenaContents", wand, EDITTOOL);
	if (tool)
	{
	    var = GetVar(tool, CLASSID);
	    if (var && GetInt(var) == CLASS_FINGERWANDTOOL)
	    {
		drawCabinets = true;
	    }
	    else
	    {
		drawCabinets = false;
	    }
	}
    }

    /*Draw the contents of the arena*/
    contents = GetVar(arena, CONTENTS);

    /*First draw opaque objects*/
    anyTransparent = false;
    drawList = LISTOF(contents);
    drawingTransparent = false;

    while (drawList)
    {
	if (GetPredicate(drawList -> thing, ISTRANSPARENT))
	{
	    anyTransparent = true;
	}
	qualityDrawnSoFar = DQ_MIN;

	DrawArenaObject(drawList -> thing);

	if (maxDrawingQuality < DQ_FULL)
	{
	    NeedToDrawFullQuality(drawList -> thing);
	}
	drawList = drawList -> next;
    }

    MakeVar(arena, WAND);
    var = GetVar(arena, WAND);
    if (var)
    {
	DrawArenaObject(var);
    }

    /*Now draw transparent objects*/
    if (anyTransparent && rgbp && maxDrawingQuality >= DQ_FULL)
    {
	BeginMask(true, true, true, false);
	blendfunction(BF_ONE, BF_ONE);

	drawingTransparent = true;
	drawList = LISTOF(contents);
	while (drawList)
	{
	    minDrawingQuality = DQ_FULL;	
	    maxDrawingQuality = DQ_FULL;
	    qualityDrawnSoFar = DQ_MIN;
	    DrawArenaObject(drawList -> thing);

	    if (maxDrawingQuality < DQ_FULL)
	    {
		NeedToDrawFullQuality(drawList -> thing);
	    }
	    drawList = drawList -> next;
	}
	blendfunction(BF_ONE, BF_ZERO);
	EndMask();
	drawingTransparent = false;
    }

    if (maxDrawingQuality >= DQ_FULL)
    {
	/*End the timing of the full drawing*/
	SetVar(arena, DRAWINGTIME, NewInt(GetPictureTime()));
    }

    StopArena(DRAWSPACE);
#endif
}

static ObjPtr DrawArena(object)
ObjPtr object;
/*Draws an arena and everything it contains*/
{
#ifdef GRAPHICS
    int left, right, bottom, top;
    ObjPtr eyePosnObj;			/*Eyeposition object*/
    ObjPtr tempObj;			/*Temporary object*/
    ObjPtr dRotMatrix;
    ObjPtr observer, observers;		/*Observer of the space*/
    ObjPtr panel;			/*Front and back panel*/
    ObjPtr var;				/*A random variable*/
    int viewType;			/*Type of view*/
    int stereoMode;			/*Stereo mode*/

    Get2DIntBounds(object, &left, &right, &bottom, &top);
    if (!CAVEP)
    {
	if (IsDrawingRestricted(left, right, bottom, top)) return ObjFalse;
    }

    EraseAll();

    /*Update the rotation matrix and fly through*/
    MakeVar(object, OBSERVER);
    observer = GetObjectVar("DrawArena", object, OBSERVER);

    var = GetVar(observer, VIEWTYPE);
    if (var)
    {
	viewType = GetInt(var);
    }
    else
    {
	viewType = VT_PERSPECTIVE;
    }

    var = GetVar(observer, STEREOMODE);
    if (var)
    {
	stereoMode = GetInt(var);
    }
    else
    {
	stereoMode = SM_MONO;
    }

    if (drawingMode == DRAW_SCREEN)
    {
	switch(stereoMode)
	{
	    case SM_MONO:
		DrawArenaContents(object, left, right, bottom, top, VIEW_CENTER | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		break;
	    case SM_CROSSEYED:
		DrawArenaContents(object, left, right, bottom, top, VIEW_XLEFT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		DrawArenaContents(object, left, right, bottom, top, VIEW_XRIGHT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		break;
	    case SM_EVENODD:
		DrawArenaContents(object, left, right, bottom, top, VIEW_EOLEFT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		DrawArenaContents(object, left, right, bottom, top, VIEW_EORIGHT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		break;
	    case SM_ODDEVEN:
		DrawArenaContents(object, left, right, bottom, top, VIEW_OELEFT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		DrawArenaContents(object, left, right, bottom, top, VIEW_OERIGHT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		break;
	    case SM_WALLEYED:
		DrawArenaContents(object, left, right, bottom, top, VIEW_WLEFT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		DrawArenaContents(object, left, right, bottom, top, VIEW_WRIGHT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		break;
	    case SM_CYBERSCOPE:
		DrawArenaContents(object, left, right, bottom, top, VIEW_CSLEFT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		DrawArenaContents(object, left, right, bottom, top, VIEW_CSRIGHT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		break;
	    case SM_LEFTONTOP:
		DrawArenaContents(object, left, right, bottom, top, VIEW_CLEFT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		DrawArenaContents(object, left, right, bottom, top, VIEW_CRIGHT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		break;
	    case SM_REDCYAN:
		if (rgbp)
		{
		    DrawArenaContents(object, left, right, bottom, top, VIEW_RCLEFT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		    DrawArenaContents(object, left, right, bottom, top, VIEW_RCRIGHT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		}
		else
		{
		    DrawArenaContents(object, left, right, bottom, top, VIEW_CENTER | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		}
		break;
	    case SM_CYANRED:
		if (rgbp)
		{
		    DrawArenaContents(object, left, right, bottom, top, VIEW_CRLEFT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		    DrawArenaContents(object, left, right, bottom, top, VIEW_CRRIGHT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		}
		else
		{
		    DrawArenaContents(object, left, right, bottom, top, VIEW_CENTER | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		}
		break;
	    case SM_REDGREEN:
		if (rgbp)
		{
		    DrawArenaContents(object, left, right, bottom, top, VIEW_RGLEFT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		    DrawArenaContents(object, left, right, bottom, top, VIEW_RGRIGHT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		}
		else
		{
		    DrawArenaContents(object, left, right, bottom, top, VIEW_CENTER | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		}
		break;
	    case SM_GREENRED:
		if (rgbp)
		{
		    DrawArenaContents(object, left, right, bottom, top, VIEW_GRLEFT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		    DrawArenaContents(object, left, right, bottom, top, VIEW_GRRIGHT | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		}
		else
		{
		    DrawArenaContents(object, left, right, bottom, top, VIEW_CENTER | (viewType == VT_ORTHOGRAPHIC ? VIEW_MOD_ORTHO : 0));
		}
		break;
	}
    }
#endif
    return ObjTrue;
}

static ObjPtr MakeArenaWindowDrawn(curWindow)
WinInfoPtr curWindow;
/*Method to make an obj window drawn.  Returns true iff drawing done*/
{
    int whichBuffer;

    if (IsDoubleBuf(curWindow))
    {
	/*Double buffer, determine which buffer we're drawing*/
	ObjPtr var;

	var = GetVar((ObjPtr) curWindow, WHICHBUFFER);
	if (var)
	{
	    whichBuffer = GetInt(var);
	}
	else
	{
	    whichBuffer = 0;
	}
    }
    else
    {
	/*Single buffer, just one buffer*/
	whichBuffer = 0;
    }

    {
	ObjPtr boundsInvalid;
	invalidFound = false;
	boundsInvalid = BoundsInvalid((ObjPtr) curWindow,
			    GetVarChangeCount((ObjPtr) curWindow,
				    whichBuffer ? BUFFER1 : BUFFER0));
	if (boundsInvalid)
	{
	    SetVar((ObjPtr) curWindow, whichBuffer ? BUFFER1 : BUFFER0, ObjTrue);
    
#ifndef DONTCLIP
	    if ((IsArray(boundsInvalid) && !GetPredicate((ObjPtr) curWindow, SHOWBOUNDS)))
	    {
		real elements[4];
		Array2CArray(elements, boundsInvalid);
		SetClipRect((int) elements[0], (int) elements[1], (int) elements[2], (int) elements[3]);
		DrawWindowInfo(curWindow);
		RestoreClipRect();
	    }
	    else
#endif
	    {
		DrawWindowInfo(curWindow);
	    }
    
	    if (IsDoubleBuf(curWindow))
	    {
		SetVar((ObjPtr) curWindow, WHICHBUFFER, NewInt(whichBuffer ? 0 : 1));
	    }
    
	    return ObjTrue;
	}
    }

    return ObjFalse;
}

#ifdef HAVE_PROTOTYPES
WinInfoPtr NewArenaWindow(char *title)
#else
WinInfoPtr NewArenaWindow(title)
char *title;
#endif
/*Creates a new arena window*/
{
    long flags;
    WinInfoPtr retVal;
    WindowID theWindow;
    ObjPtr contents;

    if (curArenaWindow) return curArenaWindow;

    flags = WINDBUF + WINZBUF + WINRGB + WINFIXEDLOC;

    /*Edit flags*/
    if (!hasRGB && (flags & WINRGB))
    {
	flags &= ~WINRGB;
    }
    if (!hasCmap) flags |= WINRGB;

    if (showWindows && !CAVEP)
    {
	theWindow = NewGLWindow("Arena", scrWidth, scrHeight, 0, 0,flags);
    }
    else
    {
	theWindow = 0;
    }

    retVal = NewWinInfo((ObjPtr) arenaWindowClass, theWindow, 
		flags, title,  scrWidth, scrHeight, 0, 0);
    SetVar((ObjPtr) retVal, CONTENTS, NewList());	/*Contents*/
    if (showWindows && !CAVEP)
    {
	CEWS(retVal, scrWidth, scrHeight);
    }

#ifdef SELNEWWINDOWS
    inputWindow = retVal;
#endif

    contents = GetListVar("NewArenaWindow", (ObjPtr) retVal, CONTENTS);
    if (contents)
    {
	PrefixList(contents, NewArena(0, scrWidth, 0, scrHeight));
    }

    SelWindow(retVal);

    curArenaWindow = retVal;

    return retVal;
}

static ObjPtr ArenaBoundsInvalid(object, changeCount)
ObjPtr object;
unsigned long changeCount;
/*For an arena, tests to see if it needs drawing.  Returns
  NULLOBJ	if it does not
  array[4]	giving bounds if it does
  ObjTrue	it it needs to be redrawn but does not know where
*/
{
    ObjPtr contents;
    ObjPtr myBounds;
    real boundsElements[4];
    real testElements[4];
    ObjPtr test;
    real myBoundsElements[4];
    ObjPtr scrollBar;
    Bool firstTime = true;
    Bool doubleNoBounds = false;
    ObjPtr retVal = NULLOBJ;
    ObjPtr time;
    ObjPtr lights, observer, renderer;
    ObjPtr wand;

    MakeVar(object, APPEARANCE);
    MakeVar(object, BOUNDS);

    myBounds = GetFixedArrayVar("ArenaBoundsInvalid", object, BOUNDS, 1, 4L);
    Array2CArray(myBoundsElements, myBounds);

    if (GetVarChangeCount(object, CHANGEDBOUNDS) > changeCount)
    {
	/*Object is not good, so return the bounds*/
	ObjPtr myBounds;

	myBounds = GetVar(object, CHANGEDBOUNDS);

	if (myBounds && IsArray(myBounds) && RANK(myBounds) == 1
		&& DIMS(myBounds)[0] == 4)
	{
	    retVal = myBounds;
	}
	else
	{
	    retVal = ObjTrue;
	}
    }

    contents = GetVar(object, CONTENTS);
    if (contents && IsList(contents))
    {
	/*Still, maybe some of the contents need to be drawn*/
 	ThingListPtr runner;

	if (GetVarChangeCount(contents, CHANGEDBOUNDS) > changeCount)
	{
	    /*Object is not good, so return the bounds*/
	    ObjPtr myBounds;
	
	    myBounds = GetVar(object, CHANGEDBOUNDS);
	
	    if (myBounds && IsArray(myBounds) && RANK(myBounds) == 1
		    && DIMS(myBounds)[0] == 4)
	    {
		retVal = myBounds;
	    }
	    else
	    {
		retVal = ObjTrue;
	    }
	}

	runner = LISTOF(contents);
	while (runner)
	{
	    test = BoundsInvalid(runner -> thing, changeCount);
	    if (test)
	    {
		/*Hey, the kid needs redrawing*/
		retVal = test;
	    }
	    runner = runner -> next;
	}
    }

    MakeVar(object, LIGHTS);
    lights = GetVar(object, LIGHTS);
    if (lights && IsList(lights))
    {
	/*Still, maybe some of the lights need to be drawn*/
 	ThingListPtr runner;

	runner = LISTOF(lights);
	while (runner)
	{
	    test = BoundsInvalid(runner -> thing, changeCount);
	    if (test)
	    {
		/*Hey, the light needs redrawing*/
		retVal =  myBounds;
	    }
	    runner = runner -> next;
	}
    }

    MakeVar(object, OBSERVER);
    observer = GetVar(object, OBSERVER);
    if (observer)
    {
	test = BoundsInvalid(observer, changeCount);
	if (test)
	{
	    /*Hey, the observer needs redrawing*/
	    retVal = myBounds;
	}
    }

    MakeVar(object, WAND);
    wand = GetVar(object, WAND);
    if (wand)
    {
	test = BoundsInvalid(wand, changeCount);
	if (test)
	{
	    /*Hey, the wand needs redrawing*/
	    retVal = myBounds;
	}
    }

    MakeVar(object, RENDERER);
    renderer = GetVar(object, RENDERER);
    if (renderer)
    {
	test = BoundsInvalid(renderer, changeCount);
	if (test)
	{
	    /*Hey, the renderer needs redrawing*/
	    retVal =  myBounds;
	}
    }

    return retVal;
}

#ifdef HAVE_PROTOTYPES
static ObjPtr NewWand(char *name)
#else
static ObjPtr NewWand(name)
char *name;
#endif
/*Creates and returns a new wand*/
{
    ObjPtr retVal;

    retVal = NewObject(wandClass, 0L);

    SetVar(retVal, NAME, NewString(name));

    return retVal;
}

static ObjPtr MakeArenaObserver(arena)
ObjPtr arena;
/*Makes the observer of an arena*/
{
    ObjPtr observer;
    real posn[3];
    real clip[2];
    Quaternion rotQuat;
    ObjPtr anArray;

    SetVar(arena, OBSERVER, observer = NewObserver());

    SetVar(observer, VIEWFIELD, NewReal(75.0));
    posn[0] = 0.0;
    if (CAVEP)
    {
	posn[1] = 0.0;
	posn[2] = 0.0;
    }
    else
    {
	posn[1] = -1.0;
	posn[2] = 0.5;
    }

    anArray = NewRealArray(1, 3L);
    CArray2Array(anArray, posn);
    SetVar(observer, LOCATION, anArray);
    SetVar(observer, FOCUSDIST, NewReal(CAVEP ? 0.01 : 1.0));

    clip[0] = 0.1;
    clip[1] = 20.0;
    anArray = NewRealArray(1, 2L);
    CArray2Array(anArray, clip);
    SetVar(observer, VIEWCLIP, anArray);

    rotQuat[0] = -SQ22;
    rotQuat[1] = SQ22;
    rotQuat[2] = 0.0;
    rotQuat[3] = 0.0;

    anArray = NewRealArray(1, 4L);
    CArray2Array(anArray, rotQuat);
    SetVar(observer, ROTQUAT, anArray);

    return ObjTrue;
}

static ObjPtr MakeArenaWand(arena)
ObjPtr arena;
/*Makes an arena's wand*/
{
    ObjPtr wand;

    SetVar(arena, WAND, wand = NewWand("Wand"));
    SetVar(wand, PARENT, wand);

    return ObjTrue;
}

static ObjPtr MakeArenaLights(arena)
ObjPtr arena;
/*Makes the lights of an arena*/
{
    ObjPtr lights;
    ThingListPtr runner;
    int k;

    lights = NewList();
    for (k = 0; k < NARENALIGHTS; ++k)
    {
	ObjPtr light;

	light = NewArenaLight(k);

	/*Set the light's parent*/
	SetVar(light, PARENT, lights);

	PostfixList(lights, light);
    }

    SetVar(arena, LIGHTS, lights);

    return ObjTrue;
}

static ObjPtr KeyDownArena(arena, key, flags)
ObjPtr arena;
int key;
long flags;
/*Key down in an arena*/
{
    ObjPtr var;
    ObjPtr observer;
    real axis[3];
    real phi;
    real mv;
    real eyePosn[3];
    real forward[3], up[3], side[3];
    real clips[2];

    phi = KEYROTFACTOR;
    mv = KEYMOVEFACTOR;

    MakeVar(arena, OBSERVER);
    observer = GetObjectVar("KeyDownArena", arena, OBSERVER);
    if (observer)
    {
	GetObserverLocation(eyePosn, observer);
	GetAdjustedVectors(forward, up, observer);
	CROSS(up, forward, side);

	if (flags & F_OPTIONDOWN)
	{
	    /*Rotation*/
	    switch(key)
	    {
		case FK_LEFT_ARROW:
		    /*Left rotation*/
		    SpinObserver(observer, up, -phi, false);
		    break;
		case FK_RIGHT_ARROW:
		    SpinObserver(observer, up, phi, false);
		    break;
	    }
	}
	else
	{
	    /*It's a move*/
	    switch(key)
	    {
		case FK_LEFT_ARROW:
		    /*Left motion*/
		    eyePosn[0] += side[0] * mv;
		    eyePosn[1] += side[1] * mv;
		    eyePosn[2] += side[2] * mv;
		    var = NewRealArray(1, 3L);
		    CArray2Array(var, eyePosn);
		    SetVar(observer, LOCATION, var);
		    ImInvalid(observer);
		    break;
		case FK_RIGHT_ARROW:
		    eyePosn[0] -= side[0] * mv;
		    eyePosn[1] -= side[1] * mv;
		    eyePosn[2] -= side[2] * mv;
		    var = NewRealArray(1, 3L);
		    CArray2Array(var, eyePosn);
		    SetVar(observer, LOCATION, var);
		    ImInvalid(observer);
		    break;
		case FK_UP_ARROW:
		    /*Forward motion*/
		    eyePosn[0] += forward[0] * mv;
		    eyePosn[1] += forward[1] * mv;
		    eyePosn[2] += forward[2] * mv;
		    var = NewRealArray(1, 3L);
		    CArray2Array(var, eyePosn);
		    SetVar(observer, LOCATION, var);
		    ImInvalid(observer);
		    break;
		case FK_DOWN_ARROW:
		    eyePosn[0] -= forward[0] * mv;
		    eyePosn[1] -= forward[1] * mv;
		    eyePosn[2] -= forward[2] * mv;
		    var = NewRealArray(1, 3L);
		    CArray2Array(var, eyePosn);
		    SetVar(observer, LOCATION, var);
		    ImInvalid(observer);
		    break;
	    }
	}
    }
    return ObjTrue;
}

static ObjPtr TurnDialArena(arena, whichDial, delta, flags)
ObjPtr arena;
int whichDial;
real delta;
long flags;
/*Turns dial whichDial in arena arena by delta*/
{
    ObjPtr var;
    ObjPtr observer;
    real axis[3];
    real phi;
    real mv;
    real eyePosn[3];
    real forward[3], up[3], side[3];
    real clips[2];

    phi = delta * DIALROTFACTOR * M_PI_2 / 5.0;
    mv = delta * DIALMOVEFACTOR * M_PI_2;

    MakeVar(arena, OBSERVER);
    observer = GetObjectVar("TurnDialArena", arena, OBSERVER);
    if (observer)
    {
	GetObserverLocation(eyePosn, observer);
	GetAdjustedVectors(forward, up, observer);
	CROSS(up, forward, side);

	switch (whichDial)
	{
	    case 5:		/*Motion in x direction*/
		if (flags & F_OPTIONDOWN)
		{
		    eyePosn[0] += mv;
		}
		else
		{
		    eyePosn[0] -= side[0] * mv;
		    eyePosn[1] -= side[1] * mv;
		    eyePosn[2] -= side[2] * mv;
		}
		var = NewRealArray(1, 3L);
		CArray2Array(var, eyePosn);
		SetVar(observer, LOCATION, var);
		ImInvalid(observer);
		break;
	    case 6:		/*Spin about y*/
		if (flags & F_OPTIONDOWN)
		{
		    axis[0] = 0.0;
		    axis[1] = 1.0;
		    axis[2] = 0.0;
		    SpinObserver(observer, axis, phi, false);
		}
		else
		{
		    SpinObserver(observer, up, phi, false);
		}
		break;
	    case 3:		/*Motion in y direction*/
		if (flags & F_OPTIONDOWN)
		{
		    eyePosn[1] += mv;
		}
		else
		{
		    eyePosn[0] += up[0] * mv;
		    eyePosn[1] += up[1] * mv;
		    eyePosn[2] += up[2] * mv;
		}
		var = NewRealArray(1, 3L);
		CArray2Array(var, eyePosn);
		SetVar(observer, LOCATION, var);
		ImInvalid(observer);
		break;
	    case 4:		/*Spin to top/bottom*/
		if (flags & F_OPTIONDOWN)
		{
		    axis[0] = -1.0;
		    axis[1] = 0.0;
		    axis[2] = 0.0;
		    SpinObserver(observer, axis, phi, false);
		}
		else
		{
		    SpinObserver(observer, side, phi, false);
		}
		break;
	    case 7:		/*Motion in z direction, or forward*/
		if (flags & F_OPTIONDOWN)
		{
		    eyePosn[2] -= mv;
		}
		else
		{
		    eyePosn[0] += forward[0] * mv;
		    eyePosn[1] += forward[1] * mv;
		    eyePosn[2] += forward[2] * mv;

		    var = GetVar(observer, FOCUSDIST);
		    if (var)
		    {
			SetVar(observer, FOCUSDIST, NewReal(GetReal(var) - mv));
		    }
		    var = GetVar(observer, VIEWCLIP);
		    if (var)
		    {
			Array2CArray(clips, var);
			clips[0] -= mv;
			clips[1] -= mv;
			var = NewRealArray(1, 2L);
			CArray2Array(var, clips);
			SetVar(observer, VIEWCLIP, var);
		    }
		}
		var = NewRealArray(1, 3L);
		CArray2Array(var, eyePosn);
		SetVar(observer, LOCATION, var);
		ImInvalid(observer);
		break;
	    case 2:		/*Spin about z*/
		if (flags & F_OPTIONDOWN)
		{
		    axis[0] = 0.0;
		    axis[1] = 0.0;
		    axis[2] = 1.0;
		    SpinObserver(observer, axis, -phi, false);
		}
		else
		{
		    SpinObserver(observer, forward, phi, false);
		}
		break;
	    case 0:		/*Change clipping plane*/
		var = GetVar(observer, VIEWCLIP);
		if (var)
		{
		    Array2CArray(clips, var);
		    if (flags & F_OPTIONDOWN)
		    {
			clips[1] += mv;
			if (clips[1] < clips[0] + 0.001) clips[1] = clips[0] + 0.001;
		    }
		    else
		    {
			clips[0] += mv;
			if (clips[1] < clips[0] + 0.001) clips[0] = clips[1] - 0.001;
		    }
		    var = NewRealArray(1, 2L);
		    CArray2Array(var, clips);
		    SetVar(observer, VIEWCLIP, var);
		    ImInvalid(observer);
		}
		break;
	    case 1:		/*Change angle or focus width*/
		if (flags & F_OPTIONDOWN)
		{
		    /*Change binocular spacing*/
		    var = GetVar(observer, BINOCULARITY);
		    if (var)
		    {
			real val;
			val = GetReal(var) + delta * DIALBINOCFACTOR;
			if (val < 0.0) val = 0.0;
			SetVar(observer, BINOCULARITY, NewReal(val));
			ImInvalid(observer);
		    }
		}
		else
		{
		    /*Change angle or field width*/
		    Bool isOrtho = false;

		    var = GetVar(observer, VIEWTYPE);
		    if (var)
		    {
			if (GetInt(var) == VT_ORTHOGRAPHIC)
			{
			    isOrtho = true;
			}
		    }

		    var = GetVar(observer, VIEWFIELD);
		    if (var)
		    {
			if (isOrtho)
			{
			    real val;
			    val = GetReal(var) + delta * DIALWIDTHFACTOR;
			    if (val < 0.001) val = 0.001;
			    SetVar(observer, VIEWFIELD, NewReal(val));
			}
			else
			{
			    real val;
			    val = GetReal(var) + delta * DIALANGLEFACTOR * 180.0;
			    if (val < 0.1) val = 0.1;
			    if (val > 179.0) val = 179.0;
			    SetVar(observer, VIEWFIELD, NewReal(val));
			}
			ImInvalid(observer);
		    }
		}
		break;
	}
    }
    return ObjTrue;
}

static MakeMethod MakeArenaAppearance(arena)
ObjPtr arena;
/*Makes an arena's appearance*/
{
    SetVar(arena, APPEARANCE, ObjTrue);
    ImInvalid(arena);
    return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
static void TransformCAVEToArena(ObjPtr observer, real source[3], real dest[3])
#else
static void TransformCAVEToArena(observer, source, dest)
ObjPtr observer;
real source[3];
real dest[3];
#endif
/*Transforms source to dest in the CAVE with respect to observer*/
{
    real temp[3];
    ObjPtr var;
    Quaternion rot;

    temp[0] = source[0];
    temp[1] = source[1];
    temp[2] = source[2];

    temp[1] -= CAVEORIGINOFFSET;

    temp[0] /= CAVESCALE;
    temp[1] /= CAVESCALE;
    temp[2] /= CAVESCALE;

    var = GetVar(observer, ROTQUAT);
    if (!var) return;
    Array2CArray(rot, var);

    QRot(temp, rot, dest);
}

static ObjPtr ArenaWandInteract(arena, intType, loc, dir, flags)
ObjPtr arena;
int intType;
real dir[3];
real loc[3];
long flags;
/*Interacts wand with an arena*/
{
    ObjPtr contents;
    FuncTyp method;
    ObjPtr retVal = NULLOBJ;

    contents = GetVar(arena, CONTENTS);
    if (contents)
    {
	method = GetMethod(contents, WANDINTERACT);
	if (method)
	{
	    retVal = (*method)(contents, intType, loc, dir, flags);
	}
    }

    if (intType == IT_FEEL)
    {
	if (retVal == curRayIntersect)
	{
	    /*Intersection has not changed*/
	    return retVal;
	}

	if (curRayIntersect)
	{
	    method = GetMethod(curRayIntersect, RAYLEAVE);
	    if (method)
	    {
		(*method)(curRayIntersect);
	    }
	}

	curRayIntersect = retVal;
	    
	if (curRayIntersect)
	{
	    method = GetMethod(curRayIntersect, RAYENTER);
	    if (method)
	    {
		(*method)(curRayIntersect);
	    }
	}
    }

    return retVal;
}

#ifdef HAVE_PROTOTYPES
real PointToLineDistance(real p[3], real s[3], real dir[3])
#else
real PointToLineDistance(p, s, dir)
real p[3];
real s[3];
real dir[3];
#endif
/*Returns the distance from point p to the line from s in direction dir*/
{
    real s1, s2, s3, s4;

#define x0	(p[0])
#define y0	(p[1])
#define z0	(p[2])
#define x1	(s[0])
#define y1	(s[1])
#define z1	(s[2])
#define a	(dir[0])
#define b	(dir[1])
#define c	(dir[2])

    s1 = (y0 - y1) * c - (z0 - z1) * b;
    s2 = (z0 - z1) * a - (x0 - x1) * c;
    s3 = (x0 - x1) * b - (y0 - y1) * a;
    s4 = SQUARE(a) + SQUARE(b) + SQUARE(c);
    return rsqrt((SQUARE(s1) + SQUARE(s2) + SQUARE(s3)) / s4);

#undef x0
#undef y0
#undef z0
#undef x1
#undef y1
#undef z1
#undef a
#undef b
#undef c
}

#ifdef HAVE_PROTOTYPES
void IdleWand(ObjPtr wand)
#else
void IdleWand(wand)
ObjPtr wand;
#endif
/*Idles the wand*/
{
    ObjPtr var;
    Quaternion rot;
    real dir[3], loc[3];
    FuncTyp method;

    var = GetVar(wand, LOCATION);
    if (!var) return;
    Array2CArray(loc, var);

    dir[0] = 1.0;
    dir[1] = 0.0;
    dir[2] = 0.0;

    var = GetVar(wand, ROTQUAT);
    if (!var) return;
    Array2CArray(rot, var);

    QRot(dir, rot, dir);

    if (GetPredicate(wand, WANDDOWNP))
    {
	method = GetMethod(wand, STILLDOWN);
	if (method)
	{
	    (*method)(wand);
	}
    }
    else
    {
	method = GetMethod(openArena, WANDINTERACT);
	if (method)
	{
	    (*method)(openArena, IT_FEEL, loc, dir, 0);
	}
    }
}

#ifdef HAVE_PROTOTYPES
Bool DistanceHit(real maxDist, real p[3], real s[3], real dir[3], real *actDist)
#else
Bool DistanceHit(maxDist, p, s, dir)
real maxDist; real p[3]; real s[3]; real dir[3]; real *actDist;
#endif
/*Returns true iff there's a distance hit within maxDist from p to the vector
  from s in the dir direction.  If actDist stuffs the actual distance there.*/
{
    real distance;

    distance = PointToLineDistance(p, s, dir);

    if (actDist) *actDist = distance;
    if (distance < maxDist)
    {
	return true;
    }
    else
    {
	return false;
    }
}

static void AddAWindowToArena(window)
WinInfoPtr window;
/*Adds a window to the arena, glue for AddWindowToArena to be compatible
  with ForAllWindows*/
{
    if (window != curArenaWindow)
    {
	AddWindowToArena(window);
    }
}

#ifdef HAVE_PROTOTYPES
void IdleArena(void)
#else
void IdleArena()
#endif
/*Idles the arena*/
{
    if (openArena)
    {
	ObjPtr wand;

	ForAllWindows(AddAWindowToArena);

	wand = GetVar(openArena, WAND);

	if (wand)
	{
	    IdleWand(wand);
	}
    }
}

#ifdef HAVE_PROTOTYPES
void CurArenaWandUp(int whichButton, long flags)
#else
void CurArenaWandUp(whichButton, flags)
int whichButton, long flags;
#endif
/*Button up in the current arena*/
{
    ObjPtr wand;
    FuncTyp method;

    if (openArena)
    {
	wand = GetVar(openArena, WAND);

	if (!wand) return;
	{
	    switch (whichButton)
	    {
		case WB_LEFT:
		    method = GetMethod(wand, LEFTWANDUP);
		    if (method)
		    {
			ObjPtr retVal;
			retVal = (*method)(wand, flags);
		    }
		    break;
		case WB_MIDDLE:
		    method = GetMethod(wand, MIDDLEWANDUP);
		    if (method)
		    {
			ObjPtr retVal;
			retVal = (*method)(wand, flags);
		    }
		    break;
		case WB_RIGHT:
		    method = GetMethod(wand, STOPMENU);
		    if (method)
		    {
			ObjPtr retVal;
			retVal = (*method)(wand);
		    }
		    break;
	    }
	    SetVar(wand, WANDDOWNP, ObjFalse);
	}
    }
}

#ifdef HAVE_PROTOTYPES
void CurArenaWandDown(int whichButton, long flags)
#else
void CurArenaWandDown(whichButton, flags)
int whichButton, long flags;
#endif
/*Button down in the current arena*/
{
    FuncTyp method;

    if (openArena)
    {
	ObjPtr wand;

	wand = GetVar(openArena, WAND);

	if (wand)
	{
	
	    SetVar(wand, WANDDOWNP, ObjTrue);
	
	    switch (whichButton)
	    {
		case WB_LEFT:
		    method = GetMethod(wand, LEFTWANDDOWN);
		    if (method)
		    {
			ObjPtr retVal;
			retVal = (*method)(wand, flags);
		    }
		    break;
		case WB_MIDDLE:
		    method = GetMethod(wand, MIDDLEWANDDOWN);
		    if (method)
		    {
			ObjPtr retVal;
			retVal = (*method)(wand, flags);
		    }
		    break;
		case WB_RIGHT:
		    method = GetMethod(wand, STARTMENU);
		    if (method)
		    {
			ObjPtr retVal;
			retVal = (*method)(wand);
		    }
		    break;
	    }
	}
    }
}

static ObjPtr SpaceMoveToolMiddleWandDown(tool, wand)
ObjPtr tool;
ObjPtr wand;
/*Middle wand down with a space move tool*/
{
    ObjPtr var, space, observer;
    ThingListPtr runner;

    var = GetVar(wand, LOCATION);
    SetVar(tool, STARTSCALE, var);

    var = GetVar(openArena, CONTENTS);
    if (var)
    {
	/*Go through and set the STARTSCALE of all the spaces*/
	for (runner = LISTOF(var); runner; runner = runner -> next)
	{
	    ObjPtr contents2;
	    ThingListPtr runner2;
	    contents2 = GetVar(runner -> thing, CONTENTS);
	    if (contents2)
	    {
		for (runner2 = LISTOF(contents2); runner2; runner2 = runner2 -> next)
		{
		    space = GetVar(runner2 -> thing, SPACE);
		    if (space)
		    {
			observer = GetVar(space, OBSERVER);
			if (observer)
			{
			    SetVar(space, STARTSCALE, GetVar(observer, SPACESCALE));
			    SetVar(observer, AUTOADJUST, ObjFalse);
			}
		    }
		}
	    }
	}
    }

    return ObjTrue;
}

static ObjPtr SpaceMoveToolMiddleWandUp(tool, wand)
ObjPtr tool;
ObjPtr wand;
/*Middle wand up with a space move tool*/
{
    ThingListPtr runner;
    ObjPtr var;
    real wandLoc[3], startScale[3];
    real spaceScale[3];
    real scaleRatio;
    ObjPtr space, observer;

    var = GetVar(tool, STARTSCALE);
    if (var)
    {
	/*It's a location change*/
	Array2CArray(startScale, var);
    
	var = GetVar(wand, LOCATION);
	if (!var) return ObjFalse;
	Array2CArray(wandLoc, var);
 
	scaleRatio = rexp(ARENASCALEPOWRATIO * -(wandLoc[2] - startScale[2]));

	var = GetVar(openArena, CONTENTS);
	if (var)
	{
	    /*Go through and set the STARTLOC of all the spaces*/
	    for (runner = LISTOF(var); runner; runner = runner -> next)
	    {
		ObjPtr contents2;
		ThingListPtr runner2;
		contents2 = GetVar(runner -> thing, CONTENTS);
		if (contents2)
		{
		    for (runner2 = LISTOF(contents2); runner2; runner2 = runner2 -> next)
		    {
			space = GetVar(runner2 -> thing, SPACE);
			if (space)
			{
			    observer = GetVar(space, OBSERVER);
			    if (observer)
			    {
				var = GetVar(space, STARTSCALE);
				if (var)
				{
				    Array2CArray(spaceScale, var);
				    spaceScale[0] *= scaleRatio;
				    spaceScale[1] *= scaleRatio;
				    spaceScale[2] *= scaleRatio;
				}
				var = NewRealArray(1, 3L);
				CArray2Array(var, spaceScale);
				SetVar(observer, SPACESCALE, var);
				ImInvalid(space);
			    }
			}
		    }
		}
	    }
	}
    }

    SetVar(tool, STARTSCALE, NULLOBJ);

    return ObjTrue;
}

static ObjPtr SpaceMoveToolLeftWandDown(tool, wand, flags)
ObjPtr tool;
ObjPtr wand;
long flags;
/*Left wand down with a space move tool*/
{
    ObjPtr var, space;
    ThingListPtr runner;

    var = GetVar(wand, LOCATION);
    SetVar(tool, STARTLOC, var);

    var = GetVar(openArena, CONTENTS);
    if (var)
    {
	/*Go through and set the STARTLOC of all the spaces*/
	for (runner = LISTOF(var); runner; runner = runner -> next)
	{
	    ObjPtr contents2;
	    ThingListPtr runner2;
	    contents2 = GetVar(runner -> thing, CONTENTS);
	    if (contents2)
	    {
		for (runner2 = LISTOF(contents2); runner2; runner2 = runner2 -> next)
		{
		    space = GetVar(runner2 -> thing, SPACE);
		    if (space)
		    {
			SetVar(space, STARTLOC, GetVar(space, LOCATION));
		    }
		}
	    }
	}
    }

    return ObjTrue;
}

static ObjPtr SpaceMoveToolLeftWandUp(tool, wand, flags)
ObjPtr tool;
ObjPtr wand;
long flags;
/*Left wand up with a space move tool*/
{
    ThingListPtr runner;
    ObjPtr var;
    real wandLoc[3], startLoc[3];
    real spaceLoc[3];
    ObjPtr space;

    var = GetVar(tool, STARTLOC);
    if (var)
    {
	/*It's a location change*/
	Array2CArray(startLoc, var);
    
	var = GetVar(wand, LOCATION);
	if (!var) return ObjFalse;
	Array2CArray(wandLoc, var);
    
	var = GetVar(openArena, CONTENTS);
	if (var)
	{
	    /*Go through and set the STARTLOC of all the spaces*/
	    for (runner = LISTOF(var); runner; runner = runner -> next)
	    {
		ObjPtr contents2;
		ThingListPtr runner2;
		contents2 = GetVar(runner -> thing, CONTENTS);
		if (contents2)
		{
		    for (runner2 = LISTOF(contents2); runner2; runner2 = runner2 -> next)
		    {
			space = GetVar(runner2 -> thing, SPACE);
			if (space)
			{
			    var = GetVar(space, STARTLOC);
			    if (var)
			    {
				Array2CArray(spaceLoc, var);
				spaceLoc[0] += wandLoc[0] - startLoc[0];
				spaceLoc[1] += wandLoc[1] - startLoc[1];
				spaceLoc[2] += wandLoc[2] - startLoc[2];
			    }
			    var = NewRealArray(1, 3L);
			    CArray2Array(var, spaceLoc);
			    SetVar(space, LOCATION, var);
			    ImInvalid(space);
			}
		    }
		}
	    }
	}
	SetVar(tool, STARTLOC, NULLOBJ);
    }

    return ObjTrue;
}

static ObjPtr SpaceMoveToolStillDown(tool, wand)
ObjPtr tool;
ObjPtr wand;
/*Still down with a space move tool*/
{
    ThingListPtr runner;
    ObjPtr var;
    real wandLoc[3], startLoc[3], startScale[3];
    real spaceLoc[3], spaceScale[3], scaleRatio;
    ObjPtr space, observer;

    var = GetVar(tool, STARTLOC);
    if (var)
    {
	/*It's a location change*/
	Array2CArray(startLoc, var);
    
	var = GetVar(wand, LOCATION);
	if (!var) return ObjFalse;
	Array2CArray(wandLoc, var);
    
	var = GetVar(openArena, CONTENTS);
	if (var)
	{
	    /*Go through and set the STARTLOC of all the spaces*/
	    for (runner = LISTOF(var); runner; runner = runner -> next)
	    {
		ObjPtr contents2;
		ThingListPtr runner2;
		contents2 = GetVar(runner -> thing, CONTENTS);
		if (contents2)
		{
		    for (runner2 = LISTOF(contents2); runner2; runner2 = runner2 -> next)
		    {
			space = GetVar(runner2 -> thing, SPACE);
			if (space)
			{
			    var = GetVar(space, STARTLOC);
			    if (var)
			    {
				Array2CArray(spaceLoc, var);
				spaceLoc[0] += wandLoc[0] - startLoc[0];
				spaceLoc[1] += wandLoc[1] - startLoc[1];
				spaceLoc[2] += wandLoc[2] - startLoc[2];
			    }
			    var = NewRealArray(1, 3L);
			    CArray2Array(var, spaceLoc);
			    SetVar(space, LOCATION, var);
			    ImInvalid(space);
			}
		    }
		}
	    }
	}
    }
    else
    {
	var = GetVar(tool, STARTSCALE);
	if (var)
	{
	    /*It's a scaling change*/
	    Array2CArray(startScale, var);
	
	    var = GetVar(wand, LOCATION);
	    if (!var) return ObjFalse;
	    Array2CArray(wandLoc, var);
     
	    scaleRatio = rexp(ARENASCALEPOWRATIO * -(wandLoc[2] - startScale[2]));
    
	    var = GetVar(openArena, CONTENTS);
	    if (var)
	    {
		/*Go through and set the STARTLOC of all the spaces*/
		for (runner = LISTOF(var); runner; runner = runner -> next)
		{
		    ObjPtr contents2;
		    ThingListPtr runner2;
		    contents2 = GetVar(runner -> thing, CONTENTS);
		    if (contents2)
		    {
			for (runner2 = LISTOF(contents2); runner2; runner2 = runner2 -> next)
			{
			    space = GetVar(runner2 -> thing, SPACE);
			    if (space)
			    {
				observer = GetVar(space, OBSERVER);
				if (observer)
				{
				    var = GetVar(space, STARTSCALE);
				    if (var)
				    {
					Array2CArray(spaceScale, var);
					spaceScale[0] *= scaleRatio;
					spaceScale[1] *= scaleRatio;
					spaceScale[2] *= scaleRatio;
				    }
				    var = NewRealArray(1, 3L);
				    CArray2Array(var, spaceScale);
				    SetVar(observer, SPACESCALE, var);
				    ImInvalid(space);
				}
			    }
			}
		    }
		}
	    }
	}
    }

    return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
static void EmitWindowStream(ObjPtr window, real p[3])
#else
static void EmitWindowStream(window, p)
ObjPtr window;
real p[3];
#endif
/*Emits a stream into a window*/
{
    ObjPtr space;
    ObjPtr contents;
    ThingListPtr runner;
    FuncTyp method;
    ObjPtr emitter;

    space = GetVar(window, SPACE);
    if (space)
    {
	/*Transform the point into the space*/
	ObjPtr observer;
	ObjPtr var;
	real loc[3];
	real scale[3];
	real newP[3];
	Quaternion rot;

	newP[0] = p[0];
	newP[1] = p[1];
	newP[2] = p[2];

	observer = GetObjectVar("EmitWindowStream", space, OBSERVER);
	if (!observer) return;

	var = GetFixedArrayVar("EmitWindowStream", space, LOCATION, 1, 3L);
	if (!var) return;
	Array2CArray(loc, var);

	var = GetFixedArrayVar("EmitWindowStream", space, ROTQUAT, 1, 4L);
	if (!var) return;
	Array2CArray(rot, var);

	newP[0] -= loc[0];
	newP[1] -= loc[1];
	newP[2] -= loc[2];

/*DIKEO maybe its inverse?*/
	QRot(newP, rot, newP);

	var = GetFixedArrayVar("EmitWindowStream", observer, SPACESCALE, 1, 3L);
	if (!var) return;
	Array2CArray(scale, var);

	newP[0] *= scale[0];
	newP[1] *= scale[1];
	newP[2] *= scale[2];

	var = GetFixedArrayVar("EmitWindowStream", observer, SPACEORIGIN, 1, 3L);
	if (!var) return;
	Array2CArray(loc, var);

	newP[0] += loc[0];
	newP[1] += loc[1];
	newP[2] += loc[2];

	printf("newP = %g %g %g\n", newP[0], newP[1], newP[2]);
	contents = GetVar(space, CONTENTS);
	if (contents)
	{
	    for (runner = LISTOF(contents); runner; runner = runner -> next)
	    {
		method = GetMethod(runner -> thing, ADDEMITTERS);
		if (method)
		{
		    printf("Adding an emitter\n");
		    emitter = NewEmitter(newP);
		    (*method)(runner -> thing, emitter);
		}
	    }
	}
    }
}

#ifdef HAVE_PROTOTYPES
static void EmitWallStream(ObjPtr wall, real p[3])
#else
static void EmitWallStream(wall, p)
ObjPtr wall;
real p[3];
#endif
/*Emits a stream into a wall*/
{
    ObjPtr contents;
    ThingListPtr runner;

    if (wall)
    {
	contents = GetVar(wall, CONTENTS);
	if (contents)
	{
	    for (runner = LISTOF(contents); runner; runner = runner -> next)
	    {
		EmitWindowStream(runner -> thing, p);
	    }
	}
    }
}

#ifdef HAVE_PROTOTYPES
static void EmitArenaStream(real p[3])
#else
static void EmitArenaStream(p)
real p[3];
#endif
/*Emits a stream into an arena*/
{
    ObjPtr contents;
    ThingListPtr runner;

    if (openArena)
    {
	contents = GetVar(openArena, CONTENTS);
	if (contents)
	{
	    for (runner = LISTOF(contents); runner; runner = runner -> next)
	    {
		EmitWallStream(runner -> thing, p);
	    }
	}
    }
}

static ObjPtr RakeToolLeftWandDown(tool, wand, flags)
ObjPtr tool;
ObjPtr wand;
long flags;
/*Left wand down in a rake tool*/
{
    ObjPtr var;
    Quaternion rot;
    real dir[3], loc[3], p1[3], p2[3], p3[3];
    FuncTyp method;
    ObjPtr sound;

    if (openArena)
    {
	if (wand)
	{
	    var = GetVar(wand, LOCATION);
	    if (!var) return;
	
	    Array2CArray(loc, var);
	
	    var = GetVar(wand, ROTQUAT);
	    if (!var) return;
	    Array2CArray(rot, var);
	
	    dir[0] = 0.8;
	    dir[1] = -0.1;
	    dir[2] = 0.0;
	
	    QRot(dir, rot, p1);

	    p1[0] += loc[0];
	    p1[1] += loc[1];
	    p1[2] += loc[2];

	    dir[0] = 0.8;
	    dir[1] = 0.0;
	    dir[2] = 0.0;
	
	    QRot(dir, rot, p2);

	    p2[0] += loc[0];
	    p2[1] += loc[1];
	    p2[2] += loc[2];
	
	    dir[0] = 0.8;
	    dir[1] = 0.1;
	    dir[2] = 0.0;
	
	    QRot(dir, rot, p3);

	    p3[0] += loc[0];
	    p3[1] += loc[1];
	    p3[2] += loc[2];

#if 0
#ifdef CAVESOUND
	    sound = NewSound("pssh.aiff");
	    Play(sound);
#endif
#endif

	    EmitArenaStream(p1);
	    EmitArenaStream(p2);
	    EmitArenaStream(p3);
	}
    }

    return ObjTrue;
}

static ObjPtr FingerToolLeftWandDown(tool, wand, flags)
ObjPtr tool;
ObjPtr wand;
long flags;
/*Left wand down in a finger tool*/
{
    ObjPtr var;
    Quaternion rot;
    real dir[3], loc[3];
    FuncTyp method;

    if (openArena)
    {
	if (wand)
	{
	    var = GetVar(wand, LOCATION);
	    if (!var) return;
	
	    Array2CArray(loc, var);
	
	    dir[0] = 1.0;
	    dir[1] = 0.0;
	    dir[2] = 0.0;
	
	    var = GetVar(wand, ROTQUAT);
	    if (!var) return;
	    Array2CArray(rot, var);
	
	    QRot(dir, rot, dir);

	    method = GetMethod(openArena, WANDINTERACT);
	    if (method)
	    {
		ObjPtr retVal;
		retVal = (*method)(openArena, IT_SELECTDOWN, loc, dir, flags);
	
		if (!retVal && !(flags & F_EXTEND)) DeselectAll();
	    }
	}
    }

    return ObjTrue;
}

static ObjPtr FingerToolLeftWandUp(tool, wand, flags)
ObjPtr tool;
ObjPtr wand;
long flags;
/*Left wand down in a finger tool*/
{
    ObjPtr var;
    Quaternion rot;
    real dir[3], loc[3];
    FuncTyp method;

    if (openArena)
    {
	if (wand)
	{
	    var = GetVar(wand, LOCATION);
	    if (!var) return;
	
	    Array2CArray(loc, var);
	
	    dir[0] = 1.0;
	    dir[1] = 0.0;
	    dir[2] = 0.0;
	
	    var = GetVar(wand, ROTQUAT);
	    if (!var) return;
	    Array2CArray(rot, var);
	
	    QRot(dir, rot, dir);
	
	    method = GetMethod(openArena, WANDINTERACT);
	    if (method)
	    {
		ObjPtr retVal;
		retVal = (*method)(openArena, IT_SELECTUP, loc, dir, flags);
	    }
	}
    }

    return ObjTrue;
}
static ObjPtr FingerToolStillDown(tool, wand, flags)
ObjPtr tool;
ObjPtr wand;
long flags;
/*Button still down in a finger tool*/
{
    ObjPtr var;
    Quaternion rot;
    real dir[3], loc[3];
    FuncTyp method;

    if (openArena)
    {
	if (wand)
	{
	    var = GetVar(wand, LOCATION);
	    if (!var) return;
	
	    Array2CArray(loc, var);
	
	    dir[0] = 1.0;
	    dir[1] = 0.0;
	    dir[2] = 0.0;
	
	    var = GetVar(wand, ROTQUAT);
	    if (!var) return;
	    Array2CArray(rot, var);
	
	    QRot(dir, rot, dir);
	
	    method = GetMethod(openArena, WANDINTERACT);
	    if (method)
	    {
		ObjPtr retVal;
		retVal = (*method)(openArena, IT_STILLDOWN, loc, dir, flags);
	    }
	}
    }

    return ObjTrue;
}


static ObjPtr WandLeftWandDown(wand, flags)
ObjPtr wand;
long flags;
/*Left wand down in a wand*/
{
    ObjPtr tool;
    FuncTyp method;

    tool = GetVar(wand, EDITTOOL);
    if (tool)
    {
	method = GetMethod(tool, LEFTWANDDOWN);
	if (method)
	{
	    (*method)(tool, wand, flags);
	}
    }

    return ObjTrue;
}

static ObjPtr WandLeftWandUp(wand, flags)
ObjPtr wand;
long flags;
/*Left wand down in a wand*/
{
    ObjPtr tool;
    FuncTyp method;

    tool = GetVar(wand, EDITTOOL);
    if (tool)
    {
	method = GetMethod(tool, LEFTWANDUP);
	if (method)
	{
	    (*method)(tool, wand, flags);
	}
    }

    return ObjTrue;
}

static ObjPtr WandMiddleWandDown(wand, flags)
ObjPtr wand;
long flags;
/*Middle wand down in a wand*/
{
    ObjPtr tool;
    FuncTyp method;

    tool = GetVar(wand, EDITTOOL);
    if (tool)
    {
	method = GetMethod(tool, MIDDLEWANDDOWN);
	if (method)
	{
	    (*method)(tool, wand, flags);
	}
	else
	{
	    method = GetMethod(tool, LEFTWANDDOWN);
	    if (method)
	    {
		(*method)(tool, wand, flags | F_EXTEND);
	    }
	}
    }

    return ObjTrue;
}

static ObjPtr WandMiddleWandUp(wand, flags)
ObjPtr wand;
long flags;
/*Middle wand down in a wand*/
{
    ObjPtr tool;
    FuncTyp method;

    tool = GetVar(wand, EDITTOOL);
    if (tool)
    {
	method = GetMethod(tool, MIDDLEWANDUP);
	if (method)
	{
	    (*method)(tool, wand, flags);
	}
	else
	{
	    method = GetMethod(tool, LEFTWANDUP);
	    if (method)
	    {
		(*method)(tool, wand, flags | F_EXTEND);
	    }
	}
    }

    return ObjTrue;
}

static ObjPtr MoveArenaWandTo(arena, x, y, z)
ObjPtr arena;
real x, y, z;
/*Moves the wand in the arena to x, y, and z, which are given in global coords*/
{
    ObjPtr observer;
    ObjPtr var;
    ObjPtr wand;
    real xyz[3];
    real xyzXform[3];

    MakeVar(arena, WAND);
    wand = GetObjectVar("MoveArenaWandTo", arena, WAND);
    if (!wand) return ObjFalse;

    MakeVar(arena, OBSERVER);
    observer = GetObjectVar("MoveArenaWandTo", arena, OBSERVER);
    if (!observer) return ObjFalse;

    xyz[0] = x;
    xyz[1] = y;
    xyz[2] = z;

    TransformCAVEToArena(observer, xyz, xyzXform);

    var = NewRealArray(1, 3L);
    CArray2Array(var, xyzXform);
    SetVar(wand, LOCATION, var);
    ImInvalid(wand);
    ImInvalid(arena);

    return ObjTrue;
}

static ObjPtr RotateArenaWandTo(arena, q0, q1, q2, q3)
ObjPtr arena;
real q0, q1, q2, q3;
/*Rotates the wand to a new quaternion orientation*/
{
    Quaternion qw, qr, qi, qa;
    ObjPtr var;
    ObjPtr wand;
    ObjPtr observer;

    MakeVar(arena, WAND);
    wand = GetObjectVar("RotateArenaWandTo", arena, WAND);
    if (!wand) return ObjFalse;

    qw[0] = q0;
    qw[1] = q1;
    qw[2] = q2;
    qw[3] = q3;

    MakeVar(arena, OBSERVER);
    observer = GetVar(arena, OBSERVER);
    if (!observer) return ObjFalse;

    /*Rotate quaternion by observer quaternion*/
    var = GetVar(observer, ROTQUAT);
    if (!var) return ObjFalse;

    Array2CArray(qr, var);
    QInvert(qr, qi);
    QMult(qw, qr, qa);

    var = NewRealArray(1, 4L);
    CArray2Array(var, qa);
    SetVar(wand, ROTQUAT, var);
    ImInvalid(wand);
    ImInvalid(arena);

    return ObjTrue;
}

static ObjPtr DrawFingerWandTool(tool)
ObjPtr tool;
/*Draws the finger wand tool*/
{
    ObjPtr var;

#ifdef GRAPHICS
    var = GetVar(tool, WANDTOOLPICTURE);
    if (var)
    {
	float c[3];

	c[0] = 1.0;
	c[1] = 1.0;
	c[2] = 0.0;
	c3f(c);
#if 0
    DrawSpaceLine(0.0, 0.0, 0.0, 2.0, 0.0, 0.0);
#endif
	callobj(GetInt(var));
    }
#endif

    return ObjTrue;
}

static ObjPtr DrawRakeWandTool(tool)
ObjPtr tool;
/*Draws the rake wand tool*/
{
    ObjPtr var;

#ifdef GRAPHICS
    var = GetVar(tool, WANDTOOLPICTURE);
    if (var)
    {
	callobj(GetInt(var));
    }
#endif

    return ObjTrue;
}
static ObjPtr DrawArenaWand(wand)
ObjPtr wand;
/*Draws a wand in an arena*/
{
    ObjPtr var;
    real loc[3];
    real end[3];
    real m[3][3];
    float v[3];
    Matrix xform;
    int i, j, k;
    Quaternion rot, inv;
    FuncTyp method;
    ObjPtr tool;
    ObjPtr allTools;
    int nRows, nColumns;
    int whichTool, nTools;
    ThingListPtr runner;

    if (GetPredicate(wand, MENUONP))
    {
#ifdef GRAPHICS
	Object o;

	lmcolor(LMC_COLOR);
	pushmatrix();
    
	var = GetVar(wand, MENULOCATION);
	if (!var) return ObjFalse;
	Array2CArray(loc, var);
    
	var = GetVar(wand, MENUROT);
	if (!var) return ObjFalse;
	Array2CArray(rot, var);

	translate(loc[0], loc[1], loc[2]);
	QInvert(rot, inv);
	QuaternionToMatrix(inv, m);
	MATCOPY(xform, Identity);
	for (i = 0; i < 3; ++i)
	{
	    for (j = 0; j < 3; ++j)
	    {
		xform[i][j] = m[i][j];
	    }
	}
	multmatrix(xform);

	/*Now we're set up, actually draw the menu*/

	allTools = GetListVar("DrawArenaWand", wand, ALLTOOLS);
	if (!allTools) return ObjFalse;
	tool = GetObjectVar("DrawArenaWand", wand, EDITTOOL);
	if (!tool) return ObjFalse;

	var = GetVar(wand, WHICHITEM);
	if (var)
	{
	    whichTool = GetInt(var);
	}
	else
	{
	    whichTool = -1;
	}

	nTools = ListCount(allTools);

	nRows = 1 + ((nTools - 1) / ARENAMENUCOLUMNS);

        if (nTools >= 3)
	{
	    nColumns = 3;
	}
	else
	{
	    nColumns = nTools;
	}

	/*Draw the background rectangle*/

	BeginMask(true, true, true, false);

	SetUIColor(UIWHITE);

	bgnpolygon();
	v[0] = 0.0;
	v[1] = 0.0;
	v[2] = 0.0;
	v3f(v);
	v[1] = -nRows * ARENAMENUSIZE;
	v3f(v);
	v[0] = nColumns * ARENAMENUSIZE;
	v3f(v);
	v[1] = 0.0;
	v3f(v);
	endpolygon();

	/*Draw the tool stuff*/
	runner = LISTOF(allTools);
	for (j = 0; j < nRows; ++j)
	{
	    for (i = 0; i < ARENAMENUCOLUMNS; ++i)
	    {
		if (runner)
		{
		    pushmatrix();
		    translate(i * ARENAMENUSIZE + ARENAMENUSIZE * 0.5,
			      -j * ARENAMENUSIZE - ARENAMENUSIZE * 0.5,
			      0.0);
		    scale(ARENAMENUSIZE, ARENAMENUSIZE, ARENAMENUSIZE);
		    if (i + j * ARENAMENUCOLUMNS == whichTool)
		    {
			/*Draw yellow background*/
			SetUIColor(UIYELLOW);
			bgnpolygon();
			v[0] = -0.5;
			v[1] = -0.5;
			v[2] = 0.0;
			v3f(v);
			v[0] = 0.5;
			v3f(v);
			v[1] = 0.5;
			v3f(v);
			v[0] = -0.5;
			v3f(v);
			endpolygon();
		    }
		    var = GetVar(runner -> thing, CACHEGRAPHICS);
		    if (var)
		    {
			o = GetInt(var);
			SetUIColor(UIRED);
			callobj(o);
		    }
		    popmatrix();
		    runner = runner -> next;
		}
	    }
	}
    
	EndMask();
	BeginMask(false, false, false, true);

	bgnpolygon();
	v[0] = 0.0;
	v[1] = 0.0;
	v[2] = 0.0;
	v3f(v);
	v[1] = -nRows * ARENAMENUSIZE;
	v3f(v);
	v[0] = nColumns * ARENAMENUSIZE;
	v3f(v);
	v[1] = 0.0;
	v3f(v);
	endpolygon();

	EndMask();

	popmatrix();
	lmcolor(LMC_DIFFUSE);
#endif
    }
    else
    {
	tool = GetVar(wand, EDITTOOL);
	if (!tool) return ObjTrue;
    
#ifdef GRAPHICS
	pushmatrix();
    
	var = GetVar(wand, LOCATION);
	if (!var) return ObjFalse;
	Array2CArray(loc, var);
	
	var = GetVar(wand, ROTQUAT);
	if (!var) return ObjFalse;
	Array2CArray(rot, var);

	translate(loc[0], loc[1], loc[2]);
	QInvert(rot, inv);
	QuaternionToMatrix(inv, m);
	MATCOPY(xform, Identity);
	for (i = 0; i < 3; ++i)
	{
	    for (j = 0; j < 3; ++j)
	    {
		xform[i][j] = m[i][j];
	    }
	}
	multmatrix(xform);
    
	method = GetMethod(tool, DRAWARENA);
	if (method)
	{
	    (*method)(tool);
	}
    
	popmatrix();
#endif
    }

    return ObjTrue;
}

static ObjPtr WandStillDown(wand)
ObjPtr wand;
/*True iff a wand is still down*/
{
    ObjPtr var;

    if (GetPredicate(wand, MENUONP))
    {
	Quaternion menuRot, inv;
	int row, column, whichItem;
	real menuLoc[3];
	real wandLoc[3];

	var = GetVar(wand, MENULOCATION);
	if (!var)
	{
	    return ObjFalse;
	}
	Array2CArray(menuLoc, var);

	var = GetVar(wand, LOCATION);
	if (!var)
	{
	    return ObjFalse;
	}
	Array2CArray(wandLoc, var);

	var = GetVar(wand, MENUROT);
	if (!var)
	{
	    return ObjFalse;
	}
	Array2CArray(menuRot, var);
	QInvert(menuRot, inv);

	wandLoc[0] -= menuLoc[0];
	wandLoc[1] -= menuLoc[1];
	wandLoc[2] -= menuLoc[2];

	QRot(wandLoc, inv, wandLoc);

	/*Calculate which menu item*/
	column = rfloor(wandLoc[0] / ARENAMENUSIZE);
	row = rfloor(-wandLoc[1] / ARENAMENUSIZE);

	if (row < 0 || row >= ARENAMENUCOLUMNS || column < 0)
	{
	    whichItem = -1;
	}
	else
	{
	    whichItem = column + row * ARENAMENUCOLUMNS;
	    var = GetVar(wand, ALLTOOLS);
	    if (var && whichItem >= ListCount(var))
	    {
		whichItem = -1;
	    }
	}

	SetVar(wand, WHICHITEM, NewInt(whichItem));
    }
    else
    {
	FuncTyp method;
	ObjPtr tool;

	tool = GetVar(wand, EDITTOOL);
	if (tool)
	{
	    method = GetMethod(tool, STILLDOWN);
	    if (method)
	    {
		return (*method)(tool, wand);
	    }
	}
    }

    return ObjTrue;
}

static ObjPtr StartWandMenu(wand)
ObjPtr wand;
/*Starts a wand menu*/
{
    ObjPtr allTools;
    ObjPtr tool;
    ObjPtr var;
    int whichTool;
    int nTools;
    real dir[3], loc[3], temp[3];
    real menuLoc[3];
    Quaternion rot;
    Quaternion menuRot, inv, tq;
    int row, column;
    FuncTyp method;

    allTools = GetListVar("StartWandMenu", wand, ALLTOOLS);
    if (!allTools) return ObjFalse;
    tool = GetObjectVar("StartWandMenu", wand, EDITTOOL);
    if (!tool) return ObjFalse;

    nTools = ListCount(allTools);
    whichTool = WhichListIndex(allTools, tool);
    SetVar(wand, WHICHITEM, NewInt(whichTool));

    column = whichTool % ARENAMENUCOLUMNS;
    row = (whichTool - column) / ARENAMENUCOLUMNS;

    var = GetVar(wand, LOCATION);
    if (!var) return;
    Array2CArray(loc, var);

    var = GetVar(wand, ROTQUAT);
    if (!var) return;
    Array2CArray(rot, var);
    QInvert(rot, inv);
    tq[0] = -.5;
    tq[1] = .5;
    tq[2] = -.5;
    tq[3] = -.5;
    QMult(tq, rot, menuRot);
    var = NewRealArray(1, 4L);
    CArray2Array(var, menuRot);
    SetVar(wand, MENUROT, var);

    menuLoc[0] = loc[0];
    menuLoc[1] = loc[1];
    menuLoc[2] = loc[2];

    dir[0] = ARENAMENUDISTANCE;
    dir[1] = ARENAMENUSIZE * column + ARENAMENUSIZE * 0.5;
    dir[2] = -ARENAMENUSIZE * row + ARENAMENUSIZE * 0.5;

    QRot(dir, rot, temp);
    menuLoc[0] += temp[0];
    menuLoc[1] += temp[1];
    menuLoc[2] += temp[2];
    
    var = NewRealArray(1, 3L);
    CArray2Array(var, menuLoc);
    SetVar(wand, MENULOCATION, var);

    SetVar(wand, MENUONP, ObjTrue);

    ImInvalid(wand);

    return ObjTrue;
}

static ObjPtr StopWandMenu(wand)
ObjPtr wand;
/*Stops a wand menu*/
{
    ObjPtr var, tool;
    int whichTool;

    var = GetVar(wand, WHICHITEM);
    if (var)
    {
	whichTool = GetInt(var);

	var = GetVar(wand, ALLTOOLS);
	if (var)
	{
	    tool = GetListElem(var, whichTool);
	    if (tool)
	    {
		SetVar(wand, EDITTOOL, tool);
		ImInvalid(wand);
	    }
	}
    }

    SetVar(wand, MENUONP, ObjFalse);
    return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
void InitArenas(void)
#else
void InitArenas()
#endif
{
    real loc[3];
    Quaternion quat;
    ObjPtr var;
    ObjPtr tool;
    ObjPtr list;
    int matIndex;
#ifdef GRAPHICS
    short c[3];
    float n[3];
    float v[3];
    Object o;
#endif

    wallClass = NewObject(NULLOBJ, 0L);
    DefineFixedClass(wallClass, CLASS_WALL, "wall");
    AddToReferenceList(wallClass);
    loc[0] = loc[1] = loc[2];
    var = NewRealArray(1, 3L);
    CArray2Array(var, loc);
    SetVar(wallClass, ARENALOCATION, var);
    quat[0] = 1.0;
    quat[1] = 0.0;
    quat[2] = 0.0;
    quat[3] = 0.0;
    var = NewRealArray(1, 4L);
    CArray2Array(var, quat);
    SetVar(wallClass, ARENAROTATION, var);
    SetMethod(wallClass, APPEARANCE, MakeWallAppearance);
    SetMethod(wallClass, DRAWARENA, DrawWallInArena);
    SetMethod(wallClass, DRAWARENACONTENTS, DrawWallContentsInArena);
    SetMethod(wallClass, WANDINTERACT, WallWandInteract);

    wandToolClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(wandToolClass);

    fingerWandToolClass = NewObject(wandToolClass, 0L);
    AddToReferenceList(fingerWandToolClass);
    DefineFixedClass(fingerWandToolClass, CLASS_FINGERWANDTOOL, "fingerWandTool");
    SetMethod(fingerWandToolClass, LEFTWANDDOWN, FingerToolLeftWandDown);
    SetMethod(fingerWandToolClass, LEFTWANDUP, FingerToolLeftWandUp);
    SetMethod(fingerWandToolClass, STILLDOWN, FingerToolStillDown);
    SetMethod(fingerWandToolClass, DRAWARENA, DrawFingerWandTool);
#ifdef GRAPHICS
    o = genobj();
    makeobj(o);
    bgnpolygon();
    v[0] = -0.300000;
    v[1] = -0.500000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.300000;
    v[1] = -0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.500000;
    v[1] = 0.500000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.300000;
    v[1] = 0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.100000;
    v[1] = 0.100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.500000;
    v[1] = -0.300000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    closeobj();

    SetVar(fingerWandToolClass, CACHEGRAPHICS, NewInt(o));
#endif
#ifdef GRAPHICS
    o = genobj();
    makeobj(o);
    bgnpolygon();
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 3.00000;
    v[1] = 0.00382683;
    v[2] = 0.00923879;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00382683;
    v[2] = 0.00923879;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = 1.00000;
    n3f(n);
    v[0] = 0.00000;
    v[1] = 0.00000;
    v[2] = 0.0100000;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00000;
    v[2] = 0.0100000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = 0.00000;
    v[2] = 0.0100000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00000;
    v[2] = 0.0100000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.00000;
    v[1] = -0.00382683;
    v[2] = 0.00923879;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.00382683;
    v[2] = 0.00923879;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = -0.00382683;
    v[2] = 0.00923879;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.00382683;
    v[2] = 0.00923879;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.707107;
    n[2] = 0.707107;
    n3f(n);
    v[0] = 0.00000;
    v[1] = -0.00707107;
    v[2] = 0.00707107;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.00707107;
    v[2] = 0.00707107;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = -0.00707107;
    v[2] = 0.00707107;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.00707107;
    v[2] = 0.00707107;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.923880;
    n[2] = 0.382683;
    n3f(n);
    v[0] = 0.00000;
    v[1] = -0.00923879;
    v[2] = 0.00382683;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.00923879;
    v[2] = 0.00382683;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = -0.00923879;
    v[2] = 0.00382683;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.00923879;
    v[2] = 0.00382683;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.00000;
    v[1] = -0.0100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.0100000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = -0.0100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.0100000;
    v[2] = 0.00000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.923880;
    n[2] = -0.382683;
    n3f(n);
    v[0] = 0.00000;
    v[1] = -0.00923879;
    v[2] = -0.00382683;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.00923879;
    v[2] = -0.00382683;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = -0.00923879;
    v[2] = -0.00382683;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.00923879;
    v[2] = -0.00382683;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.707107;
    n[2] = -0.707107;
    n3f(n);
    v[0] = 0.00000;
    v[1] = -0.00707107;
    v[2] = -0.00707107;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.00707107;
    v[2] = -0.00707107;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = -0.00707107;
    v[2] = -0.00707107;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.00707107;
    v[2] = -0.00707107;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.382683;
    n[2] = -0.923880;
    n3f(n);
    v[0] = 0.00000;
    v[1] = -0.00382683;
    v[2] = -0.00923879;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.00382683;
    v[2] = -0.00923879;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = -0.00382683;
    v[2] = -0.00923879;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.00382683;
    v[2] = -0.00923879;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = -1.00000;
    n3f(n);
    v[0] = 0.00000;
    v[1] = 0.00000;
    v[2] = -0.0100000;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00000;
    v[2] = -0.0100000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = 0.00000;
    v[2] = -0.0100000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00000;
    v[2] = -0.0100000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = -0.923880;
    n3f(n);
    v[0] = 0.00000;
    v[1] = 0.00382683;
    v[2] = -0.00923879;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00382683;
    v[2] = -0.00923879;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = 0.00382683;
    v[2] = -0.00923879;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00382683;
    v[2] = -0.00923879;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.707107;
    n[2] = -0.707107;
    n3f(n);
    v[0] = 0.00000;
    v[1] = 0.00707107;
    v[2] = -0.00707107;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00707107;
    v[2] = -0.00707107;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = 0.00707107;
    v[2] = -0.00707107;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00707107;
    v[2] = -0.00707107;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.923880;
    n[2] = -0.382683;
    n3f(n);
    v[0] = 0.00000;
    v[1] = 0.00923879;
    v[2] = -0.00382683;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00923879;
    v[2] = -0.00382683;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = 0.00923879;
    v[2] = -0.00382683;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00923879;
    v[2] = -0.00382683;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.00000;
    v[1] = 0.0100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.0100000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = 0.0100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.0100000;
    v[2] = 0.00000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.923880;
    n[2] = 0.382683;
    n3f(n);
    v[0] = 0.00000;
    v[1] = 0.00923879;
    v[2] = 0.00382683;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00923879;
    v[2] = 0.00382683;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = 0.00923879;
    v[2] = 0.00382683;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00923879;
    v[2] = 0.00382683;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.707107;
    n[2] = 0.707107;
    n3f(n);
    v[0] = 0.00000;
    v[1] = 0.00707107;
    v[2] = 0.00707107;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00707107;
    v[2] = 0.00707107;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 3.00000;
    v[1] = 0.00707107;
    v[2] = 0.00707107;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00707107;
    v[2] = 0.00707107;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.00000;
    v[1] = 0.00382683;
    v[2] = 0.00923879;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00382683;
    v[2] = 0.00923879;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 3.00000;
    v[1] = 0.00000;
    v[2] = 0.0100000;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.00382683;
    v[2] = 0.00923879;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.00707107;
    v[2] = 0.00707107;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.00923879;
    v[2] = 0.00382683;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.0100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.00923879;
    v[2] = -0.00382683;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.00707107;
    v[2] = -0.00707107;
    v3f(v);
    v[0] = 3.00000;
    v[1] = -0.00382683;
    v[2] = -0.00923879;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00000;
    v[2] = -0.0100000;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00382683;
    v[2] = -0.00923879;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00707107;
    v[2] = -0.00707107;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00923879;
    v[2] = -0.00382683;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.0100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00923879;
    v[2] = 0.00382683;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00707107;
    v[2] = 0.00707107;
    v3f(v);
    v[0] = 3.00000;
    v[1] = 0.00382683;
    v[2] = 0.00923879;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = -1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.00000;
    v[1] = 0.00000;
    v[2] = 0.0100000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.00382683;
    v[2] = 0.00923879;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.00707107;
    v[2] = 0.00707107;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.00923879;
    v[2] = 0.00382683;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.0100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.00923879;
    v[2] = -0.00382683;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.00707107;
    v[2] = -0.00707107;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.00382683;
    v[2] = -0.00923879;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00000;
    v[2] = -0.0100000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00382683;
    v[2] = -0.00923879;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00707107;
    v[2] = -0.00707107;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00923879;
    v[2] = -0.00382683;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.0100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00923879;
    v[2] = 0.00382683;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00707107;
    v[2] = 0.00707107;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.00382683;
    v[2] = 0.00923879;
    v3f(v);
    endpolygon();

    closeobj();
    SetVar(fingerWandToolClass, WANDTOOLPICTURE, NewInt(o));
#endif

    spaceMoveToolClass = NewObject(wandToolClass, 0L);
    AddToReferenceList(spaceMoveToolClass);
    SetMethod(spaceMoveToolClass, LEFTWANDDOWN, SpaceMoveToolLeftWandDown);
    SetMethod(spaceMoveToolClass, LEFTWANDUP, SpaceMoveToolLeftWandUp);
    SetMethod(spaceMoveToolClass, MIDDLEWANDDOWN, SpaceMoveToolMiddleWandDown);
    SetMethod(spaceMoveToolClass, MIDDLEWANDUP, SpaceMoveToolMiddleWandUp);
    SetMethod(spaceMoveToolClass, STILLDOWN, SpaceMoveToolStillDown);
#ifdef GRAPHICS
    o = genobj();
    makeobj(o);
    bgnpolygon();
    v[0] = 0.00000;
    v[1] = -0.500000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.200000;
    v[1] = -0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.300000;
    v[1] = -0.100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.300000;
    v[1] = -0.200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.500000;
    v[1] = 0.00000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.300000;
    v[1] = 0.200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.300000;
    v[1] = 0.100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.200000;
    v[1] = 0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = 0.500000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.200000;
    v[1] = 0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.100000;
    v[1] = 0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.100000;
    v[1] = 0.100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.300000;
    v[1] = 0.100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.300000;
    v[1] = 0.200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.500000;
    v[1] = 0.00000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.300000;
    v[1] = -0.200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.300000;
    v[1] = -0.100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.100000;
    v[1] = -0.100000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.100000;
    v[1] = -0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.200000;
    v[1] = -0.300000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    closeobj();
 
    SetVar(spaceMoveToolClass, CACHEGRAPHICS, NewInt(o));
#endif

    rakeWandToolClass = NewObject(wandToolClass, 0L);
    AddToReferenceList(rakeWandToolClass);
    SetMethod(rakeWandToolClass, DRAWARENA, DrawRakeWandTool);
    SetMethod(rakeWandToolClass, LEFTWANDDOWN, RakeToolLeftWandDown);

#ifdef GRAPHICS
    o = genobj();
    makeobj(o);
    bgnpolygon();
    v[0] = 0.0500000;
    v[1] = -0.500000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.0500000;
    v[1] = 0.150000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.200000;
    v[1] = 0.200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.250000;
    v[1] = 0.250000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.250000;
    v[1] = 0.500000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.200000;
    v[1] = 0.500000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.200000;
    v[1] = 0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.500000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.0500000;
    v[1] = 0.500000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.0500000;
    v[1] = 0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.0500000;
    v[1] = 0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.0500000;
    v[1] = 0.500000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.100000;
    v[1] = 0.500000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.100000;
    v[1] = 0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.200000;
    v[1] = 0.300000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.200000;
    v[1] = 0.500000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.250000;
    v[1] = 0.500000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.250000;
    v[1] = 0.250000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.200000;
    v[1] = 0.200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.100000;
    v[1] = 0.200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.0500000;
    v[1] = 0.150000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = -0.0500000;
    v[1] = -0.500000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    closeobj();
    SetVar(rakeWandToolClass, CACHEGRAPHICS, NewInt(o));
#endif

#ifdef GRAPHICS

    o = genobj();
    makeobj(o);
    bgnpolygon();
    c[0] = 255;
    c[1] = 0;
    c[2] = 255;
    c3s(c);
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = 1.00000;
    n3f(n);
    v[0] = 0.100000;
    v[1] = 0.00000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00000;
    v[2] = 0.0200000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = 0.00000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.00000;
    v[2] = 0.0200000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.100000;
    v[1] = -0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = -0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.707107;
    n[2] = 0.707107;
    n3f(n);
    v[0] = 0.100000;
    v[1] = -0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = -0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.923880;
    n[2] = 0.382683;
    n3f(n);
    v[0] = 0.100000;
    v[1] = -0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = -0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.100000;
    v[1] = -0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0200000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = -0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.0200000;
    v[2] = 0.00000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.923880;
    n[2] = -0.382683;
    n3f(n);
    v[0] = 0.100000;
    v[1] = -0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = -0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.707107;
    n[2] = -0.707107;
    n3f(n);
    v[0] = 0.100000;
    v[1] = -0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = -0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.382683;
    n[2] = -0.923880;
    n3f(n);
    v[0] = 0.100000;
    v[1] = -0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = -0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = -1.00000;
    n3f(n);
    v[0] = 0.100000;
    v[1] = 0.00000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00000;
    v[2] = -0.0200000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = 0.00000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.00000;
    v[2] = -0.0200000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = -0.923880;
    n3f(n);
    v[0] = 0.100000;
    v[1] = 0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = 0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.707107;
    n[2] = -0.707107;
    n3f(n);
    v[0] = 0.100000;
    v[1] = 0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = 0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.923880;
    n[2] = -0.382683;
    n3f(n);
    v[0] = 0.100000;
    v[1] = 0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = 0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.100000;
    v[1] = 0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0200000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = 0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.0200000;
    v[2] = 0.00000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.923880;
    n[2] = 0.382683;
    n3f(n);
    v[0] = 0.100000;
    v[1] = 0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = 0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.707107;
    n[2] = 0.707107;
    n3f(n);
    v[0] = 0.100000;
    v[1] = 0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = 0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.100000;
    v[1] = 0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.00000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = -1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.100000;
    v[1] = 0.00000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.00000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.100000;
    v[1] = 0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.382683;
    n[1] = 0.00000;
    n[2] = -0.923880;
    n3f(n);
    v[0] = 0.707654;
    v[1] = 0.120000;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.707654;
    v[1] = -0.120000;
    v[2] = -0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = -1.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.120000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.120000;
    v[2] = -0.0200000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = 0.120000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.120000;
    v[2] = -0.0200000;
    v3f(v);
    n[0] = -0.382683;
    n[1] = 0.00000;
    n[2] = -0.923880;
    n3f(n);
    v[0] = 0.692346;
    v[1] = -0.120000;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.692346;
    v[1] = 0.120000;
    v[2] = -0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.692346;
    v[1] = 0.120000;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.692346;
    v[1] = -0.120000;
    v[2] = -0.0184776;
    v3f(v);
    n[0] = -0.707107;
    n[1] = 0.00000;
    n[2] = -0.707107;
    n3f(n);
    v[0] = 0.685858;
    v[1] = -0.120000;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.685858;
    v[1] = 0.120000;
    v[2] = -0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.685858;
    v[1] = 0.120000;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.685858;
    v[1] = -0.120000;
    v[2] = -0.0141421;
    v3f(v);
    n[0] = -0.923880;
    n[1] = 0.00000;
    n[2] = -0.382683;
    n3f(n);
    v[0] = 0.681522;
    v[1] = -0.120000;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.681522;
    v[1] = 0.120000;
    v[2] = -0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.681522;
    v[1] = 0.120000;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.681522;
    v[1] = -0.120000;
    v[2] = -0.00765367;
    v3f(v);
    n[0] = -1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.680000;
    v[1] = -0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.680000;
    v[1] = 0.120000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.680000;
    v[1] = 0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.680000;
    v[1] = -0.120000;
    v[2] = 0.00000;
    v3f(v);
    n[0] = -0.923880;
    n[1] = 0.00000;
    n[2] = 0.382683;
    n3f(n);
    v[0] = 0.681522;
    v[1] = -0.120000;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.681522;
    v[1] = 0.120000;
    v[2] = 0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.681522;
    v[1] = 0.120000;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.681522;
    v[1] = -0.120000;
    v[2] = 0.00765367;
    v3f(v);
    n[0] = -0.707107;
    n[1] = 0.00000;
    n[2] = 0.707107;
    n3f(n);
    v[0] = 0.685858;
    v[1] = -0.120000;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.685858;
    v[1] = 0.120000;
    v[2] = 0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.685858;
    v[1] = 0.120000;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.685858;
    v[1] = -0.120000;
    v[2] = 0.0141421;
    v3f(v);
    n[0] = -0.382683;
    n[1] = 0.00000;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.692346;
    v[1] = -0.120000;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.692346;
    v[1] = 0.120000;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.692346;
    v[1] = 0.120000;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.692346;
    v[1] = -0.120000;
    v[2] = 0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = 1.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.120000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.120000;
    v[2] = 0.0200000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.700000;
    v[1] = 0.120000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.120000;
    v[2] = 0.0200000;
    v3f(v);
    n[0] = 0.382683;
    n[1] = 0.00000;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.707654;
    v[1] = -0.120000;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.707654;
    v[1] = 0.120000;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.707654;
    v[1] = 0.120000;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.707654;
    v[1] = -0.120000;
    v[2] = 0.0184776;
    v3f(v);
    n[0] = 0.707107;
    n[1] = 0.00000;
    n[2] = 0.707107;
    n3f(n);
    v[0] = 0.714142;
    v[1] = -0.120000;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.714142;
    v[1] = 0.120000;
    v[2] = 0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.714142;
    v[1] = 0.120000;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.714142;
    v[1] = -0.120000;
    v[2] = 0.0141421;
    v3f(v);
    n[0] = 0.923880;
    n[1] = 0.00000;
    n[2] = 0.382683;
    n3f(n);
    v[0] = 0.718478;
    v[1] = -0.120000;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.718478;
    v[1] = 0.120000;
    v[2] = 0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.718478;
    v[1] = 0.120000;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.718478;
    v[1] = -0.120000;
    v[2] = 0.00765367;
    v3f(v);
    n[0] = 1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.720000;
    v[1] = -0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.720000;
    v[1] = 0.120000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.720000;
    v[1] = 0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.720000;
    v[1] = -0.120000;
    v[2] = 0.00000;
    v3f(v);
    n[0] = 0.923880;
    n[1] = 0.00000;
    n[2] = -0.382683;
    n3f(n);
    v[0] = 0.718478;
    v[1] = -0.120000;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.718478;
    v[1] = 0.120000;
    v[2] = -0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.718478;
    v[1] = 0.120000;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.718478;
    v[1] = -0.120000;
    v[2] = -0.00765367;
    v3f(v);
    n[0] = 0.707107;
    n[1] = 0.00000;
    n[2] = -0.707107;
    n3f(n);
    v[0] = 0.714142;
    v[1] = -0.120000;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.714142;
    v[1] = 0.120000;
    v[2] = -0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.714142;
    v[1] = 0.120000;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.714142;
    v[1] = -0.120000;
    v[2] = -0.0141421;
    v3f(v);
    n[0] = 0.382683;
    n[1] = 0.00000;
    n[2] = -0.923880;
    n3f(n);
    v[0] = 0.707654;
    v[1] = -0.120000;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.707654;
    v[1] = 0.120000;
    v[2] = -0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.00000;
    n[1] = 1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.120000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.692346;
    v[1] = 0.120000;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.685858;
    v[1] = 0.120000;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.681522;
    v[1] = 0.120000;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.680000;
    v[1] = 0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.681522;
    v[1] = 0.120000;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.685858;
    v[1] = 0.120000;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.692346;
    v[1] = 0.120000;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.120000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.707654;
    v[1] = 0.120000;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.714142;
    v[1] = 0.120000;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.718478;
    v[1] = 0.120000;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.720000;
    v[1] = 0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.718478;
    v[1] = 0.120000;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.714142;
    v[1] = 0.120000;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.707654;
    v[1] = 0.120000;
    v[2] = -0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.00000;
    n[1] = -1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.120000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.692346;
    v[1] = -0.120000;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.685858;
    v[1] = -0.120000;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.681522;
    v[1] = -0.120000;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.680000;
    v[1] = -0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.681522;
    v[1] = -0.120000;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.685858;
    v[1] = -0.120000;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.692346;
    v[1] = -0.120000;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.120000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.707654;
    v[1] = -0.120000;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.714142;
    v[1] = -0.120000;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.718478;
    v[1] = -0.120000;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.720000;
    v[1] = -0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.718478;
    v[1] = -0.120000;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.714142;
    v[1] = -0.120000;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.707654;
    v[1] = -0.120000;
    v[2] = -0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.800000;
    v[1] = -0.0923463;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0923463;
    v[2] = 0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = 1.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.100000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.100000;
    v[2] = 0.0200000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.100000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.100000;
    v[2] = 0.0200000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.107654;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.107654;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.107654;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.107654;
    v[2] = 0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.707107;
    n[2] = 0.707107;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.114142;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.114142;
    v[2] = 0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.114142;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.114142;
    v[2] = 0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.923880;
    n[2] = 0.382683;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.118478;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.118478;
    v[2] = 0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.118478;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.118478;
    v[2] = 0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.120000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.120000;
    v[2] = 0.00000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.923880;
    n[2] = -0.382683;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.118478;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.118478;
    v[2] = -0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.118478;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.118478;
    v[2] = -0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.707107;
    n[2] = -0.707107;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.114142;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.114142;
    v[2] = -0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.114142;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.114142;
    v[2] = -0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.382683;
    n[2] = -0.923880;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.107654;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.107654;
    v[2] = -0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.107654;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.107654;
    v[2] = -0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = -1.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.100000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.100000;
    v[2] = -0.0200000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.100000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.100000;
    v[2] = -0.0200000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = -0.923880;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.0923463;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0923463;
    v[2] = -0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.0923463;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0923463;
    v[2] = -0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.707107;
    n[2] = -0.707107;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.0858579;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0858579;
    v[2] = -0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.0858579;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0858579;
    v[2] = -0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.923880;
    n[2] = -0.382683;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.0815224;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0815224;
    v[2] = -0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.0815224;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0815224;
    v[2] = -0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.0800000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0800000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.0800000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0800000;
    v[2] = 0.00000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.923880;
    n[2] = 0.382683;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.0815224;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0815224;
    v[2] = 0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.0815224;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0815224;
    v[2] = 0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.707107;
    n[2] = 0.707107;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.0858579;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0858579;
    v[2] = 0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.0858579;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0858579;
    v[2] = 0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.0923463;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0923463;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.800000;
    v[1] = -0.100000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.107654;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.114142;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.118478;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.118478;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.114142;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.107654;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.100000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0923463;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0858579;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0815224;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0800000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0815224;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0858579;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0923463;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = -1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.100000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.107654;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.114142;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.118478;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.118478;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.114142;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.107654;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.100000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0923463;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0858579;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0815224;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0800000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0815224;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0858579;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0923463;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.800000;
    v[1] = 0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = 1.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.00000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.00000;
    v[2] = 0.0200000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.00000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00000;
    v[2] = 0.0200000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.707107;
    n[2] = 0.707107;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.923880;
    n[2] = 0.382683;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0200000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0200000;
    v[2] = 0.00000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.923880;
    n[2] = -0.382683;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.707107;
    n[2] = -0.707107;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.382683;
    n[2] = -0.923880;
    n3f(n);
    v[0] = 0.700000;
    v[1] = -0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = -0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = -1.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.00000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.00000;
    v[2] = -0.0200000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.00000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00000;
    v[2] = -0.0200000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = -0.923880;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.707107;
    n[2] = -0.707107;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.923880;
    n[2] = -0.382683;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0200000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0200000;
    v[2] = 0.00000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.923880;
    n[2] = 0.382683;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.707107;
    n[2] = 0.707107;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.800000;
    v[1] = 0.00000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = -0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.00000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = -1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.00000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = -0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00765367;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0141421;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0184776;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0200000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0184776;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0141421;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.00765367;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.800000;
    v[1] = 0.107654;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.107654;
    v[2] = 0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = 1.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.100000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.100000;
    v[2] = 0.0200000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.100000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.100000;
    v[2] = 0.0200000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.0923463;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0923463;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.0923463;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0923463;
    v[2] = 0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.707107;
    n[2] = 0.707107;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.0858579;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0858579;
    v[2] = 0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.0858579;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0858579;
    v[2] = 0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.923880;
    n[2] = 0.382683;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.0815224;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0815224;
    v[2] = 0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.0815224;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0815224;
    v[2] = 0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.0800000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0800000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.0800000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0800000;
    v[2] = 0.00000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.923880;
    n[2] = -0.382683;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.0815224;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0815224;
    v[2] = -0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.0815224;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0815224;
    v[2] = -0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.707107;
    n[2] = -0.707107;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.0858579;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0858579;
    v[2] = -0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.0858579;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0858579;
    v[2] = -0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -0.382683;
    n[2] = -0.923880;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.0923463;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0923463;
    v[2] = -0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.0923463;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0923463;
    v[2] = -0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = -1.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.100000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.100000;
    v[2] = -0.0200000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.100000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.100000;
    v[2] = -0.0200000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = -0.923880;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.107654;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.107654;
    v[2] = -0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.107654;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.107654;
    v[2] = -0.0184776;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.707107;
    n[2] = -0.707107;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.114142;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.114142;
    v[2] = -0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.114142;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.114142;
    v[2] = -0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.923880;
    n[2] = -0.382683;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.118478;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.118478;
    v[2] = -0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.118478;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.118478;
    v[2] = -0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.120000;
    v[2] = 0.00000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.120000;
    v[2] = 0.00000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.923880;
    n[2] = 0.382683;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.118478;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.118478;
    v[2] = 0.00765367;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.118478;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.118478;
    v[2] = 0.00765367;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.707107;
    n[2] = 0.707107;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.114142;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.114142;
    v[2] = 0.0141421;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.800000;
    v[1] = 0.114142;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.114142;
    v[2] = 0.0141421;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 0.382683;
    n[2] = 0.923880;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.107654;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.107654;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.800000;
    v[1] = 0.100000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0923463;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0858579;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0815224;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0800000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0815224;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0858579;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.0923463;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.100000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.107654;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.114142;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.118478;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.118478;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.114142;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.800000;
    v[1] = 0.107654;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = -1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.700000;
    v[1] = 0.100000;
    v[2] = 0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0923463;
    v[2] = 0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0858579;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0815224;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0800000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0815224;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0858579;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.0923463;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.100000;
    v[2] = -0.0200000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.107654;
    v[2] = -0.0184776;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.114142;
    v[2] = -0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.118478;
    v[2] = -0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.120000;
    v[2] = 0.00000;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.118478;
    v[2] = 0.00765367;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.114142;
    v[2] = 0.0141421;
    v3f(v);
    v[0] = 0.700000;
    v[1] = 0.107654;
    v[2] = 0.0184776;
    v3f(v);
    endpolygon();

    closeobj();

    SetVar(rakeWandToolClass, WANDTOOLPICTURE, NewInt(o));
#endif

    wandClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(wandClass);
    loc[0] = loc[1] = loc[2] = 0.0;
    var = NewRealArray(1, 3L);
    CArray2Array(var, loc);
    SetVar(wandClass, LOCATION, var);
    list = NewList();
    SetVar(wandClass, EDITTOOL, tool = NewObject(fingerWandToolClass, 0L));
    PostfixList(list, tool);
    tool = NewObject(spaceMoveToolClass, 0L);
    PostfixList(list, tool);
    tool = NewObject(rakeWandToolClass, 0L);
    PostfixList(list, tool);
    SetVar(wandClass, ALLTOOLS, list);

    quat[0] = -SQ22;
    quat[1] = 0.0;
    quat[2] = 0.0;
    quat[3] = SQ22;
    var = NewRealArray(1, 4L);
    CArray2Array(var, quat);
    SetVar(wandClass, ROTQUAT, var);

    SetMethod(wandClass, STILLDOWN, WandStillDown);
    SetMethod(wandClass, DRAWARENA, DrawArenaWand);
    SetMethod(wandClass, STARTMENU, StartWandMenu);
    SetMethod(wandClass, STOPMENU, StopWandMenu);
    SetMethod(wandClass, LEFTWANDDOWN, WandLeftWandDown);
    SetMethod(wandClass, LEFTWANDUP, WandLeftWandUp);
    SetMethod(wandClass, MIDDLEWANDDOWN, WandMiddleWandDown);
    SetMethod(wandClass, MIDDLEWANDUP, WandMiddleWandUp);

    arenaClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(arenaClass);
    SetMethod(arenaClass, DRAW, DrawArena);
    SetMethod(arenaClass, BOUNDSINVALID, ArenaBoundsInvalid);
    SetMethod(arenaClass, OBSERVER, MakeArenaObserver);
    SetMethod(arenaClass, LIGHTS, MakeArenaLights);
    SetMethod(arenaClass, WAND, MakeArenaWand);
    SetVar(arenaClass, SHOWMOTGUIDES, ObjTrue);
    SetMethod(arenaClass, TURNDIAL, TurnDialArena);
    SetMethod(arenaClass, KEYDOWN, KeyDownArena);
    DeclareDependency(arenaClass, APPEARANCE, CONTENTS);
    DeclareDependency(arenaClass, APPEARANCE, LIGHTS);
    DeclareDependency(arenaClass, APPEARANCE, OBSERVER);
    DeclareDependency(arenaClass, APPEARANCE, WAND);
    SetMethod(arenaClass, APPEARANCE, MakeArenaAppearance);
    SetMethod(arenaClass, MOVEWANDTO, MoveArenaWandTo);
    SetMethod(arenaClass, ROTATEWANDTO, RotateArenaWandTo);
    SetMethod(arenaClass, WANDINTERACT, ArenaWandInteract);

    arenaWindowClass = (WinInfoPtr) NewObject((ObjPtr) objWindowClass, 0L);
    AddToReferenceList(arenaWindowClass);
    SetMethod((ObjPtr) arenaWindowClass, MAKEDRAWN, MakeArenaWindowDrawn);

#ifdef GRAPHICS
    matIndex = 0;
    material[matIndex++] = DIFFUSE;
    material[matIndex++] = 0.0;
    material[matIndex++] = 1.;
    material[matIndex++] = 0.0;
    material[matIndex++] = LMNULL;
    lmdef(DEFMATERIAL, ARENAMATERIAL, matIndex, material);

#include "ScianGeo3DDataset.h"
    SetVar(icon3DScalar, CACHEGRAPHICS, NewInt(o));
    SetVar(icon3DVector, CACHEGRAPHICS, NewInt(o));
#include "ScianGeo2DDataset.h"
    SetVar(icon2DScalar, CACHEGRAPHICS, NewInt(o));
    SetVar(icon2DVector, CACHEGRAPHICS, NewInt(o));

#endif
}

#ifdef HAVE_PROTOTYPES
void KillArenas(void)
#else
void KillArenas()
#endif
/*Kills the arena system*/
{
    RemoveFromReferenceList(arenaWindowClass);
    RemoveFromReferenceList(arenaClass);
    RemoveFromReferenceList(fingerWandToolClass);
    RemoveFromReferenceList(rakeWandToolClass);
    RemoveFromReferenceList(spaceMoveToolClass);
    RemoveFromReferenceList(wandToolClass);
    RemoveFromReferenceList(wandClass);
    RemoveFromReferenceList(wallClass);
}
